// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package common;

import javax.media.opengl.GL2;

import model.Coordinate;

/**
 * Rendering arrows in a Gl context.
 * Color must be set before calling the methods
 * The class is not thread safe!
 */
public class ArrowRenderer {
	private final static float[][] p1 = new float[8][3];
	private final static float[][] p2 = new float[8][3];
	private final static float[][] normals = new float[8][3];
	private final static float ONE_OVER_SQRT2 = 0.7071067811865475727f;
	
	/**
	 * Renders an arrow with a standard head size of 3 
	 * For very short arrows, the thickness is automatically adjusted if too large
	 * @param gl 
	 * @param origin Coordinate to start the arrow from 
	 * @param direction vector that defines direction and 
	 * @param thickness arrow shaft thickness
	 */
	public static void renderArrow(GL2 gl, Coordinate origin, float[] direction, float thickness){
		renderArrow(gl, origin, direction, thickness, 3f);
	}
	
	/**
	 * Renders an arrow with a variable scaling of head to shaft thickness
	 * For very short arrows, the thickness is automatically adjusted if too large
	 * @param gl 
	 * @param origin Coordinate to start the arrow from 
	 * @param direction vector that defines direction and 
	 * @param thickness arrow shaft thickness
	 * @param headThickScale scaling factor between head an shaft thickness
	 */
	public static void renderArrow(GL2 gl, Coordinate origin, float[] direction, float thickness, float headThickScale){
		float lenght = VectorOps.length(direction);
		if (thickness > lenght * 2) thickness = lenght * .1f;
		if (lenght < 0.001f) return;
		float headThickness = headThickScale*thickness;
		
		float[] u = new float[3];
		
		if (Math.abs(direction[0]) >= Math.abs(direction[1])){
			u[0] = -direction[2];
			u[1] = 0f;
			u[2] = direction[0];
		} else {
			u[0] = 0f;
			u[1] = direction[2];
			u[2] = -direction[1];
		}
		
		u = VectorOps.normalize(u);
        float[]	v = VectorOps.crossProduct(u, direction);
		v = VectorOps.normalize(v);
		
		normals[0][0] = u[0]; normals[0][1] = u[1]; normals[0][2] = u[2];
		normals[1][0] = (u[0]+v[0]) * ONE_OVER_SQRT2; normals[1][1] = (u[1]+v[1]) * ONE_OVER_SQRT2; normals[1][2] = (u[2]+v[2]) * ONE_OVER_SQRT2;
		normals[2][0] = v[0]; normals[2][1] = v[1]; normals[2][2] = v[2];
		normals[3][0] = (-u[0]+v[0]) * ONE_OVER_SQRT2;normals[3][1] = (-u[1]+v[1]) * ONE_OVER_SQRT2; normals[3][2] = (-u[2]+v[2]) * ONE_OVER_SQRT2; 
		normals[4][0] = -u[0]; normals[4][1] = -u[1]; normals[4][2] = -u[2];
		normals[5][0] = (-u[0]-v[0]) * ONE_OVER_SQRT2; normals[5][1] = (-u[1]-v[1]) * ONE_OVER_SQRT2; normals[5][2] = (-u[2]-v[2]) * ONE_OVER_SQRT2;	
		normals[6][0] = -v[0]; normals[6][1] = -v[1]; normals[6][2] = -v[2];
		normals[7][0] = (u[0]-v[0]) * ONE_OVER_SQRT2; normals[7][1] = (u[1]-v[1]) * ONE_OVER_SQRT2; normals[7][2] = (u[2]-v[2]) * ONE_OVER_SQRT2;
		
		float[] norm_dir = VectorOps.normalize(direction);
		float[] shift = VectorOps.multiply(norm_dir, lenght-2*headThickness);
		
		for (int i=0; i<8; i++){
			p1[i][0] = origin.getX() + normals[i][0]*thickness;
			p1[i][1] = origin.getY() + normals[i][1]*thickness;
			p1[i][2] = origin.getZ() + normals[i][2]*thickness;
			
			p2[i][0] = origin.getX() + normals[i][0]*headThickness + shift[0];
			p2[i][1] = origin.getY() + normals[i][1]*headThickness + shift[1];
			p2[i][2] = origin.getZ() + normals[i][2]*headThickness + shift[2];
		}
		
		gl.glBegin(GL2.GL_TRIANGLE_STRIP);
		for (int i=0; i<8; i++){
			gl.glNormal3f(normals[i][0], normals[i][1], normals[i][2]);
			gl.glVertex3f(p1[i][0], p1[i][1], p1[i][2]);
			gl.glVertex3f(p1[i][0]+shift[0], p1[i][1]+shift[1], p1[i][2]+shift[2]);
		}
		gl.glNormal3f(normals[0][0], normals[0][1], normals[0][2]);
		gl.glVertex3f(p1[0][0], p1[0][1], p1[0][2]);
		gl.glVertex3f(p1[0][0]+shift[0], p1[0][1]+shift[1], p1[0][2]+shift[2]);
		gl.glEnd();
		
		
		//Bottom plate at origin
		gl.glBegin(GL2.GL_TRIANGLE_FAN);
		gl.glNormal3f(-norm_dir[0], -norm_dir[1], -norm_dir[2]);
		gl.glVertex3f(origin.getX(), origin.getY(), origin.getZ());
		gl.glVertex3f(p1[0][0], p1[0][1], p1[0][2]);
		gl.glVertex3f(p1[1][0], p1[1][1], p1[1][2]);
		gl.glVertex3f(p1[2][0], p1[2][1], p1[2][2]);
		gl.glVertex3f(p1[3][0], p1[3][1], p1[3][2]);
		gl.glVertex3f(p1[4][0], p1[4][1], p1[4][2]);
		gl.glVertex3f(p1[5][0], p1[5][1], p1[5][2]);
		gl.glVertex3f(p1[6][0], p1[6][1], p1[6][2]);
		gl.glVertex3f(p1[7][0], p1[7][1], p1[7][2]);
		gl.glVertex3f(p1[0][0], p1[0][1], p1[0][2]);
		gl.glEnd();
		
		//Arrow cap bottom
		gl.glBegin(GL2.GL_TRIANGLE_FAN);
		gl.glNormal3f(-norm_dir[0], -norm_dir[1], -norm_dir[2]);
		gl.glVertex3f(origin.getX()+shift[0], origin.getY()+shift[1], origin.getZ()+shift[2]);
		gl.glVertex3f(p2[0][0], p2[0][1], p2[0][2]);
		gl.glVertex3f(p2[1][0], p2[1][1], p2[1][2]);
		gl.glVertex3f(p2[2][0], p2[2][1], p2[2][2]);
		gl.glVertex3f(p2[3][0], p2[3][1], p2[3][2]);
		gl.glVertex3f(p2[4][0], p2[4][1], p2[4][2]);
		gl.glVertex3f(p2[5][0], p2[5][1], p2[5][2]);
		gl.glVertex3f(p2[6][0], p2[6][1], p2[6][2]);
		gl.glVertex3f(p2[7][0], p2[7][1], p2[7][2]);
		gl.glVertex3f(p2[0][0], p2[0][1], p2[0][2]);
		gl.glEnd();
		
		
		//Arrow cap
		gl.glBegin(GL2.GL_TRIANGLE_FAN);
		gl.glNormal3f(norm_dir[0], norm_dir[1], norm_dir[2]);
		gl.glVertex3f(origin.getX()+direction[0], origin.getY()+direction[1], origin.getZ()+direction[2]);
		gl.glNormal3f(normals[0][0], normals[0][1], normals[0][2]);
		gl.glVertex3f(p2[0][0], p2[0][1], p2[0][2]);
		gl.glNormal3f(normals[7][0], normals[7][1], normals[7][2]);
		gl.glVertex3f(p2[7][0], p2[7][1], p2[7][2]);
		gl.glNormal3f(normals[6][0], normals[6][1], normals[6][2]);
		gl.glVertex3f(p2[6][0], p2[6][1], p2[6][2]);
		gl.glNormal3f(normals[5][0], normals[5][1], normals[5][2]);
		gl.glVertex3f(p2[5][0], p2[5][1], p2[5][2]);
		gl.glNormal3f(normals[4][0], normals[4][1], normals[4][2]);
		gl.glVertex3f(p2[4][0], p2[4][1], p2[4][2]);
		gl.glNormal3f(normals[3][0], normals[3][1], normals[3][2]);
		gl.glVertex3f(p2[3][0], p2[3][1], p2[3][2]);
		gl.glNormal3f(normals[2][0], normals[2][1], normals[2][2]);
		gl.glVertex3f(p2[2][0], p2[2][1], p2[2][2]);
		gl.glNormal3f(normals[1][0], normals[1][1], normals[1][2]);
		gl.glVertex3f(p2[1][0], p2[1][1], p2[1][2]);
		gl.glNormal3f(normals[0][0], normals[0][1], normals[0][2]);
		gl.glVertex3f(p2[0][0], p2[0][1], p2[0][2]);
		gl.glEnd();
	}
}
