// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package crystalStructures;

import java.util.*;

import common.Tupel;
import crystalStructures.CrystalStructureProperties.CrystalProperty;

import model.*;
import model.BurgersVector.BurgersVectorType;
import model.polygrain.Grain;
import model.polygrain.mesh.Mesh;
import model.skeletonizer.processors.*;
import model.skeletonizer.processors.BurgersVectorAnalyzer.ClassificationPattern;

public class B2NiTi extends BCCStructure{

	private static final ArrayList<ClassificationPattern> bvClassifcationPattern = new ArrayList<ClassificationPattern>();
	
	static{
		//1/2<111>
		bvClassifcationPattern.add(new ClassificationPattern(111, 2, 4, 111, 2, BurgersVectorType.PARTIAL));
		//<100>
		bvClassifcationPattern.add(new ClassificationPattern(100, 1, 2, 100, 1, BurgersVectorType.PERFECT));
		//1/2<111> identified as 1/4<112>
		bvClassifcationPattern.add(new ClassificationPattern(112, 4, 4, 111, 2, BurgersVectorType.PARTIAL));
	}
	
	public B2NiTi() {
		Iterator<CrystalProperty> iter = crystalProperties.iterator();
		while (iter.hasNext()){
			CrystalProperty cp = iter.next();
			if (cp.id.equals("gbFilterDist"))
				iter.remove();
		}
	}
	
	@Override
	public float getGrainBoundaryFilterDistance() {
		return 0f;
	}
	
	private NearestNeighborFilter identMartensitePhaseFilter = new NearestNeighborFilter() {
		@Override
		public boolean accept(Atom a) {
			if (( (a.getType() == 3) &&
					(a.getRBV()==null || a.getRBV().getSquaredRBVLengh() < latticeConstant*latticeConstant*0.15f)))
				return true;
			return false;
		}
	};
	
	@Override
	protected CrystalStructure deriveNewInstance() {
		return new B2NiTi();
	}
	
	@Override
	protected String getIDName() {
		return "NiTi";
	}
	
	@Override
	public int identifyAtomType(Atom atom, NearestNeighborBuilder<Atom> nnb) {
		ArrayList<Tupel<Atom, float[]>> d_cont  = nnb.getNeighAndNeighVec(atom);
		if (d_cont.size()<=10) return 6;
				
		int co_x0_same = 0;
		int co_x0_other = 0;
//		int co_x1_other = 0;
//		int co_x2_other = 0;
		int co_x3_other = 0;
		int co_x4_other = 0;
		int co_x5_other = 0;
		int co_x6_other = 0;
		int co_x7_other = 0;
		int num_same = 0;
		int num_other = 0;
		
		for (int i = 0; i < d_cont.size(); i++) {
			int type_i = d_cont.get(i).getO1().getElement()%2;
			
			float v_x = d_cont.get(i).getO2()[0];
			float v_y = d_cont.get(i).getO2()[1];
			float v_z = d_cont.get(i).getO2()[2];
			
			float v_length = (float)Math.sqrt(v_x*v_x+v_y*v_y+v_z*v_z);
						
			if (type_i == (atom.getElement()%2)) num_same++;
			else num_other++;
			
			for (int j = 0; j < i; j++) {
				int type_j = d_cont.get(j).getO1().getElement()%2;
				if (type_i == type_j){
					float u_x = d_cont.get(j).getO2()[0];
					float u_y = d_cont.get(j).getO2()[1];
					float u_z = d_cont.get(j).getO2()[2];
					
					float u_length = (float)Math.sqrt(u_x*u_x+u_y*u_y+u_z*u_z);
					float a = v_x*u_x+v_y*u_y+v_z*u_z;
					a /= (v_length*u_length);
					
					boolean same = type_i == atom.getElement()%2;
					
					if (same){
						if (a < -.95) co_x0_same++;
					} else {
						if (a < -.95) co_x0_other++;
						else if (a < -.90){}// co_x1_other++;
						else if (a < -.60){}// co_x2_other++;
						else if (a < -.30) co_x3_other++;
						else if (a < -.20) co_x4_other++;
						else if (a < .20) co_x5_other++;
						else if (a < 0.45) co_x6_other++;
						else if (a < 0.7) co_x7_other++;
					}
				}
			}
		}
		
		if (num_same==6 && num_other == 8 && co_x0_other == 2 && co_x0_same == 1 && co_x3_other+co_x4_other>=10) return 5;
		if ( atom.getElement()%2 == 0 && num_same >= 5 && num_other>=7 && co_x5_other+co_x7_other > 3 && co_x3_other<12 && co_x6_other<12) return 3;  //MartTi
		if ( atom.getElement()%2 == 1 && num_same >= 5 && num_other>=7 && co_x3_other <= co_x4_other+co_x5_other && co_x6_other<12) return 3;  //MartNi
		if (num_same == 6 && num_other>=num_same && (co_x3_other >= 7 || (co_x3_other == 6 && co_x6_other==12)) && co_x6_other > 9 && co_x0_other >= 3 && co_x0_same >= 2) return 0;  //B2
		
		return 4;
	}
	
	@Override
	public float[] getSphereSizeScalings(){
		float[] size = new float[getNumberOfElements()];
		size[0] = 1f;
		size[1] = 0.85f;
		return size;
	}
	
	public float getPerfectBurgersVectorLength(){
		return latticeConstant;
	}
	
	@Override
	public ArrayList<Atom> getDislocationDefectAtoms(AtomData data) {
		float minRBVLength = latticeConstant*0.35f;
		minRBVLength *= minRBVLength;
		float maxRBVLength = latticeConstant*3f;
		maxRBVLength *= maxRBVLength;
		
		ArrayList<Atom> defectAtoms = new ArrayList<Atom>();
		for (Atom a : data.getAtoms()) {
			int type = a.getType();
			if (type > 1 && type < 6){
				if (data.isRbvAvailable()){
					if (a.getRBV()!=null && a.getRBV().getSquaredRBVLengh()>minRBVLength && a.getRBV().getSquaredRBVLengh()<maxRBVLength)
						defectAtoms.add(a);
				}
				else defectAtoms.add(a);
			}
		}
		return defectAtoms;
	}
	
	@Override
	public TreeSet<Atom> getStackingFaultAtoms(AtomData data){
		return new TreeSet<Atom>();
	}
	
	@Override
	public String getNameForType(int i) {
		switch (i) {
			case 0: return "B2";
			case 1: return "Unused";
			case 2: return "Unused";
			case 3: return "Possible martensite";
			case 4: return "Defect";
			case 5: return "APB";
			case 6: return "free surface";
			case 7: return "Unused";
			default: return "unknown";
		}
	}
	
	@Override
	public ArrayList<ClassificationPattern> getBurgersVectorClassificationPattern() {
		return bvClassifcationPattern;
	}
	
	@Override
	public int getNumberOfElements(){
		return 2;
	}
	
	public List<Grain> identifyGrains(AtomData data) {
		List<Grain> grains = new Vector<Grain>();
		
		//In case the grain column is not used, but grain boundaries are given in the metadata, enable the function
		//nasty workaround for some older files
		boolean importedGrainsOnly = false;
		Object o = data.getFileMetaData("grain");
		PolygrainMetadata meta = null;
		if (o!=null && o instanceof PolygrainMetadata)
			meta = (PolygrainMetadata)o;
		if (meta != null && meta.meshes.size() == meta.grainOrientation.size()) importedGrainsOnly = true;
		
		if (!data.isGrainsImported() && !importedGrainsOnly){
			//create a nearest neighbor mesh for possible martensite atoms
			data.findNearestNeighbors(getNearestNeighborSearchRadius(), identMartensitePhaseFilter);
			//Detect grains from data
			TreeSet<Atom> notAssignedAtoms = new TreeSet<Atom>();
			for (Atom a : data.getAtoms()) {
				//if ( mayBePartOfSecondPhase(a) )  notAssignedAtoms.add(a);
				//Only possible martensite atoms have nearest neighbors
				if ( a.getNeigh()!=null ) notAssignedAtoms.add(a);
			}
			while (!notAssignedAtoms.isEmpty()) {
				Atom a = notAssignedAtoms.pollFirst();
				if (a.getNeigh().length <= 9) continue;
				
				ArrayList<Atom> grain = new ArrayList<Atom>();
				//From each item in the grain array the bfs is continued
				//There is no need to keep a second list, the index of processed
				//atoms is enough
				int queuePosition = 0;
				int grainID = grains.size();
				grain.add(a);
				a.setGrain(grainID);
				
				while (queuePosition < grain.size()) {
					Atom c = grain.get(queuePosition++);
					
					// in case some same element neighbors exists, continue a breath
					// first search at this atoms
					if (c.getNeigh().length > 9) {
						for (int i = 0; i < c.getNeigh().length; i++) {
							if (c.getNeigh()[i].getGrain() == Atom.DEFAULT_GRAIN) {
								c.getNeigh()[i].setGrain(grainID);
								grain.add(c.getNeigh()[i]);
								notAssignedAtoms.remove(c.getNeigh()[i]);
							}
						}
					}
				}
	
				if (grain.size() > 20) {
					CrystalStructure secondPhase = new MonoclinicNiTi(this.getLatticeConstant(),
							this.getNearestNeighborSearchRadius());
					Mesh mesh = new Mesh(grain, secondPhase);
					Grain g = new Grain(mesh, grain, grains.size(), secondPhase, data.getBounds());
					grains.add(g);
				} else {
					for (int i=0; i<grain.size(); i++) {
						grain.get(i).setGrain(Atom.DEFAULT_GRAIN);
					}
				}
			}
		} else {
			if (meta != null && meta.meshes.size() == meta.grainOrientation.size()){
				for (int i : meta.meshes.keySet()){
					CrystalStructure monoclinicNiTi = new MonoclinicNiTi(this.getLatticeConstant(),
							this.getNearestNeighborSearchRadius());
					Mesh mesh = new Mesh(meta.meshes.get(i).triangles, meta.meshes.get(i).vertices);
					Grain g;
					if (meta!=null && meta.grainOrientation.containsKey(i))
						g = new Grain(mesh, new ArrayList<Atom>(), i, monoclinicNiTi, meta.grainOrientation.get(i));
					else g = new Grain(mesh, new ArrayList<Atom>(), i, monoclinicNiTi, data.getBounds());
					grains.add(g);
				}
			}
			else {
				//Grains are given in input, just process the sets
				HashMap<Integer, ArrayList<Atom>> grainsAtomLists = new HashMap<Integer, ArrayList<Atom>>();
				for (int i=0; i<data.getAtoms().size(); i++){
					Atom a = data.getAtoms().get(i);
					if (a.getGrain() != Atom.IGNORED_GRAIN && a.getGrain() != Atom.DEFAULT_GRAIN){	
						if (!grainsAtomLists.containsKey(a.getGrain())){
							grainsAtomLists.put(a.getGrain(), new ArrayList<Atom>());
						}
						grainsAtomLists.get(a.getGrain()).add(a);
					}
				}
				
				//Mesh creation
				for (Integer i : grainsAtomLists.keySet()){
					ArrayList<Atom> grain = grainsAtomLists.get(i);
					if (grain.size() > 20){
						CrystalStructure monoclinicNiTi = new MonoclinicNiTi(this.getLatticeConstant(),
								this.getNearestNeighborSearchRadius());
						Mesh mesh = new Mesh(grain, monoclinicNiTi);
						Grain g;
						if (meta!=null && meta.grainOrientation.containsKey(i))
							g = new Grain(mesh, grain, i, monoclinicNiTi, meta.grainOrientation.get(i));
						else g = new Grain(mesh, grain, i, monoclinicNiTi, data.getBounds());
						grains.add(g);
					}
				}
			}
		}
		return grains;
	}
	
	public float getSkeletonizationMeshingThreshold(){
		return getNearestNeighborSearchRadius()*1.28f;
	}
	
	@Override
	public List<SkeletonPreprocessor> getSkeletonizerPreProcessors(){
		Vector<SkeletonPreprocessor>  list = new Vector<SkeletonPreprocessor>();
		
		list.add(new RBVAngleFilterPreprocessor(45));
		
		list.add(new MeshLineSenseCenteringPreprocessor());
		list.add(new ReMeshingPreprocessor());
		list.add(new RBVAngleFilterPreprocessor(45));
		
		list.add(new MeshCleaningPreprocessor());
		
		return list;
	}
	
	@Override
	public boolean skeletonizeOverMultipleGrains(){
		return true;
	}
	
	@Override
	public boolean createRBVbeforeGrains(){
		return true;
	}
	
	public boolean orderGrainsBySize(){
		return true;
	}
}
