// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package crystalStructures;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.Collection;
import java.util.Properties;

import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;


public class CrystalStructureProperties {
	
	public static JPanel createPropertyContainer(Collection<CrystalProperty> cp){
		JPanel configPanel = new JPanel(new GridBagLayout());
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.fill = GridBagConstraints.BOTH;
		gbc.gridx = 0; gbc.gridy = 0;
		gbc.weightx = 1;
		
		configPanel.setBorder(new TitledBorder(new EtchedBorder(1), "Crystal structure options"));
		for (CrystalProperty c : cp){
			configPanel.add(new JLabel(c.label), gbc); gbc.gridx++;
			configPanel.add(c.getEditor(), gbc);  gbc.gridx = 0;
			gbc.gridy++;
		}
		return configPanel;
	}
	
	public static void readProperties(Collection<CrystalProperty> cp, Reader r) throws IOException{
		Properties prop = new Properties();
		prop.load(r);
		for (CrystalProperty c : cp){
			c.load(prop);
		}
	}
	
	public static void storeProperties(Collection<CrystalProperty> cp, Writer w) throws IOException{
		Properties prop = new Properties();
		for (CrystalProperty c : cp){
			c.save(prop);
		}
		prop.store(w, "CrystalStructureOptions");
	}
	
	static abstract class CrystalProperty{
		protected String id, label;
		
		public CrystalProperty(String id, String label) {
			this.id = id;
			this.label = label;
		}
		public abstract JComponent getEditor();
		
		public String getLabel(){
			return label;
		}
		
		protected abstract void save(Properties prop);
		protected abstract void load(Properties prop);
	}
	
	static class IntegerCrystalProperty extends CrystalProperty{
		int value;
		int max, min;
		
		public IntegerCrystalProperty(String id, String label, int defaultValue, int min, int max) {
			super(id, label);
			this.min = min;
			this.max = max;
			this.value = defaultValue;
			assert (min>value  || max<value);
		}
		
		@Override
		protected void save(Properties prop) {
			prop.setProperty(id, Integer.toString(value));
		}
		
		@Override
		protected void load(Properties prop) {
			String s = prop.getProperty(id, Integer.toString(value));
			value = Integer.parseInt(s);
		}
		
		@Override
		public JComponent getEditor() {
			final JSpinner spinner = new JSpinner(new SpinnerNumberModel(value, min, max, 1));
			spinner.addChangeListener(new ChangeListener() {
				@Override
				public void stateChanged(ChangeEvent arg0) {
					value = ((Integer)(((JSpinner)arg0.getSource()).getValue())).intValue();
				}
			});
			return spinner;
		}
		
	}
	
	static class FloatCrystalProperty extends CrystalProperty{
		float value;
		float max, min;
		
		public FloatCrystalProperty(String id, String label,float defaultValue, float min, float max) {
			super(id, label);
			this.min = min;
			this.max = max;
			this.value = defaultValue;
			assert (min<=value && max>=value);
		}
		
		public float getValue() {
			return value;
		}
		
		@Override
		protected void save(Properties prop) {
			prop.setProperty(id, Float.toString(value));
		}
		
		@Override
		protected void load(Properties prop) {
			String s = prop.getProperty(id, Float.toString(value));
			value = Float.parseFloat(s);
		}
		
		@Override
		public JComponent getEditor() {
			final JSpinner spinner = new JSpinner(new SpinnerNumberModel(value, min, max, (max-min)/1000.));
			spinner.addChangeListener(new ChangeListener() {
				@Override
				public void stateChanged(ChangeEvent arg0) {
					value = ((Double)(((JSpinner)arg0.getSource()).getValue())).floatValue();
				}
			});
			return spinner;
		}
	}
	
	static class BooleanCrystalProperty extends CrystalProperty{
		boolean value;
		
		public BooleanCrystalProperty(String id, String label, boolean defaultValue) {
			super(id, label);
			this.value = defaultValue;
		}
			
		public boolean getValue(){
			return value;
		}
		
		@Override
		protected void save(Properties prop) {
			prop.setProperty(id, Boolean.toString(value));
		}
		
		@Override
		protected void load(Properties prop) {
			String s = prop.getProperty(id, Boolean.toString(value));
			value = Boolean.parseBoolean(s);
		}
		
		@Override
		public JComponent getEditor() {
			final JCheckBox checkBox = new JCheckBox(); 
			checkBox.setSelected(value);
			checkBox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					value = ((JCheckBox)arg0.getSource()).isSelected();
				}
			});
			
			return checkBox;
		}
	}
}
