// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package crystalStructures;

import java.util.*;

import common.VectorOps;
import crystalStructures.CrystalStructureProperties.BooleanCrystalProperty;
import crystalStructures.CrystalStructureProperties.FloatCrystalProperty;
import model.*;
import model.BurgersVector.BurgersVectorType;
import model.polygrain.Grain;
import model.polygrain.mesh.Mesh;
import model.skeletonizer.processors.MeshCleaningPreprocessor;
import model.skeletonizer.processors.MeshLineSenseCenteringPreprocessor;
import model.skeletonizer.processors.SkeletonPreprocessor;
import model.skeletonizer.processors.BurgersVectorAnalyzer.ClassificationPattern;

public class FCCStructure extends CrystalStructure {

	private final static int FCC = 1;
	private final static int HCP = 2;
	
	private static double[][] neighPerfFCC = new double[][]{
		{0., 0.5, 0.5},
		{0.,-0.5,-0.5},
		{0.,-0.5, 0.5},
		{0., 0.5,-0.5},
			
		{0.5 , 0., 0.5},
		{-0.5, 0.,-0.5},
		{-0.5, 0., 0.5},
		{0.5 , 0.,-0.5},
			
		{0.5 , 0.5 , 0.},
		{-0.5, -0.5, 0.},
		{-0.5, 0.5 , 0.},
		{0.5 , -0.5, 0.}
	};
	
	private static final ArrayList<ClassificationPattern> bvClassifcationPattern = new ArrayList<ClassificationPattern>();
	static{
		//Shockley partial <211> with two adjacent stacking fault planes
		bvClassifcationPattern.add(new ClassificationPattern(211, 4, 9, 211, 6, 2, 2, BurgersVectorType.PARTIAL));
		//Perfect <110> with no adjacent stacking fault planes
		bvClassifcationPattern.add(new ClassificationPattern(110, 1, 3, 110, 2, 0, 0, BurgersVectorType.PERFECT));
		//Ideal Shockley partials, where the adjacent stacking faults are not precisely identified 
		bvClassifcationPattern.add(new ClassificationPattern(211, 6, 6, 211, 6, 0, 10, BurgersVectorType.PARTIAL));
		//Slightly disordered partial <211> recognized as <632>
		bvClassifcationPattern.add(new ClassificationPattern(632, 10, 30, 211, 6, 2, 3, BurgersVectorType.PARTIAL));
		//Slightly disordered partial <211> recognized as 1/5<211> or 1/8<211> 
		bvClassifcationPattern.add(new ClassificationPattern(211, 5, 8, 211, 6, 2, 10, BurgersVectorType.PARTIAL));
		//More severed disordered partial <211> recognized as 1/5<211> or 1/11<211> but with correct number of adjacent stacking faults
		bvClassifcationPattern.add(new ClassificationPattern(211, 5, 11, 211, 6, 2, 2, BurgersVectorType.PARTIAL));
		//Stair Rod <100> with extremely clear signals
		bvClassifcationPattern.add(new ClassificationPattern(100, 3, 6, 100, 3, 4, 4, BurgersVectorType.STAIR_ROD));
		//Stair Rod 1/3<101> with perfect signals
		bvClassifcationPattern.add(new ClassificationPattern(110, 3, 3, 110, 3, 4, 10, BurgersVectorType.STAIR_ROD));
		//Stair Rod 1/6<101> with perfect signals
		bvClassifcationPattern.add(new ClassificationPattern(110, 6, 8, 110, 6, 4, 10, BurgersVectorType.STAIR_ROD));
		//Stair Rod 1/6<101> with not so close signals, but correct number of adjacent surfaces
		bvClassifcationPattern.add(new ClassificationPattern(110, 6, 11, 110, 6, 4, 4, BurgersVectorType.STAIR_ROD));
		//Stair Rod 1/6<310> with perfect signals
		bvClassifcationPattern.add(new ClassificationPattern(310, 6, 8, 310, 6, 4, 4, BurgersVectorType.STAIR_ROD));
		//Stair Rod 1/6<123> with perfect signals
		bvClassifcationPattern.add(new ClassificationPattern(123, 6, 8, 123, 6, 4, 4, BurgersVectorType.STAIR_ROD));
		//Frank partial 1/3<111> with perfect signals
		bvClassifcationPattern.add(new ClassificationPattern(111, 3, 3, 111, 3, 0, 4, BurgersVectorType.FRANK_PARTIAL));
	}
	
	
	private FloatCrystalProperty grainBoundaryFilterDistanceProperty = 
			new FloatCrystalProperty("gbFilterDist", "Grain boundary filter distance (0 -> disabled)", 6f, 0f, 25f);
	protected BooleanCrystalProperty highTempProperty = 
			new BooleanCrystalProperty("highTempADA", "optimize ADA for tempature > 150K", false);
	
	public FCCStructure() {
		super();
		crystalProperties.add(grainBoundaryFilterDistanceProperty);
		crystalProperties.add(highTempProperty);
	}
	
	@Override
	protected CrystalStructure deriveNewInstance() {
		return new FCCStructure();
	}
	
	@Override
	protected String getIDName() {
		return "FCC";
	}
	
	@Override
	public float getDefaultNearestNeighborSearchScaleFactor(){
		return 0.848f;
	}
	
	@Override
	public float getPerfectBurgersVectorLength(){
		return latticeConstant*0.7071067f;
	}
	
	@Override
	public boolean hasStackingFaults(){
		return true;
	}
	
	@Override
	public float[][] getStackingFaultNormals(CrystalRotationTools crt){
		float[][] stackingFaultNormals = new float[4][0];
		
		stackingFaultNormals[0] = VectorOps.normalize(crt.getInXYZ(new float[]{ 1f, 1f, 1f})); 
		stackingFaultNormals[1] = VectorOps.normalize(crt.getInXYZ(new float[]{-1f, 1f,-1f}));
		stackingFaultNormals[2] = VectorOps.normalize(crt.getInXYZ(new float[]{-1f,-1f, 1f}));
		stackingFaultNormals[3] = VectorOps.normalize(crt.getInXYZ(new float[]{ 1f,-1f,-1f}));
		
		return stackingFaultNormals;
	}
	
	@Override
	public int identifyAtomType(Atom atom, NearestNeighborBuilder<Atom> nnb) {
		float threshold = highTempProperty.getValue() ? -0.945f : -0.965f;
		
		ArrayList<float[]> neigh = nnb.getNeighVec(atom);
		/*
		 * type=0: bcc
		 * type=1: fcc
		 * type=2: hcp
		 * type=3: unassigned
		 * type=4: less than 12 neighbors
		 * type=5: more than 12 neighbors
		 * type=6: less than 10 neighbors
		 */
		if (neigh.size() < 10) return 6;
		else if (neigh.size() < 12) return 4;
		else if (neigh.size() > 14) return 7;
		else {
			int co_x0 = 0;
			int co_x1 = 0;
			int co_x2 = 0;
			for (int i = 0; i < neigh.size(); i++) {				
				float v_x = neigh.get(i)[0];
				float v_y = neigh.get(i)[1];
				float v_z = neigh.get(i)[2];
				
				float v_length = (float)Math.sqrt(v_x*v_x+v_y*v_y+v_z*v_z);
				
				for (int j = 0; j < i; j++) {
					float u_x = neigh.get(j)[0];
					float u_y = neigh.get(j)[1];
					float u_z = neigh.get(j)[2];
					
					float u_length = (float)Math.sqrt(u_x*u_x+u_y*u_y+u_z*u_z);
					float a = v_x*u_x+v_y*u_y+v_z*u_z;
					a /= (v_length*u_length);
					
					if (a < threshold) co_x0++; // 0.945
					else if (a < -.915f) co_x1++;
					else if (a < -.775f) co_x2++; // 0.755
				}
			}

			if (co_x0 == 7 && neigh.size() == 14) return 0;
			else if (co_x0 == 6 && neigh.size() == 12) return FCC;
			else if (co_x0 == 3 && co_x1 <= 1 && co_x2 > 2 && neigh.size() == 12) return HCP;
			else if (neigh.size() > 12) return 5;
			else if (neigh.size() == 12) return 3;
			else return 4;
		}
	}
	
	@Override
	public ArrayList<Atom> getDislocationDefectAtoms(AtomData data) {
		ArrayList<Atom> defectAtoms = new ArrayList<Atom>();
		
		float minRBVLength = latticeConstant*0.7071067f*0.1f; //latticeConstant*0.7071067f -> perfect burgers vector 1/2[110]
		minRBVLength *= minRBVLength;
		float maxRBVLength = latticeConstant*0.7071067f*2.5f;
		maxRBVLength *= maxRBVLength;
		
		for (Atom a : data.getAtoms()) {
			if (a.getType() >= 3 && a.getType() <= 5 && (!ImportStates.POLY_MATERIAL.isActive() || a.getGrain() != Atom.IGNORED_GRAIN)){
				if (data.isRbvAvailable()){
					if (a.getRBV()!=null && a.getRBV().getSquaredRBVLengh()>minRBVLength && a.getRBV().getSquaredRBVLengh()<maxRBVLength)
						defectAtoms.add(a);
				}
				else defectAtoms.add(a);
			}
		}
		
		return defectAtoms;
	}
	
	@Override
	public TreeSet<Atom> getStackingFaultAtoms(AtomData data){
		TreeSet<Atom> sfAtoms = new TreeSet<Atom>();
		for (int i=0; i<data.getAtoms().size(); i++){
			Atom a = data.getAtoms().get(i);
			if (a.getType() == HCP && (!ImportStates.POLY_MATERIAL.isActive() || a.getGrain() != Atom.IGNORED_GRAIN))
				sfAtoms.add(a);
		}
		return sfAtoms;
	}

	@Override
	public boolean isRBVToBeCalculated(Atom a) {
		if (ImportStates.POLY_MATERIAL.isActive() && a.getGrain() == Atom.IGNORED_GRAIN) return false;
		
		int type = a.getType();
		if (type != FCC && type != HCP && type != 6) return true;
		return false;
	}
	
	@Override
	public int getNumberOfTypes() {
		return 8;
	}

	@Override
	public String getNameForType(int i) {
			switch (i) {
			case 0: return "bcc";
			case 1: return "fcc";
			case 2: return "hcp";
			case 3: return "12 neigh., unassigned";
			case 4: return "10-11 neighbors";
			case 5: return "13-14 neighbors";
			case 6: return "<10 neighbors";
			case 7: return ">14 neighbors";
			default: return "unknown";
		}
	}
	
	
	@Override
	public double[][] getPerfectNearestNeighborsUnrotated() {
		return neighPerfFCC.clone();
	}
	
	@Override
	public float getRBVIntegrationRadius() {
		return 0.707107f*latticeConstant;
	}
	
	@Override
	public int getNumberOfNearestNeighbors() {
		return 12;
	}
	
	@Override
	public ArrayList<ClassificationPattern> getBurgersVectorClassificationPattern() {
		return bvClassifcationPattern;
	}
	
	public List<SkeletonPreprocessor> getSkeletonizerPreProcessors(){
		Vector<SkeletonPreprocessor>  list = new Vector<SkeletonPreprocessor>();
		list.add(new MeshCleaningPreprocessor());
		list.add(new MeshLineSenseCenteringPreprocessor());
		
		return list;
	}
	
	@Override
	public int getDefaultType(){
		return 1;
	}
	
	@Override
	public List<Grain> identifyGrains(AtomData data) {
		List<Grain> grains = new Vector<Grain>();
		
		if (!data.isGrainsImported()){
			/**
			 * HIGHLY EXPERIMENTAL STUFF
			 * do not expect to work properly
			 * In fact the opposite is true 
			 */
			data.findNearestNeighbors(getNearestNeighborSearchRadius(), new NearestNeighborFilter() {
				@Override
				public boolean accept(Atom a) {
					if (a.getType()!=6) return true;
					return false;
				}
			});
			//Detect grains from data
			TreeSet<Atom> notAssignedAtoms = new TreeSet<Atom>();
			for (Atom a : data.getAtoms()) {
				if ( a.getNeigh()!=null ) notAssignedAtoms.add(a);
			}
			while (!notAssignedAtoms.isEmpty()) {
				Atom a = notAssignedAtoms.pollFirst();
				if (a.getNeigh().length <= 9) continue;
				
				ArrayList<Atom> grain = new ArrayList<Atom>();
				//From each item in the grain array the bfs is continued
				//There is no need to keep a second list, the index of processed
				//atoms is enough
				int queuePosition = 0;
				int grainID = grains.size();
				grain.add(a);
				a.setGrain(grainID);
				
				while (queuePosition < grain.size()) {
					Atom c = grain.get(queuePosition++);
					int count = 0;
					for (int j=0; j<c.getNeigh().length; j++){
						if (c.getNeigh()[j].getType() == FCC || c.getNeigh()[j].getType() == HCP) count++;
					}
					// in case some same element neighbors exists, continue a breath
					// first search at this atoms
					if (count > 9) {
						for (int i = 0; i < c.getNeigh().length; i++) {
							if (c.getNeigh()[i].getGrain() == Atom.DEFAULT_GRAIN) {
								c.getNeigh()[i].setGrain(grainID);
								grain.add(c.getNeigh()[i]);
								notAssignedAtoms.remove(c.getNeigh()[i]);
							}
						}
					}
				}
	
				if (grain.size() > 20) {
					Mesh mesh = new Mesh(grain, this);
					Grain g = new Grain(mesh, grain, grains.size(), this, data.getBounds());
					grains.add(g);
				} else {
					for (int i=0; i<grain.size(); i++) {
						grain.get(i).setGrain(Atom.DEFAULT_GRAIN);
					}
				}
			}
		} else {
			grains = super.identifyGrains(data);
		}
		return grains;
	}
	
	
	@Override
	public int getSurfaceType() {
		return 6;
	}
	
	@Override
	public float getGrainBoundaryFilterDistance() {
		return grainBoundaryFilterDistanceProperty.getValue();
	}
	
	public float getSkeletonizationMeshingThreshold(){
		return getNearestNeighborSearchRadius() * 1.05f;
	}
}
