// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package crystalStructures;

import java.util.*;
import java.util.concurrent.Callable;
import java.util.concurrent.CyclicBarrier;

import common.*;

import model.*;
import model.polygrain.Grain;
import model.polygrain.mesh.Mesh;

public class FCCTwinnedStructure extends FCCStructure {

	private static final float[][] ATOM_COLORS = {
		{1f, 0.5f, 0f},
		{0f, 0f, 1f}, 
		{1f, 0f, 0f}, 
		{1f, 0.4f, 0.7f}, 
		{1f, 1f, 0f}, 
		{0f, 1f, 0f}, 
		{0f, 1f, 1f}, 
		{0f, 1f, 0.5f},
		{1f, 1f, 1f}
	};
	
	@Override
	public float[][] getDefaultColors(){
		float[][] f = new float[ATOM_COLORS.length][3];
		for (int i=0; i< f.length;i++){
			f[i][0] = ATOM_COLORS[i][0];
			f[i][1] = ATOM_COLORS[i][1];
			f[i][2] = ATOM_COLORS[i][2];
		}
		
		return f;
	}
	
	@Override
	protected CrystalStructure deriveNewInstance() {
		return new FCCTwinnedStructure();
	}
	
	@Override
	protected String getIDName() {
		return "FCC_Twinned";
	}
	
	@Override
	public String getNameForType(int i) {
			switch (i) {
			case 0: return "bcc";
			case 1: return "fcc";
			case 2: return "hcp";
			case 3: return "twin";
			case 4: return "10-11 neighbors";
			case 5: return "defect";
			case 6: return "<10 neighbors";
			default: return "unknown";
		}
	}
	
	@Override
	public int getNumberOfTypes() {
		return 7;
	}
	
	@Override
	public void identifyDefectAtoms(List<Atom> atoms, NearestNeighborBuilder<Atom> nnb, 
			int start, int end, CyclicBarrier barrier) {
		for (int i=start; i<end; i++){
			if (Thread.interrupted()) return;
			Atom a = atoms.get(i);
			int type = identifyAtomType(a, nnb); 
			//Relabel
			if (type == 3 || type == 7) type = 5;
			a.setType(type);
		}
		try {
			barrier.await();
		} catch (Exception e) {
			if (Thread.interrupted()) return;
		}
		
		Atom n1 = null;
		float[] normal = null;
		
		for (int i=start; i<end; i++){
			if (Thread.interrupted()) return;
			Atom a = atoms.get(i);
			
			if (a.getType() == 2){
				int countHCP = 0;
				boolean twin = true;
				ArrayList<Atom> nei = nnb.getNeigh(a);
				for (Atom n : nei){
					if (n.getType() == 2 || n.getType() == 3){
						if (countHCP == 0) {
							countHCP++;
							n1 = n;
						} else if (countHCP == 1) {
							countHCP++;
							normal = VectorOps.makeNormal(a, n1, n);
							n1 = n;
						} else {
							float[] d = VectorOps.normalize(VectorOps.sub(a, n));
							float dev = VectorOps.product(normal, d);
							if (Math.abs(dev)>0.2) {
								twin = false;
								break;
							}
							countHCP++;
						}
						
					}
				}
				
				if (twin && countHCP >= 3) a.setType(3);
			}
		}
	}
	
	@Override
	public List<Grain> identifyGrains(final AtomData data) {
		List<Grain> grains = new Vector<Grain>();
		
		if (!data.isGrainsImported()){
			/**
			 * HIGHLY EXPERIMENTAL STUFF
			 * do not expect to work properly
			 * In fact the opposite is true 
			 */
			data.findNearestNeighbors(getNearestNeighborSearchRadius(), new NearestNeighborFilter() {
				@Override
				public boolean accept(Atom a) {
					if (a.getType()!=6) return true;
					return false;
				}
			});
			//Detect grains from data
			TreeSet<Atom> notAssignedAtoms = new TreeSet<Atom>();
			for (Atom a : data.getAtoms()) {
				if ( a.getNeigh()!=null ) notAssignedAtoms.add(a);
			}
			
			final HashMap<Integer, List<Atom>> allGrains = new HashMap<Integer, List<Atom>>();
			
			int grainID = 0;
			
			while (!notAssignedAtoms.isEmpty()) {
				Atom a = notAssignedAtoms.pollFirst();
				if (a.getNeigh().length <= 9) continue;
				
				Vector<Atom> grain = new Vector<Atom>();
				//From each item in the grain array the bfs is continued
				//There is no need to keep a second list, the index of processed
				//atoms is enough
				int queuePosition = 0;
				
				grain.add(a);
				a.setGrain(grainID);
				
				while (queuePosition < grain.size()) {
					Atom c = grain.get(queuePosition++);
					int count = 0;
					for (int j=0; j<c.getNeigh().length; j++){
						if (c.getNeigh()[j].getType() == 1) count++;
					}
					// in case some same element neighbors exists, continue a breath
					// first search at this atoms, consider only atoms with mostly FCC neighbors
					if (count > 9) {
						for (int i = 0; i < c.getNeigh().length; i++) {
							if (c.getNeigh()[i].getGrain() == Atom.DEFAULT_GRAIN) {
								c.getNeigh()[i].setGrain(grainID);
								grain.add(c.getNeigh()[i]);
								notAssignedAtoms.remove(c.getNeigh()[i]);
							}
						}
					}
				}
	
				if (grain.size() > 20) {
					allGrains.put(grainID, grain);
					grainID++;
				} else {
					for (int i=0; i<grain.size(); i++) {
						grain.get(i).setGrain(Atom.DEFAULT_GRAIN);
					}
				}
			}
			
			//Second step, find the most frequent grain assignment of neighbors, 
			//then assign this type to the previously unassigned atom
			Vector<Callable<Void>> parallelTasks = new Vector<Callable<Void>>();
			final CyclicBarrier cb = new CyclicBarrier(ThreadPool.availProcessors());
			for (int i=0; i<ThreadPool.availProcessors(); i++){
				final int start = (data.getAtoms().size() * i)/ThreadPool.availProcessors();
				final int end = (data.getAtoms().size() * (i+1))/ThreadPool.availProcessors();
				
				parallelTasks.add(new Callable<Void>() {
					@Override
					public Void call() throws Exception {
						ArrayList<Tupel<Atom,Integer>> assigns = new ArrayList<Tupel<Atom, Integer>>();
						for (int j=start; j<end; j++){
							Atom a = data.getAtoms().get(j);
							if (a.getNeigh() != null && a.getNeigh().length != 0 && a.getGrain() == Atom.DEFAULT_GRAIN){
								int[] ng = new int[a.getNeigh().length];
								for (int i=0; i<a.getNeigh().length;i++){
									ng[i] = a.getNeigh()[i].getGrain();
								}
								Arrays.sort(ng);
								int maxElement = 1; int maxIndex = ng[0];
								int currentElement = 1; int currentIndex = ng[0];
								
								for (int i=1; i<ng.length; i++){
									if (ng[i] == currentIndex){
										currentElement++;
									} else {
										if (currentElement > maxElement && currentIndex != Atom.DEFAULT_GRAIN) {
											maxElement = currentElement;
											maxIndex = currentIndex;
										}
										currentElement = 1;
										currentIndex = ng[i];
									}
								}
								
								if (maxIndex != Atom.DEFAULT_GRAIN){
									assigns.add(new Tupel<Atom, Integer>(a, maxIndex));
								}
							}
						}
						
						cb.await();
						for (Tupel<Atom, Integer> a : assigns){
							a.getO1().setGrain(a.getO2());
							allGrains.get(a.getO2()).add(a.getO1());
						}
						
						return null;
					}
				});
			};
			ThreadPool.executeParallel(parallelTasks);
			
			//Now build the grains
			for (List<Atom> grain : allGrains.values()){
				Mesh mesh = new Mesh(grain, this);
				Grain g = new Grain(mesh, grain, grains.size(), this, data.getBounds());
				grains.add(g);
			}
		} else {
			grains = super.identifyGrains(data);
		}
		return grains;
	}
	
	
	@Override
	public ArrayList<Atom> getDislocationDefectAtoms(AtomData data) {
		ArrayList<Atom> defectAtoms = new ArrayList<Atom>();
		
		float minRBVLength = latticeConstant*0.7071067f*0.1f; //latticeConstant*0.7071067f -> perfect burgers vector 1/2[110]
		minRBVLength *= minRBVLength;
		float maxRBVLength = latticeConstant*0.7071067f*2.5f;
		maxRBVLength *= maxRBVLength;
		
		for (Atom a : data.getAtoms()) {
			if ( (a.getType() == 4 || a.getType() == 5) && (!ImportStates.POLY_MATERIAL.isActive() || a.getGrain() != Atom.IGNORED_GRAIN)){
				if (data.isRbvAvailable()){
					if (a.getRBV()!=null && a.getRBV().getSquaredRBVLengh()>minRBVLength && a.getRBV().getSquaredRBVLengh()<maxRBVLength)
						defectAtoms.add(a);
				}
				else defectAtoms.add(a);
			}
		}
		
		return defectAtoms;
	}
	
	@Override
	public boolean isRBVToBeCalculated(Atom a) {
		if (ImportStates.POLY_MATERIAL.isActive() && a.getGrain() == Atom.IGNORED_GRAIN) return false;
		
		int type = a.getType();
		if (type == 4 || type == 5) return true;
		return false;
	}
	
	@Override
	public TreeSet<Atom> getStackingFaultAtoms(AtomData data){
		TreeSet<Atom> sfAtoms = new TreeSet<Atom>();
		for (int i=0; i<data.getAtoms().size(); i++){
			Atom a = data.getAtoms().get(i);
			if ((a.getType() == 2 || a.getType() == 3) && (!ImportStates.POLY_MATERIAL.isActive() || a.getGrain() != Atom.IGNORED_GRAIN))
				sfAtoms.add(a);
		}
		return sfAtoms;
	}
	
	
}
