// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package crystalStructures;

import java.util.ArrayList;

import common.Tupel;

import model.Atom;
import model.NearestNeighborBuilder;

public class FeCStructure extends BCCStructure {
	
	@Override
	protected CrystalStructure deriveNewInstance() {
		return new FeCStructure();
	}
	
	@Override
	protected String getIDName() {
		return "FeC";
	}
	
	@Override
	public String getNameForType(int i) {
		switch (i) {
			case 0: return "bcc";
			case 1: return "unused";
			case 2: return "unused";
			case 3: return "14 neighbors";
			case 4: return "11-13 neighbors";
			case 5: return ">14 neighbors";
			case 6: return "<11 neighbors";
			case 7: return "Carbon";
			default: return "unknown";
		}
	}
	
	@Override
	public void buildNeighborsForRBV(Atom a, NearestNeighborBuilder<Atom> nnb){
		a.buildNeighOfSameElement(nnb);
	}
	
	@Override
	public int getNumberOfElements(){
		return 2;
	}
	
	@Override
	public float[] getSphereSizeScalings(){
		float[] size = new float[getNumberOfElements()];
		size[0] = 1f;
		size[1] = 0.43f;
		return size;
	}
	
	@Override
	public int identifyAtomType(Atom atom, NearestNeighborBuilder<Atom> nnb) {
		ArrayList<Tupel<Atom, float[]>> neigh = nnb.getNeighAndNeighVec(atom);		
		/*
		 * type=0: bcc
		 * type=1: unused
		 * type=2: unused
		 * type=3: 14 neighbors
		 * type=4: 11-13 neighbors
		 * type=5: >14 neighbors
		 * type=6: less than 11 neighbors
		 * type=7: carbon
		 */
		
		//carbon
		if (atom.getElement()%2 == 1) return 7;
		
		//count Fe neighbors for Fe atoms
		int count = 0;
		for (int i=0; i < neigh.size(); i++)
			if (neigh.get(i).getO1().getElement() % 2 == 0) count++;		
		
		if (count < 11) return 6;
		else if (count == 11) return 4;
		else if (count >= 15) return 5;
		else {
			int co_x0 = 0;
			int co_x1 = 0;
			int co_x2 = 0;
			for (int i = 0; i < neigh.size(); i++) {
				if (neigh.get(i).getO1().getElement() % 2 == 1) continue; //ignore carbon atoms
				float v_x = neigh.get(i).getO2()[0];
				float v_y = neigh.get(i).getO2()[1];
				float v_z = neigh.get(i).getO2()[2];
				
				float v_length = (float)Math.sqrt(v_x*v_x+v_y*v_y+v_z*v_z);
				
				for (int j = 0; j < i; j++) {
					if (neigh.get(j).getO1().getElement() % 2 == 1) continue; //ignore carbon atoms
					float u_x = neigh.get(j).getO2()[0];
					float u_y = neigh.get(j).getO2()[1];
					float u_z = neigh.get(j).getO2()[2];
					
					float u_length = (float)Math.sqrt(u_x*u_x+u_y*u_y+u_z*u_z);
					float a = v_x*u_x+v_y*u_y+v_z*u_z;
					a /= (v_length*u_length);
					
					if (a < -.945)
						co_x0++;
					else if (a < -.915)
						co_x1++;
					else if (a > -.75 && a< -0.67)
						co_x2++;
				}
			}
			
			if (co_x0 > 5 && co_x0+co_x1==7 && co_x2<=2 && count==14) return 0;
			else if (count==13 && neigh.size()==14 && co_x0==6) return 0;
			else if (count == 13) return 4;
			else if (count == 12) return 4;
			else return 3;
		}
	}
	
	@Override
	public boolean isRBVToBeCalculated(Atom a) {
		int type = a.getType();		
		if (type != 0 && type < 6) {			
			return true;
		}
		return false;		
	}
}
