// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package crystalStructures;

import java.util.*;
import java.util.concurrent.Callable;

import Jama.Matrix;
import model.Atom;
import model.AtomData;
import model.Configuration;
import model.CustomColumnInfo;
import model.NearestNeighborBuilder;
import common.ThreadPool;
import common.VectorOps;

public final class MonoclinicNiTi extends B2NiTi{
	public static final int IGNORED_VARIANT = -1;
	public static final int UNKNOWN_VARIANT = 13;
	
	//Both in <100>/<010>/<001> orientation
//	private static double[][] neighPerfTiTi = new double[][]{
//			{1.7999935 , 2.3601036, 0.0},
//			{-2.5458164, 2.3601036, 0.0},
//			{2.172903  ,-1.974308 , 0.0},
//			{-2.172905 ,-1.974308 , 0.0},
//			{0.0       , 0.0      , 2.8998308},
//			{0.0       , 0.0      ,-2.8998299}
//		};
	
	private static double[][] neighPerfNiNi = new double[][]{
			{1.7925072 , 1.9250374 , 0.0},
			{-2.5533028, 1.9250374 , 0.0},
			{2.172905  ,-2.4093723 , 0.0},
			{-2.172903 ,-2.4093723 , 0.0},
			{0.0       , 0.0       , 2.8998299},
			{0.0       , 0.0       ,-2.8998299}
		};
	
	private static Matrix neighPerfNiNiMatrix; 
	
	static {
		//Rotate this coordinate system to the orientation of the comparing patters
		float[][] defRot = new float[][]{{0.70710677f, -0.70710677f, 0.0f}, {0.70710677f, 0.70710677f, 0.0f}, {0.0f, 0.0f, 1.0f}};
		for (int i=0; i<neighPerfNiNi.length; i++){
			double[] d = new double[3];
			d[0] = (neighPerfNiNi[i][0]*defRot[0][0] + neighPerfNiNi[i][1]*defRot[1][0] + neighPerfNiNi[i][2]*defRot[2][0]);
			d[1] = (neighPerfNiNi[i][0]*defRot[0][1] + neighPerfNiNi[i][1]*defRot[1][1] + neighPerfNiNi[i][2]*defRot[2][1]);
			d[2] = (neighPerfNiNi[i][0]*defRot[0][2] + neighPerfNiNi[i][1]*defRot[1][2] + neighPerfNiNi[i][2]*defRot[2][2]);
			neighPerfNiNi[i] = d;
		}
		neighPerfNiNi = sortBondsNi(neighPerfNiNi);
		//neighPerfTiTi = sortBondsNi(neighPerfTiTi);
		neighPerfNiNiMatrix = new Matrix(neighPerfNiNi);
	}
	
	@Override
	protected String getIDName() {
		return "MonoclinicNiTi";
	}
	
	protected MonoclinicNiTi(float latticeConstant, float nearestNeighborSearchRadius) {
		super();
		this.latticeConstant = latticeConstant;
		this.nearestNeighborSearchRadius = nearestNeighborSearchRadius;
	}
	
	@Override
	public float getMeshingCellSize(){
		return 2f*nearestNeighborSearchRadius;
	}
	
	public static void identifyVariants(final AtomData data){
		final NearestNeighborBuilder<Atom> nnb = new NearestNeighborBuilder<Atom>(data.getBounds(), 
				Configuration.getCrystalStructure().getNearestNeighborSearchRadius(),
				Configuration.getPbc());
		
		final float[][] rot = Configuration.getCrystalRotationTools().getDefaultRotationMatrix();
		
		//Find the correct column containing the martensite variant values
		int mc = 0;
		for (int i=0; i<Configuration.getSizeCustomColumns(); i++){
			CustomColumnInfo cci = Configuration.getCustomColumnInfo(i);
			if (cci.getName() == "Variant")
				mc = i;
		}
		final int martColumn = mc;
		
		for (Atom a : data.getAtoms()){
			if (a.getElement() % 2 == 1) nnb.add(a);
			else a.setCustom(IGNORED_VARIANT, martColumn);
		}
		
		Vector<Callable<Void>> parallelTasks = new Vector<Callable<Void>>();
		for (int i=0; i<ThreadPool.availProcessors(); i++){
			final int j = i;
			parallelTasks.add(new Callable<Void>() {
				@Override
				public Void call() throws Exception {
					final int start = (data.getAtoms().size() * j)/ThreadPool.availProcessors();
					final int end = (data.getAtoms().size() * (j+1))/ThreadPool.availProcessors();
					for (int i=start; i<end; i++){
						Atom a = data.getAtoms().get(i);
						if (a.getElement() % 2 == 1 && (a.getType() == 3 || a.getType() == 4)) 
							a.setCustom(getVariant(a, nnb, rot),martColumn);
						else a.setCustom(IGNORED_VARIANT, martColumn);
					}
					return null;
				}
			});
		}
		ThreadPool.executeParallel(parallelTasks);		
	}
	
	private static double[][] sortBondsNi(double[][] unsortedBonds){
		double[][] sort = new double[3][];

		//Put all bonds into a list 
		Vector<double[]> ub = new Vector<double[]>();
		for (int i=0; i<unsortedBonds.length; i++) ub.add(unsortedBonds[i]);
		//Step 1: pick point closest to center -> sort[0]
		double nearestDist = Double.MAX_VALUE;
		int num = 0;
		for (int i=0; i<ub.size();i++){
			double dist = VectorOps.lengthSqr(ub.get(i));
			if (nearestDist>dist) {
				nearestDist = dist;
				num = i;
			}
		}
		sort[0] = ub.remove(num);
		
		//Step 2: select closest point to sort[0]
		nearestDist = Double.MAX_VALUE;
		num = 0;
		for (int i = 0; i < ub.size(); i++) {
			double dist =  VectorOps.sqrDist(ub.get(i), sort[0]);
			if (nearestDist>dist) {
				nearestDist = dist;
				num = i;
			}
		}
		sort[1] = ub.remove(num);
		
		double[] p = VectorOps.crossProduct(sort[0], sort[1]);
		double largestDist = 0.;
		num = 0;
		for (int i=0; i<ub.size();i++){
			double d = VectorOps.product(ub.get(i), p);
			if (largestDist<d) {
				largestDist = d;
				num = i;
			}
		}
		sort[2] = ub.remove(num);
		return sort;
	}
	
	private static int getVariant(Atom a, NearestNeighborBuilder<Atom> nnb, float[][] rot){
		if (a.getElement() % 2 == 0) return UNKNOWN_VARIANT;
		ArrayList<float[]> nn = nnb.getNeighVec(a);
	
		if (nn.size() < 4 ) return -1;
		
		double[][] neighVec = new double[nn.size()][3];
		
		for (int i=0; i<nn.size(); i++){
			neighVec[i][0] = nn.get(i)[0]*rot[0][0] + nn.get(i)[1]*rot[0][1] + nn.get(i)[2]*rot[0][2];
			neighVec[i][1] = nn.get(i)[0]*rot[1][0] + nn.get(i)[1]*rot[1][1] + nn.get(i)[2]*rot[1][2];
			neighVec[i][2] = nn.get(i)[0]*rot[2][0] + nn.get(i)[1]*rot[2][1] + nn.get(i)[2]*rot[2][2];
		}
		
		neighVec = sortBondsNi(neighVec);
		
		Matrix trans;
		try {
			trans = new Matrix(neighVec).solve(neighPerfNiNiMatrix);
		} catch (RuntimeException e){
			return UNKNOWN_VARIANT;
		}
		int[] m = new int[9];
		m[0] = Math.abs(trans.get(0, 0))<0.75 ? 0 : (trans.get(0, 0)>0. ? 1:-1);
		m[1] = Math.abs(trans.get(0, 1))<0.75 ? 0 : (trans.get(0, 1)>0. ? 1:-1);
		m[2] = Math.abs(trans.get(0, 2))<0.75 ? 0 : (trans.get(0, 2)>0. ? 1:-1);
		                                                 
		m[3] = Math.abs(trans.get(1, 0))<0.75 ? 0 : (trans.get(1, 0)>0. ? 1:-1);
		m[4] = Math.abs(trans.get(1, 1))<0.75 ? 0 : (trans.get(1, 1)>0. ? 1:-1);
		m[5] = Math.abs(trans.get(1, 2))<0.75 ? 0 : (trans.get(1, 2)>0. ? 1:-1);
		                                                 
		m[6] = Math.abs(trans.get(2, 0))<0.75 ? 0 : (trans.get(2, 0)>0. ? 1:-1);
		m[7] = Math.abs(trans.get(2, 1))<0.75 ? 0 : (trans.get(2, 1)>0. ? 1:-1);
		m[8] = Math.abs(trans.get(2, 2))<0.75 ? 0 : (trans.get(2, 2)>0. ? 1:-1);
		
		int ones = 0;
		for (int i=0; i<9; i++){
			if (m[i] != 0) ones++;
		}
		
		if (ones!=3)
			return -1;
		
		if (m[6] == -1 || m[7] == -1 || m[8] == -1){
			m[3] *=-1; m[4] *=-1; m[5] *=-1;
		}
		
		if (m[0] == 1 && m[4] == 1) return 0;
		if (m[0] == 1 && m[4] ==-1) return 1;
		if (m[0] ==-1 && m[4] ==-1) return 0;
		if (m[0] ==-1 && m[4] == 1) return 1;
		
		if (m[1] == 1 && m[3] == 1) return 3;
		if (m[1] == 1 && m[3] ==-1) return 2;
		if (m[1] ==-1 && m[3] ==-1) return 3;
		if (m[1] ==-1 && m[3] == 1) return 2;
		    
		if (m[1] == 1 && m[5] == 1) return 4;
		if (m[1] == 1 && m[5] ==-1) return 5;
		if (m[1] ==-1 && m[5] ==-1) return 4;
		if (m[1] ==-1 && m[5] == 1) return 5;
		    
		if (m[0] == 1 && m[5] == 1) return 7;
		if (m[0] == 1 && m[5] ==-1) return 6;
		if (m[0] ==-1 && m[5] ==-1) return 7;
		if (m[0] ==-1 && m[5] == 1) return 6;
		    
		if (m[2] == 1 && m[3] == 1) return 8;
		if (m[2] == 1 && m[3] ==-1) return 9;
		if (m[2] ==-1 && m[3] ==-1) return 8;
		if (m[2] ==-1 && m[3] == 1) return 9;
		    
		if (m[2] == 1 && m[4] == 1) return 11;
		if (m[2] == 1 && m[4] ==-1) return 10;
		if (m[2] ==-1 && m[4] ==-1) return 11;
		if (m[2] ==-1 && m[4] == 1) return 10;
		
		return UNKNOWN_VARIANT;
	}
}
