// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package gui;

import java.awt.*;

import common.VectorOps;
/**
* The Arcball rotation interface sources were created by 
* Pepijn Van Eeckhoudt.
* 
* Original ArcBall C code from Ken Shoemake, Graphics Gems IV, 1993. 
*
* Modified by C.Begau
**/
class ArcBall {
	private static final float EPSILON = 1.0e-5f;
	private static final float SPHERESIZE = 0.8f;

	private float[] startPoint;
	private boolean startVecInsideSphere = false;
	private float widthScaling, heightScaling;

	public ArcBall(int width, int height) {
		startPoint = new float[3];
		setSize(width, height);
	}

	public void setSize(int width, int height) {		
		this.widthScaling = 1.0f / ((width - 1.0f) * 0.5f);
		this.heightScaling = 1.0f / ((height - 1.0f) * 0.5f);
	}

	public void setRotationStartingPoint(Point point) {
		float[] scaledPointVec = new float[3];

		//scaling coordinates
		scaledPointVec[0] = (point.x * this.widthScaling) - 1.0f;
		scaledPointVec[1] = 1.0f - (point.y * this.heightScaling);

		float length = VectorOps.length(scaledPointVec);

		//Startpoint inside or outside the ball (sphere)?
		if (length > SPHERESIZE) {
			float scaleFactor = (float) (SPHERESIZE / Math.sqrt(length));
			startPoint[0] = scaledPointVec[0] * scaleFactor;
			startPoint[1] = scaledPointVec[1] * scaleFactor;
			startPoint[2] = 0.0f;
			startVecInsideSphere = false;
		} else {
			startPoint[0] = scaledPointVec[0];
			startPoint[1] = scaledPointVec[1];
			startPoint[2] = (float)Math.sqrt(SPHERESIZE - length);
			startVecInsideSphere = true;
		}
	}

	/**
	 * Computes the rotation on the acrBall from the startpoint of rotation to a current point
	 * @param point Current point of the cursor on screen
	 * @return the quaternion as a float[4] or null if not on screen of rotation matrix is the identity
	 */
	public float[] drag(Point point) {
		float[] scaledPointVec = new float[3];

		//scaling coordinates
		scaledPointVec[0] = (point.x * this.widthScaling) - 1.0f;
		scaledPointVec[1] = 1.0f - (point.y * this.heightScaling);

		float length = VectorOps.length(scaledPointVec);
		float[] endVec = new float[3];
		
		if (length > SPHERESIZE) {
			float scaleFactor = (float) (SPHERESIZE / Math.sqrt(length));
			endVec[0] = scaledPointVec[0] * scaleFactor;
			endVec[1] = scaledPointVec[1] * scaleFactor;
			endVec[2] = 0.0f;
		} else {
			endVec[0] = scaledPointVec[0];
			endVec[1] = scaledPointVec[1];
			endVec[2] = (float)Math.sqrt(SPHERESIZE - length);
		}
		if (!startVecInsideSphere) endVec[2] = 0f;

		// Return the quaternion equivalent to the rotation
		float[] quarternionRot = new float[4];
		float[] normVec = VectorOps.crossProduct(startPoint, endVec);

		if (VectorOps.length(normVec) > EPSILON) {
			quarternionRot[0] = normVec[0];
			quarternionRot[1] = normVec[1];
			quarternionRot[2] = normVec[2];
			quarternionRot[3] = VectorOps.product(startPoint, endVec);
		} else return null;
		return quarternionRot;
	}
}