// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package gui;

import java.util.ArrayList;
import java.util.List;
import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;
import java.util.concurrent.Callable;

import common.GLMatrix;
import common.ThreadPool;
import model.Atom;
import model.Configuration;
import model.Coordinate;

class AtomRenderData {
	public final static int MAX_ATOMS_PER_CELL = 3000;
	private int runningTasks = 0;	//Counting the number of cells waiting for or running subdivision in the ThreadPool
	
	Cell rootCell;
	
	float maxSphereSize;
	private final CellComparator cellComparator = new CellComparator();
	Vector<Cell> allCells = new Vector<AtomRenderData.Cell>();
	
	public AtomRenderData(List<Atom> atoms, List<float[]> colors, float maxSphereSize, boolean subdivide) {
//		long time = System.nanoTime();
		
		this.maxSphereSize = maxSphereSize;		
		float[] size = Configuration.getCurrentAtomData().getBounds();

		rootCell = new Cell(size[0]*0.5f, size[1]*0.5f, size[2]*0.5f, size[0], size[1], size[2]);
		
		rootCell.atoms = new ArrayList<AtomRenderData.RenderableAtom>(atoms.size());
		for (int i = 0; i<atoms.size(); i++)
			rootCell.atoms.add(new RenderableAtom(atoms.get(i), colors.get(i)));		
		
		if(subdivide){
			incRunningTasks();
			ThreadPool.submit(rootCell);
		}
		else allCells.add(rootCell);
		
		getRenderableCells();
//		long time2 = System.nanoTime();
//		System.out.println("Create "+(time2-time)/1000000);
	}
	
	private synchronized void incRunningTasks(){
		runningTasks++;
	}
	
	private synchronized void decRunningTasks(){
		runningTasks--;
	}
	
	public void sortCells(final GLMatrix modelViewMatrix){
		//Extract the column to compute z-coordinate 
		float mvmMatrix1 = modelViewMatrix.getMatrix().get(2);
		float mvmMatrix2 = modelViewMatrix.getMatrix().get(6);
		float mvmMatrix3 = modelViewMatrix.getMatrix().get(10);
		
		for (int j=0; j<allCells.size(); j++){
			Cell c = allCells.get(j);
			c.visibleZ = Float.POSITIVE_INFINITY;
			for (int i=0; i<8; i++){
				float z1 = mvmMatrix1 * c.corners[i][0] 
						+ mvmMatrix2 * c.corners[i][1] 
						+ mvmMatrix3 * c.corners[i][2];
				if (z1 < c.visibleZ) c.visibleZ = z1;
			}
		}
		
		Collections.sort(allCells, cellComparator);
	}
	
	public List<Cell> getRenderableCells() {
		while (runningTasks != 0){	//make sure that no tasks are currently running
			try {
				Thread.sleep(20l);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		return allCells;
	}
		
	class Cell implements Coordinate, Callable<Void>{
		private List<RenderableAtom> atoms;
		private float centerX, centerY, centerZ;
		private float sizeX,sizeY,sizeZ;
		private float[][] corners = new float[8][3];
		private Cell[] subCells = null;
		
		private float visibleZ;
		
		public Cell(float centerX, float centerY, float centerZ, float sizeX, float sizeY, float sizeZ) {
			super();
			this.centerX = centerX;
			this.centerY = centerY;
			this.centerZ = centerZ;
			this.sizeX = sizeX;
			this.sizeY = sizeY;
			this.sizeZ = sizeZ;
			
			this.corners[0][0] = centerX-sizeX*0.5f-maxSphereSize;
			this.corners[1][0] = centerX+sizeX*0.5f+maxSphereSize;
			this.corners[2][0] = centerX-sizeX*0.5f-maxSphereSize;
			this.corners[3][0] = centerX+sizeX*0.5f+maxSphereSize;
			this.corners[4][0] = centerX-sizeX*0.5f-maxSphereSize;
			this.corners[5][0] = centerX+sizeX*0.5f+maxSphereSize;
			this.corners[6][0] = centerX-sizeX*0.5f-maxSphereSize;
			this.corners[7][0] = centerX+sizeX*0.5f+maxSphereSize;
			
			this.corners[0][1] = centerY-sizeY*0.5f-maxSphereSize;
			this.corners[1][1] = centerY-sizeY*0.5f-maxSphereSize;
			this.corners[2][1] = centerY+sizeY*0.5f+maxSphereSize;
			this.corners[3][1] = centerY+sizeY*0.5f+maxSphereSize;
			this.corners[4][1] = centerY-sizeY*0.5f-maxSphereSize;
			this.corners[5][1] = centerY-sizeY*0.5f-maxSphereSize;
			this.corners[6][1] = centerY+sizeY*0.5f+maxSphereSize;
			this.corners[7][1] = centerY+sizeY*0.5f+maxSphereSize;
			
			this.corners[0][2] = centerZ-sizeZ*0.5f-maxSphereSize;
			this.corners[1][2] = centerZ-sizeZ*0.5f-maxSphereSize;
			this.corners[2][2] = centerZ-sizeZ*0.5f-maxSphereSize;
			this.corners[3][2] = centerZ-sizeZ*0.5f-maxSphereSize;
			this.corners[4][2] = centerZ+sizeZ*0.5f+maxSphereSize;
			this.corners[5][2] = centerZ+sizeZ*0.5f+maxSphereSize;
			this.corners[6][2] = centerZ+sizeZ*0.5f+maxSphereSize;
			this.corners[7][2] = centerZ+sizeZ*0.5f+maxSphereSize;
		}
		
		public int getNumAtoms() {
			return atoms.size();
		}

		@Override
		public float getX() {
			return centerX;
		}
		
		@Override
		public float getY() {
			return centerY;
		}
		
		@Override
		public float getZ() {
			return centerZ;
		}
		
		private void shrink(){
			if (atoms.size()==0) return;
			
			float minX = Float.POSITIVE_INFINITY;
			float maxX = Float.NEGATIVE_INFINITY;
			float minY = Float.POSITIVE_INFINITY;
			float maxY = Float.NEGATIVE_INFINITY;
			float minZ = Float.POSITIVE_INFINITY;
			float maxZ = Float.NEGATIVE_INFINITY;
						
			for (int i=0; i<atoms.size(); i++){
				Atom a = atoms.get(i).atom;
				if (a.getX() < minX) minX = a.getX();
				if (a.getX() > maxX) maxX = a.getX();
				if (a.getY() < minY) minY = a.getY();
				if (a.getY() > maxY) maxY = a.getY();
				if (a.getZ() < minZ) minZ = a.getZ();
				if (a.getZ() > maxZ) maxZ = a.getZ();
			}
			this.centerX = minX+(maxX-minX)*0.5f;
			this.sizeX = maxX-minX;
			this.centerY = minY+(maxY-minY)*0.5f;
			this.sizeY = maxY-minY;
			this.centerZ = minZ+(maxZ-minZ)*0.5f;
			this.sizeZ = maxZ-minZ;				
			
			this.corners[0][0] = centerX-sizeX*0.5f-maxSphereSize;
			this.corners[1][0] = centerX+sizeX*0.5f+maxSphereSize;
			this.corners[2][0] = centerX-sizeX*0.5f-maxSphereSize;
			this.corners[3][0] = centerX+sizeX*0.5f+maxSphereSize;
			this.corners[4][0] = centerX-sizeX*0.5f-maxSphereSize;
			this.corners[5][0] = centerX+sizeX*0.5f+maxSphereSize;
			this.corners[6][0] = centerX-sizeX*0.5f-maxSphereSize;
			this.corners[7][0] = centerX+sizeX*0.5f+maxSphereSize;
			this.corners[0][1] = centerY-sizeY*0.5f-maxSphereSize;
			this.corners[1][1] = centerY-sizeY*0.5f-maxSphereSize;
			this.corners[2][1] = centerY+sizeY*0.5f+maxSphereSize;
			this.corners[3][1] = centerY+sizeY*0.5f+maxSphereSize;
			this.corners[4][1] = centerY-sizeY*0.5f-maxSphereSize;
			this.corners[5][1] = centerY-sizeY*0.5f-maxSphereSize;
			this.corners[6][1] = centerY+sizeY*0.5f+maxSphereSize;
			this.corners[7][1] = centerY+sizeY*0.5f+maxSphereSize;
			this.corners[0][2] = centerZ-sizeZ*0.5f-maxSphereSize;
			this.corners[1][2] = centerZ-sizeZ*0.5f-maxSphereSize;
			this.corners[2][2] = centerZ-sizeZ*0.5f-maxSphereSize;
			this.corners[3][2] = centerZ-sizeZ*0.5f-maxSphereSize;
			this.corners[4][2] = centerZ+sizeZ*0.5f+maxSphereSize;
			this.corners[5][2] = centerZ+sizeZ*0.5f+maxSphereSize;
			this.corners[6][2] = centerZ+sizeZ*0.5f+maxSphereSize;
			this.corners[7][2] = centerZ+sizeZ*0.5f+maxSphereSize;
			
		}
		
		private void subdivide(){
			if (this.getNumAtoms() <= MAX_ATOMS_PER_CELL){					
				this.shrink();
				if (this.getNumAtoms()>0) allCells.add(this);
				return;
			}
			
			this.subCells = new Cell[2];
			
			int longestAxis = 0;
			if (sizeX>sizeY && sizeX>sizeZ){
				longestAxis = 0;
			} else if(sizeY>sizeZ)
				longestAxis = 1;
			else longestAxis = 2;
			
			float cutPoint=0f;
			
			//If the number of atoms per cell is large, perform only a rough balancing
			//by splitting the cell in half
			//If fewer atoms remain, switch to a more precise scheme 
			//where the cell is cut at the center of mass
			if (atoms.size()<32*MAX_ATOMS_PER_CELL){
				if (longestAxis == 0) {
					for (int i = 0; i < atoms.size(); i++)
						cutPoint += atoms.get(i).atom.getX();
				}
				else if (longestAxis == 1) {
					for (int i = 0; i < atoms.size(); i++)
						cutPoint += atoms.get(i).atom.getY();
				}
				else if (longestAxis == 2) {
					for (int i = 0; i < atoms.size(); i++)
						cutPoint += atoms.get(i).atom.getZ();
				}
				cutPoint/=atoms.size();
			} else {
				if (longestAxis == 0)
					cutPoint = this.centerX;
				if (longestAxis == 1)
					cutPoint = this.centerY;
				if (longestAxis == 2)
					cutPoint = this.centerZ;
			}
			
			float sX = this.sizeX * (longestAxis == 0 ? 0.5f : 1f);
			float sY = this.sizeY * (longestAxis == 1 ? 0.5f : 1f);
			float sZ = this.sizeZ * (longestAxis == 2 ? 0.5f : 1f);
			
			float centerShiftX = this.sizeX * (longestAxis == 0 ? 0.25f : 0f);
			float centerShiftY = this.sizeY * (longestAxis == 1 ? 0.25f : 0f);
			float centerShiftZ = this.sizeZ * (longestAxis == 2 ? 0.25f : 0f);
			
			this.subCells[0] = new Cell(centerX-centerShiftX, centerY-centerShiftY, centerZ-centerShiftZ, sX, sY, sZ); 
			this.subCells[1] = new Cell(centerX+centerShiftX, centerY+centerShiftY, centerZ+centerShiftZ, sX, sY, sZ);
			
			for (int i=0; i<2; i++)
				this.subCells[i].atoms = new ArrayList<AtomRenderData.RenderableAtom>(atoms.size()/2 + 5);
			
			if (longestAxis == 0){
				for (int i=0; i<atoms.size(); i++){
					RenderableAtom ra = atoms.get(i);
					if (ra.atom.getX() < cutPoint) subCells[0].atoms.add(ra);
					else subCells[1].atoms.add(ra);
				}
			} else if (longestAxis == 1){
				for (int i=0; i<atoms.size(); i++){
					RenderableAtom ra = atoms.get(i);
					if (ra.atom.getY() < cutPoint) subCells[0].atoms.add(ra);
					else subCells[1].atoms.add(ra);
				}
			}else if (longestAxis == 2){
				for (int i=0; i<atoms.size(); i++){
					RenderableAtom ra = atoms.get(i);
					if (ra.atom.getZ() < cutPoint) subCells[0].atoms.add(ra);
					else subCells[1].atoms.add(ra);
				}
			}
			
			this.atoms = null;
			
			incRunningTasks();
			ThreadPool.submit(subCells[0]);
			incRunningTasks();
			ThreadPool.submit(subCells[1]);
		}
		
		public float[] getSize(){
			return new float[]{corners[7][0]-corners[0][0],
					corners[7][1]-corners[0][1],
					corners[7][2]-corners[0][2]};
		}
		
		public float[] getOffset(){
			return new float[]{corners[0][0],
					corners[0][1],
					corners[0][2]};
		}
		
		public List<RenderableAtom> getAtoms() {
			return atoms;
		}

		@Override
		public Void call() throws Exception {
			this.subdivide();
			decRunningTasks();
			return null;
		}
	}
	
	public static class RenderableAtom{
		public final Atom atom;
		public float[] color;
		
		public RenderableAtom(Atom atom, float[] color) {
			this.atom = atom;
			this.color = color;
		}
	}
	
	/**
	 * Comparator to sort the cells with decreasing z-coordinates 
	 * with respect to the current modelViewMatrix 
	 */
	private class CellComparator implements Comparator<Cell>{
		@Override
		public int compare(Cell c1, Cell c2) {
			if (c1.visibleZ<c2.visibleZ) return 1;
			if (c1.visibleZ>c2.visibleZ) return -1;
			return 0;
		}
	}
}


