// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package gui;
import javax.media.opengl.GL2;

public class FrameBufferObject {
	
	private int height, width, fbo, depthBuf, texture;
	private boolean destroyed;
	
	public FrameBufferObject(int minWidth, int minHeight, GL2 gl){
		width = (int)Math.pow(2, Math.ceil(Math.log(minWidth)/Math.log(2)));
		height = (int)Math.pow(2, Math.ceil(Math.log(minHeight)/Math.log(2)));
		if (height<16) height = 16;
		if (width<16) width = 16;
		
		int[] buf = new int[1];
		gl.glGenFramebuffers(1, buf, 0);
		fbo = buf[0];
		
		gl.glBindFramebuffer(GL2.GL_FRAMEBUFFER, fbo);
		gl.glGenRenderbuffers(1, buf, 0);
		depthBuf = buf[0];
		
		gl.glBindRenderbuffer(GL2.GL_RENDERBUFFER, depthBuf);
		gl.glRenderbufferStorage(GL2.GL_RENDERBUFFER, GL2.GL_DEPTH_COMPONENT, width, height);
		gl.glFramebufferRenderbuffer(GL2.GL_FRAMEBUFFER, GL2.GL_DEPTH_ATTACHMENT, GL2.GL_RENDERBUFFER, depthBuf); 
		gl.glGenTextures(1, buf, 0);
		texture = buf[0];
				
		gl.glBindTexture(GL2.GL_TEXTURE_2D, texture);
		gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MAG_FILTER, GL2.GL_NEAREST);
		gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MIN_FILTER, GL2.GL_NEAREST);
		
		gl.glTexImage2D(GL2.GL_TEXTURE_2D, 0, GL2.GL_RGBA,  width, height, 0, GL2.GL_RGBA, GL2.GL_UNSIGNED_BYTE, null);
		gl.glFramebufferTexture2D(GL2.GL_FRAMEBUFFER, GL2.GL_COLOR_ATTACHMENT0, GL2.GL_TEXTURE_2D, texture, 0);
		checkFBO(gl);
		destroyed = false;
	}
	
	private void checkFBO(GL2 gl) {
		int error = gl.glCheckFramebufferStatus(GL2.GL_FRAMEBUFFER);
		switch (error) {
		case GL2.GL_FRAMEBUFFER_COMPLETE:
			break;
		case GL2.GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT:
			throw new RuntimeException(" FBO: Incomplete attachment");
		case GL2.GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT:
			throw new RuntimeException(" FBO: Missing attachment");
		case GL2.GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS:
			throw new RuntimeException(" FBO: Incomplete dimensions");
		case GL2.GL_FRAMEBUFFER_INCOMPLETE_FORMATS:
			throw new RuntimeException(" FBO: Incomplete formats");
		case GL2.GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER:
			throw new RuntimeException(" FBO: Incomplete draw buffer");
		case GL2.GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER:
			throw new RuntimeException(" FBO: Incomplete read buffer");
		case GL2.GL_FRAMEBUFFER_UNSUPPORTED:
			throw new RuntimeException(" FBO: Framebufferobjects unsupported");
		}
	}
	
	public void bind(GL2 gl, boolean filter){
		gl.glBindTexture(GL2.GL_TEXTURE_2D, texture);
		if (filter){
			gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MAG_FILTER, GL2.GL_LINEAR);
			gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MIN_FILTER, GL2.GL_LINEAR);	
		} else {
			gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MAG_FILTER, GL2.GL_NEAREST);
			gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MIN_FILTER, GL2.GL_NEAREST);
		}

		gl.glBindFramebuffer(GL2.GL_FRAMEBUFFER, fbo);
		gl.glPushAttrib(GL2.GL_VIEWPORT_BIT);
		gl.glViewport(0, 0, width, height);
	}
	
	public void unbind(GL2 gl){
		gl.glPopAttrib();
		gl.glBindFramebuffer(GL2.GL_FRAMEBUFFER, 0);
		gl.glBindTexture(GL2.GL_TEXTURE_2D, 0);
	}
	
	public void destroy(GL2 gl){
		if (destroyed) return;
		gl.glDeleteFramebuffers(1, new int[] { fbo }, 0);
		gl.glDeleteRenderbuffers(1, new int[] { depthBuf }, 0);
		gl.glDeleteTextures(1, new int[] { texture }, 0);
		destroyed = true;
	}
	
	public int getTextureName(){
		if (destroyed) throw new RuntimeException("FBO is not valid anymore");
		return texture;
	}
}
