// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package gui;

import gui.JMainWindow.KeyBoardAction;
import gui.ViewerGLJPanel.AtomRenderType;
import gui.ViewerGLJPanel.RenderOption;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import common.ColorTable;

import model.*;
import model.dataContainer.DataContainer;
import model.dislocationDensity.DislocationDensityTensorData;

public class JAtomicMenuPanel extends JPanel{
	private static final long serialVersionUID = 1L;
	
	private ViewerGLJPanel viewer;
	
	private JLabel totalAtomsLabel;
	private JLabel crystalStructureLabel;
	private JLabel[] boxSizeLabel = new JLabel[3];
	private JLabel[] numberOftypeLabels = new JLabel[0];
	private JIgnoreTypeCheckbox[] ignoreTypeCheckbox = new JIgnoreTypeCheckbox[0];
	private JIgnoreRBVCheckbox[] ignoreRBVCheckbox = new JIgnoreRBVCheckbox[0];
	private JTypeColorPanel[] typeColorPanel = new JTypeColorPanel[0];
	private Container typeIgnoreContainer = new Container(); 
	
	private JLabel timeStepLabel = new JLabel();
	
	private JToggleButton atomsToggleButton = new JToggleButton("Hide");
	private JToggleButton rbvToggleButton = new JToggleButton("Hide");
	
	private JButton forwardButton = new JButton(">");
	private JButton rewindButton = new JButton("<");
	private JButton fastForwardButton = new JButton(">>");
	private JButton fastRewindButton = new JButton("<<");
	
	private JRadioButton drawAsTypesButton = new JRadioButton("Atoms");
	private JRadioButton drawAsGrainsButton = new JRadioButton("Grains");
	private JRadioButton drawAsElementsButton = new JRadioButton("Elements");
	private JRadioButton drawAsCustomButton = new JRadioButton("Raw Values");
	private JRadioButton drawAsRotationButton = new JRadioButton("Lattice rotation");
	private JCheckBox drawClusterCheckBox = new JCheckBox("Grain boundaries");
	
	private Container elementIgnoreContainer = new Container();	
	private Container grainIgnoreContainer = new Container();
	private JScrollPane elementScrollPane;
	private JScrollPane grainScrollPane;
	
	private JLatticeRotationControlPanel latticeRotationPanel;
	private JCustomColumnControlPanel customPanel;
	
	private JPanel customDataPanel = new JPanel();
	private GridBagConstraints customDataPanelContraints = new GridBagConstraints();
 	
	private JDislocationMenuPanel dislocationMenu = new JDislocationMenuPanel();
	private DislocationDensityTensorData.JDislocationDensityTensorControls ddtPanel = 
		DislocationDensityTensorData.JDislocationDensityTensorControls.getInstance();
	
	private AtomData atomData;
	
	public JAtomicMenuPanel(KeyBoardAction kba){
		Container cont = new Container();
		cont.setLayout(new GridBagLayout());
		
		GridBagConstraints gbc = new GridBagConstraints();
		
		gbc.anchor = GridBagConstraints.WEST;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.gridwidth = 2;
		gbc.weightx = 0.5;
		
		gbc.gridy = 0; cont.add(totalAtomsLabel = new JLabel("Total number of atoms: "), gbc);
		gbc.gridy++; cont.add(timeStepLabel, gbc);
		gbc.gridy++; cont.add(boxSizeLabel[0] = new JLabel("Boxsize X: "), gbc);
		gbc.gridy++; cont.add(boxSizeLabel[1] = new JLabel("Boxsize Y: "), gbc);
		gbc.gridy++; cont.add(boxSizeLabel[2] = new JLabel("Boxsize Z: "), gbc);
		gbc.gridy++;
		
		JPanel infoPanel = new JPanel();
		infoPanel.setLayout(new GridBagLayout());
		GridBagConstraints gbcCont = new GridBagConstraints();
		infoPanel.setBorder(new TitledBorder(new EtchedBorder(1), "Atoms"));
		gbcCont.anchor = GridBagConstraints.WEST;
		gbcCont.fill = GridBagConstraints.HORIZONTAL;
		gbcCont.gridwidth = 3;
		gbcCont.weightx = 1;
		gbcCont.gridx = 0;
		gbcCont.gridy = 0;
		infoPanel.add(crystalStructureLabel = new JLabel(), gbcCont); gbcCont.gridy++;
		
		typeIgnoreContainer.setLayout(new GridBagLayout());
		fillIgnoreBoxPanel();
		infoPanel.add(typeIgnoreContainer, gbcCont); gbcCont.gridx++;
		
		rbvToggleButton.setSelected(true);
		rbvToggleButton.setEnabled(false);
		
		cont.add(infoPanel, gbc); gbc.gridy++;
		
		atomsToggleButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				if (atomData == null) return;
				if (atomsToggleButton.isSelected()){
					for (int i=0; i<ignoreTypeCheckbox.length; i++){
						ignoreTypeCheckbox[i].setEnabled(false);
						viewer.setTypeIgnored(i, true);
					}
				} else {
					for (int i=0; i<ignoreTypeCheckbox.length; i++){
						ignoreTypeCheckbox[i].setEnabled(true);
						viewer.setTypeIgnored(i, !ignoreTypeCheckbox[i].isSelected());
					}
				}
				viewer.updateAtoms();
			}
		});
		
		rbvToggleButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				if (atomData == null) return;
				if (rbvToggleButton.isSelected()){
					for (int i=0; i<ignoreRBVCheckbox.length; i++){
						ignoreRBVCheckbox[i].setEnabled(false);
						viewer.setBVIgnored(i, true);
					}
				} else {
					for (int i=0; i<ignoreRBVCheckbox.length; i++){
						ignoreRBVCheckbox[i].setEnabled(true);
						viewer.setBVIgnored(i, !ignoreRBVCheckbox[i].isSelected());
					}
				}
				viewer.reDraw();
			}
		});
		
		ButtonGroup bg = new ButtonGroup();
		bg.add(drawAsTypesButton);
		bg.add(drawAsElementsButton);
		bg.add(drawAsGrainsButton);
		bg.add(drawAsCustomButton);
		bg.add(drawAsRotationButton);
		
		ActionListener al = new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent arg0) {
				elementScrollPane.setVisible(false);
				if (!drawClusterCheckBox.isVisible() || !drawClusterCheckBox.isSelected())
					grainScrollPane.setVisible(false);
				latticeRotationPanel.setVisible(false);
				customPanel.setVisible(false);
				
				if (arg0.getActionCommand().equals(drawAsTypesButton.getText())){
					if (viewer!=null) viewer.setAtomRenderMethod(AtomRenderType.ATOMS);
				} else if (arg0.getActionCommand().equals(drawAsCustomButton.getText())){
					if (viewer!=null) viewer.setAtomRenderMethod(AtomRenderType.CUSTOM);
					customPanel.setVisible(true);
				}else if (arg0.getActionCommand().equals(drawAsElementsButton.getText())){
					if (viewer!=null) viewer.setAtomRenderMethod(AtomRenderType.ELEMENTS);
					elementScrollPane.setVisible(true);
				} else if (arg0.getActionCommand().equals(drawAsGrainsButton.getText())){
					if (viewer!=null) viewer.setAtomRenderMethod(AtomRenderType.GRAINS);
					grainScrollPane.setVisible(true);
				} else if (arg0.getActionCommand().equals(drawAsRotationButton.getText())){
					if (viewer!=null) viewer.setAtomRenderMethod(AtomRenderType.LATTICE_ROTATION);
					latticeRotationPanel.setVisible(true);
				}
				JAtomicMenuPanel.this.revalidate();	
			}
		};
		
		drawAsTypesButton.addActionListener(al);
		drawAsTypesButton.setActionCommand(drawAsTypesButton.getText());
		drawAsGrainsButton.addActionListener(al);
		drawAsGrainsButton.setActionCommand(drawAsGrainsButton.getText());
		drawAsElementsButton.addActionListener(al);
		drawAsElementsButton.setActionCommand(drawAsElementsButton.getText());
		drawAsCustomButton.addActionListener(al);
		drawAsCustomButton.setActionCommand(drawAsCustomButton.getText());
		drawAsRotationButton.addActionListener(al);
		drawAsRotationButton.setActionCommand(drawAsRotationButton.getText());
		
		cont.add(drawAsTypesButton, gbc); gbc.gridy++;
		cont.add(drawAsElementsButton, gbc); gbc.gridy++;
		cont.add(drawAsGrainsButton, gbc); gbc.gridy++;
		cont.add(drawAsCustomButton, gbc); gbc.gridy++;
		cont.add(drawAsRotationButton, gbc); gbc.gridy++;

		drawAsTypesButton.setVisible(false);
		drawAsElementsButton.setVisible(false);
		drawAsGrainsButton.setVisible(false);
		drawAsCustomButton.setVisible(false);
		drawAsRotationButton.setVisible(false);
		
		elementIgnoreContainer.setLayout(new GridBagLayout());
		elementScrollPane = new JScrollPane(elementIgnoreContainer, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		elementScrollPane.setBorder(new TitledBorder(new EtchedBorder(1), "Show Elements"));
		elementScrollPane.setPreferredSize(new Dimension(10, 120));
		elementScrollPane.setMinimumSize(new Dimension(10, 120));
		cont.add(elementScrollPane, gbc); gbc.gridy++;
		elementScrollPane.setVisible(false);
		
		latticeRotationPanel = new JLatticeRotationControlPanel();
		cont.add(latticeRotationPanel, gbc); gbc.gridy++;
		latticeRotationPanel.setVisible(false);
		
		customPanel = new JCustomColumnControlPanel();
		cont.add(customPanel, gbc); gbc.gridy++;
		customPanel.setVisible(false);
		
		drawClusterCheckBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				RenderOption.GRAINS.setEnabled(drawClusterCheckBox.isSelected());
				if (drawClusterCheckBox.isSelected()){
					grainScrollPane.setVisible(true);
				} else {
					grainScrollPane.setVisible(drawAsGrainsButton.isSelected());
				}
				JAtomicMenuPanel.this.revalidate();	
			}
		});
		cont.add(drawClusterCheckBox, gbc); gbc.gridy++;
		drawClusterCheckBox.setVisible(false);
		
		grainIgnoreContainer.setLayout(new GridBagLayout());
		grainScrollPane = new JScrollPane(grainIgnoreContainer, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		grainScrollPane.setBorder(new TitledBorder(new EtchedBorder(1), "Show Grains"));
		grainScrollPane.setPreferredSize(new Dimension(10, 120));
		grainScrollPane.setMinimumSize(new Dimension(10, 120));
		cont.add(grainScrollPane, gbc); gbc.gridy++;
		grainScrollPane.setVisible(false);
		
		cont.add(dislocationMenu, gbc); gbc.gridy++;
		dislocationMenu.setVisible(false);
		
		cont.add(ddtPanel, gbc); gbc.gridy++;
		ddtPanel.setVisible(false);

		cont.add(customDataPanel, gbc); gbc.gridy++;
		customDataPanel.setVisible(true);
		customDataPanel.setLayout(new GridBagLayout());
		customDataPanelContraints.fill = GridBagConstraints.HORIZONTAL;
		customDataPanelContraints.weightx = 1;
		customDataPanelContraints.gridx = 0;
		customDataPanelContraints.gridy = 0;
		
		gbc.weighty = 1.;
		cont.add(new JLabel(), gbc); gbc.gridy++;
		
		gbc.gridheight = GridBagConstraints.REMAINDER;
		gbc.anchor = GridBagConstraints.SOUTH;
		JPanel p = new JPanel();
		p.setLayout(new GridLayout(1,4));
		
		p.add(fastRewindButton);
		fastRewindButton.setToolTipText("First file in sequence (f)");
		fastRewindButton.addActionListener(kba);
		fastRewindButton.setActionCommand("f");
		fastRewindButton.setEnabled(false);
		p.add(rewindButton); 
		rewindButton.setToolTipText("Previous file in sequence (y or z)");
		rewindButton.addActionListener(kba);
		rewindButton.setActionCommand("z");
		rewindButton.setEnabled(false);
		p.add(forwardButton);
		forwardButton.setToolTipText("Next file in sequence (x)");
		forwardButton.addActionListener(kba);
		forwardButton.setActionCommand("x");
		forwardButton.setEnabled(false);
		p.add(fastForwardButton);
		fastForwardButton.setToolTipText("Last file in sequence (l)");
		fastForwardButton.addActionListener(kba);
		fastForwardButton.setActionCommand("l");
		fastForwardButton.setEnabled(false);
		cont.add(p, gbc);
		
		drawAsTypesButton.doClick(); //Set selected
		
		JScrollPane scrollPane = new JScrollPane(cont, 
				ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		this.setLayout(new GridLayout(1,1));
		this.add(scrollPane);
	}
	
	/**
	 * updates values like colors if they have modified
	 */
	public void updateValues(){
		for (int i=0; i<typeColorPanel.length; i++){
			JTypeColorPanel p = typeColorPanel[i];
			p.setBackground(Configuration.getCrystalStructure().getColor(i));
		}
	}
	
	private void fillIgnoreBoxPanel(){
		typeIgnoreContainer.removeAll();
		if (Configuration.getCrystalStructure() == null){
			numberOftypeLabels = new JLabel[0];
			ignoreTypeCheckbox = new JIgnoreTypeCheckbox[0];
			ignoreRBVCheckbox = new JIgnoreRBVCheckbox[0];
			return;
		} else 
			
		numberOftypeLabels = new JLabel[Configuration.getCrystalStructure().getNumberOfTypes()];
		ignoreTypeCheckbox = new JIgnoreTypeCheckbox[Configuration.getCrystalStructure().getNumberOfTypes()];
		ignoreRBVCheckbox = new JIgnoreRBVCheckbox[Configuration.getCrystalStructure().getNumberOfTypes()];
		typeColorPanel = new JTypeColorPanel[Configuration.getCrystalStructure().getNumberOfTypes()];
		
		GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.HORIZONTAL;
		c.weightx = 1.;
		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.NORTHWEST;
		for (int i=0; i<Configuration.getCrystalStructure().getNumberOfTypes();i++){
			typeColorPanel[i] = new JTypeColorPanel(i, Configuration.getCrystalStructure().getColor(i));
			c.gridwidth = 3;
			typeIgnoreContainer.add(numberOftypeLabels[i] = new JLabel(), c); c.gridy++;
			c.gridwidth = 1;
			c.weightx = 5;
			typeIgnoreContainer.add(typeColorPanel[i], c); c.gridx++;
			c.weightx = 1;
			typeIgnoreContainer.add(ignoreTypeCheckbox[i] = new JIgnoreTypeCheckbox(i), c); c.gridx++;
			typeIgnoreContainer.add(ignoreRBVCheckbox[i] = new JIgnoreRBVCheckbox(i), c);
			c.gridy++;
			c.gridx = 0;
		}
		c.gridwidth = 1; c.weightx = 5;
		typeIgnoreContainer.add(new JLabel(), c); c.gridx++;
		c.weightx = 1;
		typeIgnoreContainer.add(atomsToggleButton, c); c.gridx++;
		typeIgnoreContainer.add(rbvToggleButton, c);
	}
	
	public void setAtomData(AtomData data, ViewerGLJPanel viewer, boolean init){
		this.customDataPanel.removeAll();
		customDataPanelContraints.gridy = 0;
		for (DataContainer dc: data.getAdditionalData()){
			customDataPanel.add(dc.getDataControlPanel(),customDataPanelContraints);
			dc.getDataControlPanel().setViewer(viewer);
			customDataPanelContraints.gridy++;
		}
		customDataPanel.revalidate();
		
		this.viewer = viewer;
		this.atomData = data;
		
		if (init){
			fillIgnoreBoxPanel();
			int def = Configuration.getCrystalStructure().getDefaultType();
			atomsToggleButton.setSelected(false);
			rbvToggleButton.setSelected(true);	
			viewer.setBVIgnored(def, true);
			ignoreRBVCheckbox[def].setSelected(false);
			if (data.getNumberOfType(def)*4>data.getAtoms().size() && (data.getNumberOfType(def)!=data.getAtoms().size())){
				ignoreTypeCheckbox[def].setSelected(false);
				viewer.setTypeIgnored(def, true);
			}
		}
		
		for (int i=0; i<ignoreTypeCheckbox.length;i++){
			viewer.setBVIgnored(i, !(ignoreRBVCheckbox[i].isSelected() && ignoreRBVCheckbox[i].isEnabled()));
			viewer.setTypeIgnored(i, !(ignoreTypeCheckbox[i].isSelected() && ignoreTypeCheckbox[i].isEnabled()) );
		}
		
		if (atomData.getFileMetaData("timestep") != null){
			timeStepLabel.setText("Timestep: "+( (int)((float[])(atomData.getFileMetaData("timestep")))[0]) );
		} else timeStepLabel.setText(""); 
		
		totalAtomsLabel.setText("Total number of atoms: "+data.getAtoms().size());
		boxSizeLabel[0].setText("Boxsize X: "+data.getBounds()[0]);
		boxSizeLabel[1].setText("Boxsize Y: "+data.getBounds()[1]);
		boxSizeLabel[2].setText("Boxsize Z: "+data.getBounds()[2]);
		crystalStructureLabel.setText("Structure: "+Configuration.getCrystalStructure().toString());
		
		for (int i=0; i<numberOftypeLabels.length;i++){
			numberOftypeLabels[i].setText(
					Configuration.getCrystalStructure().getNameForType(i) +" ("+data.getNumberOfType(i)+")");
		}
		
		latticeRotationPanel.setViewer(viewer);
		customPanel.setViewer(viewer);
		
		elementIgnoreContainer.removeAll();
		GridBagConstraints c = new GridBagConstraints();
		c.fill = GridBagConstraints.HORIZONTAL;
		c.weightx = 1.;
		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.NORTHWEST;
		for (int i=0; i<Configuration.getNumElements(); i++){
			elementIgnoreContainer.add(new JIgnoreElementCheckbox(i, atomData.getNumAtomsOfElements(i)), c);
			c.gridy++;
		}
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weighty = 1.;
		elementIgnoreContainer.add(new JLabel(""),c);
		
		
		
		grainIgnoreContainer.removeAll();
		c = new GridBagConstraints();
		c.fill = GridBagConstraints.HORIZONTAL;
		c.weightx = 1.;
		c.gridx = 0;
		c.gridy = 0;
		c.anchor = GridBagConstraints.NORTHWEST;
		ArrayList<Integer> sortedList = new ArrayList<Integer>(Configuration.getGrainIndices());
		Collections.sort(sortedList);
		for (int i : sortedList){
			grainIgnoreContainer.add(new JIgnoreGrainCheckbox(i), c);
			c.gridy++;
		}
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.weighty = 1.;
		grainIgnoreContainer.add(new JLabel(""),c);
		
		
		if (Configuration.getNumElements() < 2){
			if (drawAsElementsButton.isSelected())
				drawAsTypesButton.doClick();
			drawAsElementsButton.setVisible(false);
		} else drawAsElementsButton.setVisible(true);
		
		if (!ImportStates.POLY_MATERIAL.isActive()){
			drawAsGrainsButton.setVisible(false);
			drawClusterCheckBox.setVisible(false);
			grainScrollPane.setVisible(false);
		} else {
			drawAsGrainsButton.setVisible(true);
			drawClusterCheckBox.setVisible(true);
			if (drawClusterCheckBox.isSelected())
				grainScrollPane.setVisible(true);
		}
		
		//Check if there are any none special columns that need to be displayed  
		boolean enableCustom = false;
		for (int i=0; i<Configuration.getSizeCustomColumns();i++)
			if (!Configuration.getCustomColumnInfo(i).isSpecialColoumn())
				enableCustom = true;
	
		enableCustom = enableCustom && ImportStates.CUSTOM_VALUES.isActive();
		
		drawAsCustomButton.setVisible(enableCustom);
		if (drawAsCustomButton.isSelected() && !enableCustom){
			drawAsTypesButton.doClick();
		}
			
		if (init) {
			customPanel.resetValues();
			customPanel.resetDropDown();
			
			for (int i=0; i< ignoreRBVCheckbox.length; i++)
				if (!rbvToggleButton.isSelected()) ignoreRBVCheckbox[i].setEnabled(data.isRbvAvailable());
			rbvToggleButton.setEnabled(data.isRbvAvailable());
		}
		
		ddtPanel.setViewer(viewer, data, init);
		ddtPanel.setVisible(ImportStates.DISLOCATION_DENSITY.isActive());
		
		drawAsRotationButton.setVisible(ImportStates.LATTICE_ROTATION.isActive());
		
		this.dislocationMenu.setAtomData(viewer);
		this.dislocationMenu.setVisible(ImportStates.SKELETONIZE.isActive());
		
		drawAsTypesButton.setVisible(ImportStates.DETECT_MARTENSITE_VARIANTS.isActive() || ImportStates.CUSTOM_VALUES.isActive() ||
				ImportStates.LATTICE_ROTATION.isActive() || Configuration.getNumElements() >= 2 ||
				ImportStates.POLY_MATERIAL.isActive());
		
		if (fastForwardButton.isEnabled() != (data.getNext() != null)) fastForwardButton.setEnabled(data.getNext() != null);
		if (forwardButton.isEnabled() != (data.getNext() != null)) forwardButton.setEnabled(data.getNext() != null);
		if (fastRewindButton.isEnabled() != (data.getPrevious() != null)) fastRewindButton.setEnabled(data.getPrevious() != null);
		if (rewindButton.isEnabled() != (data.getPrevious() != null))rewindButton.setEnabled(data.getPrevious() != null);
		
		this.revalidate();
		this.repaint();
	}
	
	private class JIgnoreTypeCheckbox extends JCheckBox{
		private static final long serialVersionUID = 1L;		
		
		public JIgnoreTypeCheckbox(final int number){
			this.setText("Atom");
			this.setSelected(true);
			this.addActionListener(new ActionListener(){
				@Override
				public void actionPerformed(ActionEvent e) {
					if (atomData != null){
						viewer.setTypeIgnored(number, !JIgnoreTypeCheckbox.this.isSelected());
						viewer.updateAtoms();
					}
				}
			});
		}
	}
	
	private class JIgnoreRBVCheckbox extends JCheckBox{
		private static final long serialVersionUID = 1L;
		
		public JIgnoreRBVCheckbox(final int number){
			this.setText("RBV");
			this.setSelected(true);
			this.setEnabled(false);
			this.addActionListener(new ActionListener(){
				@Override
				public void actionPerformed(ActionEvent e) {
					if (atomData!=null){
						viewer.setBVIgnored(number, !JIgnoreRBVCheckbox.this.isSelected());
						viewer.reDraw();
					}
				}
			});
		}
	}
	
	private class JTypeColorPanel extends JPanel{
		private static final long serialVersionUID = 1L;
		
		public JTypeColorPanel(final int number, Color c){
			this.setSize(100, 20);
			this.setBackground(Configuration.getCrystalStructure().getColor(number));
			this.setBorder(new LineBorder(Color.BLACK, 1));
			this.addMouseListener(new MouseListener() {
				@Override
				public void mouseReleased(MouseEvent e) {}				
				@Override
				public void mousePressed(MouseEvent e) {}
				@Override
				public void mouseExited(MouseEvent e) {}
				@Override
				public void mouseEntered(MouseEvent e) {}
				
				@Override
				public void mouseClicked(MouseEvent e) {
					Color newColor = JColorChooser.showDialog(null, 
							String.format("Choose color for type %d (%s)", number, 
									Configuration.getCrystalStructure().getNameForType(number)), 
									JTypeColorPanel.this.getBackground());
					if (newColor == null) return;
					JTypeColorPanel.this.setBackground(newColor);
					float[] color = new float[3];
					color[0] = newColor.getRed()/255f;
					color[1] = newColor.getGreen()/255f;
					color[2] = newColor.getBlue()/255f;
					Configuration.getCrystalStructure().setGLColors(number, color);
					viewer.reDraw();
				}
			});
		}
	}
	
	private class JIgnoreElementCheckbox extends JCheckBox{
		private static final long serialVersionUID = 1L;
		
		public JIgnoreElementCheckbox(final int number, final int numberOfAtoms){
			this.setText(Integer.toString(number)+ " ("+ Integer.toString(numberOfAtoms)+")");
			this.setSelected(!viewer.isElementIgnored(number));
			float[][] color = ColorTable.createColorTable(Configuration.getNumElements());			
			this.setBackground(new Color(color[number][0], color[number][1], color[number][2]));
			this.setSize(new Dimension(200, 20));
			this.addActionListener(new ActionListener(){
				@Override
				public void actionPerformed(ActionEvent e) {
					if (atomData!=null){
						viewer.setElementIgnored(number, !JIgnoreElementCheckbox.this.isSelected());
						viewer.updateAtoms();
					}
				}
			});
		}
	}
	
	private class JIgnoreGrainCheckbox extends JCheckBox{
		private static final long serialVersionUID = 1L;
		
		public JIgnoreGrainCheckbox(final int number){
			this.setText(Integer.toString(number));
			if (number == Atom.IGNORED_GRAIN) this.setText("none"); 
			if (number == Atom.DEFAULT_GRAIN) this.setText("default");
			this.setSelected(!viewer.isGrainIgnored(number));
			float[] color = viewer.getGrainColor(number);
			this.setBackground(new Color(color[0], color[1], color[2]));
			this.setSize(new Dimension(200, 20));
			this.addActionListener(new ActionListener(){
				@Override
				public void actionPerformed(ActionEvent e) {
					if (atomData!=null){
						viewer.setGrainIgnored(number, !JIgnoreGrainCheckbox.this.isSelected());
						viewer.updateAtoms();
					}
				}
			});
		}
	}
	
	private static class JCustomColumnControlPanel extends JPanel {
		private static final long serialVersionUID = 1L;
		
		private JSpinner lowerLimitSpinner = new JSpinner(new SpinnerNumberModel(0., -50., 50., 0.0001));
		private JSpinner upperLimitSpinner = new JSpinner(new SpinnerNumberModel(0., -50., 50., 0.0001));
		private JButton resetButton = new JButton("Now");
		private JButton resetAllButton = new JButton("Now");
		private JCheckBox filterCheckbox = new JCheckBox("Filter");
		private JCheckBox inverseFilterCheckbox = new JCheckBox("Inverse");
		private int selected = -1;
		private CustomColumnInfo selectedColumn;
		private JComboBox valueComboBox = new JComboBox();
		
		private ViewerGLJPanel viewer;
		
		public JCustomColumnControlPanel() {
			((JSpinner.NumberEditor)lowerLimitSpinner.getEditor()).getFormat().setMaximumFractionDigits(4);
			((JSpinner.NumberEditor)upperLimitSpinner.getEditor()).getFormat().setMaximumFractionDigits(4);
			 
			((SpinnerNumberModel)(lowerLimitSpinner.getModel())).setMinimum(null);
			((SpinnerNumberModel)(upperLimitSpinner.getModel())).setMinimum(null);
			((SpinnerNumberModel)(upperLimitSpinner.getModel())).setMaximum(null);
			((SpinnerNumberModel)(lowerLimitSpinner.getModel())).setMaximum(null);
			
			this.setBorder(new TitledBorder(new EtchedBorder(1), "Values"));
			
			this.setLayout(new GridBagLayout());
			
			GridBagConstraints gbc = new GridBagConstraints();
			
			gbc.anchor = GridBagConstraints.WEST;
			gbc.weightx = 1;
			gbc.fill = GridBagConstraints.BOTH;
			gbc.gridy = 0; gbc.gridx = 0;
			gbc.gridwidth = 2;
			gbc.gridx = 0; gbc.gridy++;
			this.add(valueComboBox, gbc); gbc.gridy++;
			gbc.gridwidth = 1;
			
			filterCheckbox.setSelected(false);
			this.add(new JLabel("Min."), gbc);gbc.gridx++;
			this.add(new JLabel("Max."), gbc);
			gbc.gridx = 0; gbc.gridy++;
			this.add(lowerLimitSpinner, gbc);gbc.gridx++;
			this.add(upperLimitSpinner, gbc);
			gbc.gridx = 0; gbc.gridy++;
			this.add(new JLabel("Auto (this file)"), gbc);gbc.gridx++;
			this.add(resetButton, gbc);
			gbc.gridx = 0; gbc.gridy++;
			this.add(new JLabel("Auto (all files)"), gbc);gbc.gridx++;
			this.add(resetAllButton, gbc);
			gbc.gridx = 0; gbc.gridy++;
			gbc.gridwidth = 1;
			this.add(filterCheckbox, gbc);
			gbc.gridx++;
			this.add(inverseFilterCheckbox, gbc);
			gbc.gridx = 0; gbc.gridy++;
			
			valueComboBox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					selected = valueComboBox.getSelectedIndex();
					selectedColumn = (CustomColumnInfo)valueComboBox.getSelectedItem();
					setSpinner();
					Configuration.setSelectedColumn(selectedColumn);
					viewer.updateAtoms();
				}
			});
			
			lowerLimitSpinner.addChangeListener(new ChangeListener() {
				@Override
				public void stateChanged(ChangeEvent e) {
					selectedColumn.setLowerLimit(((Number)lowerLimitSpinner.getValue()).floatValue());
					viewer.updateAtoms();
				}
			});
			
			upperLimitSpinner.addChangeListener(new ChangeListener() {
				@Override
				public void stateChanged(ChangeEvent e) {
					selectedColumn.setUpperLimit(((Number)upperLimitSpinner.getValue()).floatValue());
					viewer.updateAtoms();
				}
			});
			
			filterCheckbox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					Configuration.setFilterRange(filterCheckbox.isSelected());
					viewer.updateAtoms();
					inverseFilterCheckbox.setEnabled(filterCheckbox.isSelected());
				}
			});
			
			inverseFilterCheckbox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					Configuration.setFilterInversed(inverseFilterCheckbox.isSelected());
					viewer.updateAtoms();
				}
			});
			
			resetAllButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					selectedColumn.findRange(true);
					lowerLimitSpinner.setValue(selectedColumn.getLowerLimit());
					upperLimitSpinner.setValue(selectedColumn.getUpperLimit());
					viewer.updateAtoms();
				}
			});
			
			resetButton.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					selectedColumn.findRange(false);
					lowerLimitSpinner.setValue(selectedColumn.getLowerLimit());
					upperLimitSpinner.setValue(selectedColumn.getUpperLimit());
					viewer.updateAtoms();
				}
			});
		}
		
		@Override
		public void setVisible(boolean aFlag) {
			super.setVisible(aFlag);
			this.filterCheckbox.setSelected(Configuration.isFilterRange());
			this.inverseFilterCheckbox.setSelected(Configuration.isFilterInversed());
			this.inverseFilterCheckbox.setEnabled(this.filterCheckbox.isSelected());
		}
		
		private void setSpinner(){
			if (selectedColumn == null) return;
			
			lowerLimitSpinner.setValue(selectedColumn.getLowerLimit());
			upperLimitSpinner.setValue(selectedColumn.getUpperLimit());
			
			lowerLimitSpinner.setEnabled(!selectedColumn.isFixedRange());
			upperLimitSpinner.setEnabled(!selectedColumn.isFixedRange());
			resetAllButton.setEnabled(!selectedColumn.isFixedRange());
			resetButton.setEnabled(!selectedColumn.isFixedRange());
		}
		
		public void resetValues(){
			if (selectedColumn!= null && Configuration.getSizeCustomColumns()!=0) selectedColumn.findRange(false);
		}
		
		public void resetDropDown(){
			valueComboBox.removeAllItems();
			for (int i = 0; i<Configuration.getSizeCustomColumns(); i++)
				if (!Configuration.getCustomColumnInfo(i).isSpecialColoumn())
					valueComboBox.addItem(Configuration.getCustomColumnInfo(i));
			if (selected > Configuration.getSizeCustomColumns()){
				selected = 0;
				selectedColumn = (CustomColumnInfo)valueComboBox.getItemAt(0);
				setSpinner();
			}
			else if (Configuration.getSizeCustomColumns()>0){
				valueComboBox.setSelectedIndex(selected);
				selectedColumn = (CustomColumnInfo)valueComboBox.getItemAt(selected);
				setSpinner();
			}
			Configuration.setSelectedColumn(selectedColumn);
		}
		
		public void setViewer(ViewerGLJPanel viewer){
			if (this.viewer==null) this.viewer = viewer;
		}
	}
	
	public static class JLatticeRotationControlPanel extends JPanel {
		private static final long serialVersionUID = 1L;
		private JComboBox axisComboBox = new JComboBox(new String[]{"X-Axis","Y-Axis","Z-Axis","Absolute Tilt"});

		private JFormattedTextField lowerLimitTextBox = new JFormattedTextField(new DecimalFormat());
		private JFormattedTextField upperLimitTextBox = new JFormattedTextField(new DecimalFormat());
		private JCheckBox filterCheckbox = new JCheckBox("Filter");
		private JCheckBox inverseFilterCheckbox = new JCheckBox("Inverse");
		
		private ViewerGLJPanel viewer;
		
		public JLatticeRotationControlPanel() {
			this.setBorder(new TitledBorder(new EtchedBorder(1), "LatticeRotation"));
			this.setLayout(new GridBagLayout());
			GridBagConstraints gbc = new GridBagConstraints();

			gbc.fill = GridBagConstraints.BOTH;
			gbc.weightx = 0.33;
			gbc.gridx = 0; gbc.gridy = 2;
			gbc.gridwidth = 3;
			this.add(axisComboBox, gbc);
			axisComboBox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					int selected = axisComboBox.getSelectedIndex();
					if (viewer!=null)
						viewer.setLatticeRotationComponent(selected);
				}
			});
			
			gbc.gridwidth = 1;
			gbc.gridy = 1;
			gbc.gridx = 0;
			this.add(new JLabel("min/max angle"),gbc);
			
			gbc.gridx = 1;
			this.add(lowerLimitTextBox, gbc);
			lowerLimitTextBox.setValue(-8);
			gbc.gridx = 2;
			this.add(upperLimitTextBox, gbc);
			upperLimitTextBox.setValue(8);
			
			LimiterTextBoxChangeListener ltbcl = new LimiterTextBoxChangeListener();
			
			upperLimitTextBox.addPropertyChangeListener(ltbcl);
			lowerLimitTextBox.addPropertyChangeListener(ltbcl);
			
			gbc.gridy = 3; gbc.gridx = 0; gbc.gridwidth = 2;
			this.add(filterCheckbox, gbc);
			gbc.gridy = 4;
			this.add(inverseFilterCheckbox, gbc);
			
			
			filterCheckbox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					Configuration.setFilterRange(filterCheckbox.isSelected());
					viewer.updateAtoms();
					inverseFilterCheckbox.setEnabled(filterCheckbox.isSelected());
				}
			});
			
			inverseFilterCheckbox.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent arg0) {
					Configuration.setFilterInversed(inverseFilterCheckbox.isSelected());
					viewer.updateAtoms();
				}
			});
		}
		
		public void setViewer(ViewerGLJPanel viewer){
			if (this.viewer==null){
				this.viewer = viewer;
				axisComboBox.setSelectedIndex(0);
			}
		}
		
		@Override
		public void setVisible(boolean aFlag) {
			super.setVisible(aFlag);
			this.filterCheckbox.setSelected(Configuration.isFilterRange());
			this.inverseFilterCheckbox.setSelected(Configuration.isFilterInversed());
			this.inverseFilterCheckbox.setEnabled(this.filterCheckbox.isSelected());
		}

		private class LimiterTextBoxChangeListener implements PropertyChangeListener{
			@Override
			public void propertyChange(PropertyChangeEvent arg0) {
				if (lowerLimitTextBox.getValue()==null || viewer == null) return;
				if (upperLimitTextBox.getValue()==null || viewer == null) return;
				
				Number l = (Number)lowerLimitTextBox.getValue();
				float lower = l.floatValue();
				l = (Number)upperLimitTextBox.getValue();
				float upper = l.floatValue();
				
				if (lower!=viewer.getLatticeRotationMin() || upper!=viewer.getLatticeRotationMax()){
					viewer.setLatticeRotationMinMax(lower, upper);
				}
			}
		}
	}
}