// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package gui;

import model.*;
import model.dataContainer.DataContainer;
import model.dislocationDensity.DislocationDensityTensorData;
import model.polygrain.Grain;
import model.polygrain.mesh.*;
import model.skeletonizer.*;
import gui.AtomRenderData.Cell;
import gui.AtomRenderData.RenderableAtom;
import gui.ShaderCollection.GL3Shader;
import gui.ShaderCollection.Shader;

import java.awt.*;
import java.awt.event.*;
import java.io.IOException;
import java.nio.*;
import java.util.*;

import javax.imageio.ImageIO;
import javax.media.opengl.*;
import javax.media.opengl.awt.GLJPanel;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.util.awt.Screenshot;
import com.jogamp.opengl.util.awt.TextRenderer;
import com.jogamp.opengl.util.gl2.GLUT;

import common.*;
import common.ColorTable.ColorBarScheme;
import crystalStructures.CrystalStructure;

public class ViewerGLJPanel extends GLJPanel implements MouseMotionListener, MouseListener, 
	MouseWheelListener, GLEventListener {
	
	public enum RenderOption {
		INDENTER(false), BURGERS_VECTORS_ON_CORES(false), GRAINS(false), DDT(false), DISLOCATIONS(false),
		STACKING_FAULT(false), LEGEND(true), COORDINATE_SYSTEM(false), THOMPSON_TETRAEDER(false),
		PRINTING_MODE(false), STEREO(false), BOUNDING_BOX(true), PERSPECTIVE(false);
		
		private boolean enabled;
		private static ViewerGLJPanel viewer;
		
		private RenderOption(boolean enabled){
			this.enabled = enabled;
		}
		
		public void setEnabled(boolean enabled){
			this.enabled = enabled;
			if (viewer!=null) viewer.reDraw();
		}
		
		public boolean isEnabled(){
			return enabled;
		}
		
		private static void setViewer(ViewerGLJPanel viewer){
			RenderOption.viewer = viewer;
		}
	}
	
	public enum AtomRenderType {
		ATOMS, ELEMENTS, GRAINS, LATTICE_ROTATION, CUSTOM
	};
	
	private static final long serialVersionUID = 1L;
	private static final float CORE_THICKNESS = 5f;
	
	private AtomData atomData;
	private RenderRange renderInterval;
	
	private FrameBufferObject fboLeft;
	private FrameBufferObject fboRight;
	private FrameBufferObject fboBackground;
	
	// Rotation-Zoom-Move-Variables
	private GLMatrix rotMatrix = new GLMatrix();
	private GLMatrix dragMatrix = new GLMatrix();
	private ArcBall arcBall;
	private float moveX = 0f, moveY = 0;
	private float zoom = 1f;
	private boolean mousePressed = false;
	private Point startDragPosition;

	//Caching one half of the bounding box
	private float[] halfbounds = new float[3];
	private float[] globalMaxBounds = new float[3];
	
	private boolean performPicking = false;
	private Point pickPosition;
	private int pickModifier;
	
	private boolean exportScreenshot = false;
	private boolean exportSequence = false;
	private String exportFilename;
	private String exportType ="";
	private Dimension sizeBeforeExport;
	
	// OpenGL
	private GLUT glut;

	private String[] legendLabels = new String[3];
	private boolean drawLegendThisFrame = false;
	
	private boolean reRenderTexture = true;
	//The index is the position in the list, therefore it is some kind of map
	private ArrayList<Pickable> pickMap = new ArrayList<Pickable>();
	private int pickingShaderUniform;
	
	private boolean[] ignoreTypes;
	private boolean[] ignoreBV;
	private boolean[] ignoreElement;
	private HashMap<Integer,Boolean> ignoreGrain = new HashMap<Integer, Boolean>();
	private HashMap<Integer,float[]> grainColorTable = new HashMap<Integer, float[]>();
	
	private AtomRenderType atomRenderType = AtomRenderType.ATOMS;
	//Sphere render data
	private int[] sphereVboIndices;
	private float defaultSphereSize = 1.5f;
	private float[] sphereSize = new float[]{1.5f};
	private float maxSphereSize;
	private int sphereVBOIndexCount = 0;
	//Cube render data
	private int[] cubeVboIndices;
	
	private ArrayList<Object> highLightObjects = new ArrayList<Object>();
	
	private int width, height;
	private float latticeRotationMin = -8f, latticeRotationMax = 8f;
	private int latticeRotationComponent = 0;
	
	private GLMatrix projectionMatrix = new GLMatrix();
	private GLMatrix modelViewMatrix = new GLMatrix();
	private TextRenderer tr = null;
	
	private FrameBufferObject sphereTexLQ = null;
	private FrameBufferObject sphereTexHQ = null;
	private boolean updateBillboards = true;
	private boolean updateSphereRenderData = true;
	private boolean updateAtoms = true;
	
	private RingBuffer<CellRenderBuffer> cellRenderRingBuffer = null;
	private AtomRenderData renderData;
	
	public double openGLVersion = 0;
	
	public ViewerGLJPanel(int width, int height) {
		super();
		
		addMouseListener(this);
		addMouseMotionListener(this);
		addMouseWheelListener(this);
		addGLEventListener(this);
		
		this.setPreferredSize(new Dimension(width, height));
		this.setMinimumSize(new Dimension(100, 100));
		
		this.setFocusable(true);
		
		RenderOption.setViewer(this);
		Configuration.Options.setViewerPanel(this);
	}
	
	@Override
	public void display(GLAutoDrawable arg0) {
		GL2 gl = arg0.getGL().getGL2();
		
		if (performPicking){
			Shader.PICKING_SHADER.enable(gl);
			boolean reRenderAfterPicking = performPicking(gl, pickPosition); 
			Shader.DEFAULT_SHADER.enable(gl);;
			performPicking = false;
			if (reRenderAfterPicking) reRenderTexture = true;
			else return;
		}
		
		if (reRenderTexture){
			renderTexture(gl, RenderOption.STEREO.isEnabled());
			reRenderTexture = false;
		}

		renderSceneFromTexture(gl);
		
		//Second export iteration -> export screen
		if (exportScreenshot){
			try {
				exportScreenshot(gl);
				exportScreenshot = false;
				exportFilename = null;
				//Reset panel size
				this.setSize(sizeBeforeExport);
				createSphereRenderData(gl, Configuration.Options.BILLBOARD_RENDERING.isEnabled(), 
						Configuration.Options.HQ_RENDERING.isEnabled());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		
		//First export iteration -> draw screen in new size
		if (exportFilename != null){
			//Switch to high detailed rendering
			createSphereRenderData(gl, Configuration.Options.BILLBOARD_RENDERING.isEnabled(), true);
			exportScreenshot = true;
			reRenderTexture = true;
			this.reDraw();
		}	
	}

	@Override
	public void init(GLAutoDrawable arg0) {
		GL2 gl = arg0.getGL().getGL2();
		glut = new GLUT();
		 
		String glVersion = gl.glGetString(GL.GL_VERSION);
		glVersion = glVersion.substring(0, 3);
		openGLVersion = Double.parseDouble(glVersion);
		
		if (openGLVersion < 2.0) {
			String s = String.format("OpenGL 2.0 or higher is required, your version is: "+openGLVersion);
			System.out.println(s);
			System.exit(0);
		}
		
		createSphereRenderData(gl, Configuration.Options.BILLBOARD_RENDERING.isEnabled(), 
				Configuration.Options.HQ_RENDERING.isEnabled());

		gl.glClearColor(0f, 0f, 0f, 0.0f);
		
        gl.glEnable(GL.GL_LINE_SMOOTH);
		gl.glEnable(GL2.GL_POINT_SMOOTH);
		gl.glHint(GL2.GL_POINT_SMOOTH_HINT, GL.GL_NICEST);
        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glCullFace(GL.GL_BACK);
        gl.glEnable(GL.GL_CULL_FACE);
        
		gl.glEnable(GL.GL_BLEND);
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
        gl.glDepthFunc(GL.GL_LESS);
        
        ShaderCollection.init(gl);
        if (openGLVersion >=3.3){
        	GL3 gl3 = arg0.getGL().getGL3();
        	ShaderCollection.initGL3(gl3);
        }
        	
        
        Shader.DEFAULT_SHADER.enable(gl);
        pickingShaderUniform = gl.glGetUniformLocation(Shader.PICKING_SHADER.getProgram(), "pickcolor");    
		
        if (arg0.getGL().isGL3()){
        	gl.glEnable(GL3.GL_DEPTH_CLAMP);
        } else if (gl.isExtensionAvailable("GL_NV_depth_clamp")){
			gl.glEnable(GL2.GL_DEPTH_CLAMP);
		}
		
		tr = new TextRenderer(new Font(Font.SANS_SERIF, Font.PLAIN, 36));
		tr.setSmoothing(true);
		
		this.arcBall = new ArcBall(width, height);
		updateBillboards = true;
		updateSphereRenderData = true;

		
		if (openGLVersion >=3.3 && cellRenderRingBuffer == null){
			ArrayList<CellRenderBuffer> ringElements = new ArrayList<ViewerGLJPanel.CellRenderBuffer>();
			for (int i = 0; i < 8; i++) {
				CellRenderBuffer crb = new CellRenderBuffer();
				crb.init(gl, AtomRenderData.MAX_ATOMS_PER_CELL * 8 * Float.SIZE / 8, GL.GL_ARRAY_BUFFER);
				ringElements.add(crb);
			}			
			cellRenderRingBuffer = new RingBuffer<ViewerGLJPanel.CellRenderBuffer>(ringElements);
		}
		
		if (cubeVboIndices != null)
			gl.glDeleteBuffers(2, cubeVboIndices, 0);
			
		int[] t = {0,1,5, 0,5,4, 1,3,7, 1,7,5, 3,2,7, 2,6,7, 2,0,6, 0,4,6, 4,5,7, 4,7,6, 2,3,0, 3,1,0};
		float[] v = {0f, 0f, 0f,   1f, 0f, 0f,   0f, 1f, 0f,   1f, 1f, 0f,
				     0f, 0f, 1f,   1f, 0f, 1f,   0f, 1f, 1f,   1f, 1f, 1f};
		cubeVboIndices = new int[2];
		gl.glGenBuffers(2, cubeVboIndices, 0);
		IntBuffer tb = Buffers.newDirectIntBuffer(t.length);
		tb.put(t);
		tb.rewind();
		gl.glBindBuffer(GL.GL_ARRAY_BUFFER, cubeVboIndices[0]);
		gl.glBufferData(GL.GL_ARRAY_BUFFER, tb.capacity()*Integer.SIZE/8, tb, GL.GL_STATIC_DRAW);
		gl.glUnmapBuffer(GL.GL_ARRAY_BUFFER);
		
		FloatBuffer vb = Buffers.newDirectFloatBuffer(v.length);
		vb.put(v);
		vb.rewind();
		gl.glBindBuffer(GL.GL_ARRAY_BUFFER, cubeVboIndices[1]);
		gl.glBufferData(GL.GL_ARRAY_BUFFER, vb.capacity()*Float.SIZE/8, vb, GL.GL_STATIC_DRAW);
		gl.glUnmapBuffer(GL.GL_ARRAY_BUFFER);
		
		
	}

	private void makeBillboards(GL2 gl) {
		if (updateBillboards){
			if (sphereTexLQ != null)
				sphereTexLQ.destroy(gl);
			sphereTexLQ = SphereGenerator.getSphereFBO(gl, 32);
			if (sphereTexHQ != null)
				sphereTexHQ.destroy(gl);
			sphereTexHQ = SphereGenerator.getSphereFBO(gl, 128);
			updateBillboards = false;
		}
	}
	
	public void updateBillboards(){
		this.updateSphereRenderData = true;
		this.updateBillboards = true;
	}

	@Override
	public void reshape(GLAutoDrawable arg0, int arg1, int arg2, int arg3, int arg4) {
		GL2 gl = arg0.getGL().getGL2();
		arg0.getGL().glViewport(arg1, arg2, arg3, arg4);
		reRenderTexture = true;
		this.width = arg3;
		this.height = arg4;
		
		if (fboLeft != null) fboLeft.destroy(gl);
		if (fboRight != null) fboRight.destroy(gl);
		if (fboBackground != null) fboBackground.destroy(gl);
		fboLeft = new FrameBufferObject(width, height, gl);
		fboRight = new FrameBufferObject(width, height, gl);
		fboBackground = new FrameBufferObject(width, height, gl);
		
		this.arcBall.setSize(this.width, this.height);
	}
	
	private void setupProjection(GL2 gl, boolean loadIdentity) {
		if (loadIdentity) projectionMatrix.loadIdentity();

		// Correct the screen-aspect
		float aspect = width / (float) height;
		if (aspect > 1) {
			projectionMatrix.createOrtho(-aspect, aspect, -1f, 1f, -100f, +100f);
		} else {
			projectionMatrix.createOrtho(-1f, 1f, (-1f / aspect), (1f / aspect), -100f, +100f);
		}
		
		projectionMatrix.scale(this.zoom, this.zoom, this.zoom);
		projectionMatrix.applyMatrix(gl, GL2.GL_PROJECTION);
	}
	
	/**
	 * 
	 * @param eye -1: left eye,1 right eye, 0 centered (for picking) 
	 */
	private void setupProjectionStereo(GL2 gl, int eye) {
		projectionMatrix.loadIdentity();
		
		float zNear = 1.2f;
        float zFar  = Math.max(10f, 10f*zoom);

        float viewPaneCorrection = 0.5f;
        float focus = 0.8f;
        float eyeDist = 0.04f;
        float left, right, bottom, top;
        
        //Create asymmetric viewing-frustrums
        if (width<height){
        	left   = - viewPaneCorrection;
            right  = viewPaneCorrection;
            bottom = - viewPaneCorrection / (width/(float)height);
            top    = viewPaneCorrection / (width/(float)height);
        } else {
        	left   = - viewPaneCorrection * (width/(float)height);
            right  = viewPaneCorrection * (width/(float)height);
            bottom = - viewPaneCorrection;
            top    = viewPaneCorrection;
        }
        
        if ( eye == 1 ) {
            left  += eyeDist * focus;
            right += eyeDist * focus;
        } else if (eye == -1) {
            left  -= eyeDist * focus;
            right -= eyeDist * focus;
        }

        projectionMatrix.createFrustum(left, right, bottom, top, zNear, zFar);

        //Eye Position
        projectionMatrix.translate(0f, 0f, -3.2f);
        if ( eye == 1 ) {
        	projectionMatrix.translate(+eyeDist, 0f, 0f);
        } else if (eye == -1) {
        	projectionMatrix.translate(-eyeDist, 0f, 0f);
        }
        
        projectionMatrix.scale(this.zoom, this.zoom, this.zoom);
        
        projectionMatrix.applyMatrix(gl, GL2.GL_PROJECTION);
	}

	private void setupProjectionFlat(GL2 gl) {
		projectionMatrix.loadIdentity();
		projectionMatrix.createOrtho(0f, width, 0f, height, -100f, +100f);
		
		projectionMatrix.applyMatrix(gl, GL2.GL_PROJECTION);
	}
	
	private void setupModelView(GL2 gl) {
		modelViewMatrix.loadIdentity();

		modelViewMatrix.translate(-moveX, -moveY, 0);
		modelViewMatrix.mult(rotMatrix);
		float scale = 1f / Math.max(globalMaxBounds[0], Math.max(globalMaxBounds[1], globalMaxBounds[2])) * 0.8f;
		modelViewMatrix.scale(scale, scale, scale);
		modelViewMatrix.translate(-halfbounds[0], -halfbounds[1], -halfbounds[2]);
		
		modelViewMatrix.applyMatrix(gl, GL2.GL_MODELVIEW);
	}
	
	private void renderTexture(GL2 gl, boolean stereo){
//		long time = System.nanoTime();
		
		if (updateBillboards)
			makeBillboards(gl.getGL2());
		
		//Render Background
		fboBackground.bind(gl, false);
		setupProjectionFlat(gl);
		modelViewMatrix.loadIdentity();
		modelViewMatrix.applyMatrix(gl, GL2.GL_MODELVIEW);
		
		gl.glClearColor(1f, 1f, 1f, 1f);
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
		gl.glBegin(GL2.GL_TRIANGLE_STRIP);
		if (!RenderOption.PRINTING_MODE.isEnabled()){
			gl.glColor3f(0.8f, 0.8f, 0.8f); gl.glVertex3f(0f, 0f, 0f);
			gl.glColor3f(0.8f, 0.8f, 0.8f); gl.glVertex3f(width, 0f, 0f);
			gl.glColor3f(0.2f, 0.2f, 0.2f); gl.glVertex3f(0f, height, 0f);
			gl.glColor3f(0.2f, 0.2f, 0.2f); gl.glVertex3f(width, height, 0f);
		} else {
			gl.glColor3f(0f, 0f, 0f);
			gl.glVertex3f(0f, 0f, 0f); gl.glVertex3f(width, 0f, 0f);
			gl.glVertex3f(0f, height, 0f); gl.glVertex3f(width, height, 0f);
		}
		gl.glEnd();
		fboBackground.unbind(gl);
		
		if (stereo) {
			fboLeft.bind(gl, false);
			renderScene(gl, false, -1); //left eye
			fboLeft.unbind(gl);
			
			fboRight.bind(gl, false);
			renderScene(gl, false, 1); //right eye
			fboRight.unbind(gl);
		} else {
			fboLeft.bind(gl, false);
			renderScene(gl, false, 0); 	//central view
			fboLeft.unbind(gl);
		}
//		gl.glFinish();
//		System.out.println(1000000000./(System.nanoTime()-time));
	}

	private void renderSceneFromTexture(GL2 gl) {
		if (GLContext.getCurrent() == null) {
			System.out.println("Current context is null");
			System.exit(0);
		}
		
		setupProjectionFlat(gl);
		gl.glMatrixMode(GL.GL_TEXTURE);
		gl.glLoadIdentity();
		gl.glMatrixMode(GL2.GL_MODELVIEW);
		gl.glLoadIdentity();
		
		gl.glClearColor(0f, 0f, 0f, 0f);
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

		gl.glEnable(GL.GL_TEXTURE_2D);
		gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST);
	    gl.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST);
		Shader.TEXTURE_RENDER_SHADER.enable(gl);
		
		gl.glActiveTexture(GL.GL_TEXTURE0);
		gl.glBindTexture(GL.GL_TEXTURE_2D, fboLeft.getTextureName());
		gl.glUniform1i(gl.glGetUniformLocation(Shader.TEXTURE_RENDER_SHADER.getProgram(), "left"), 0);
		gl.glUniform1i(gl.glGetUniformLocation(Shader.TEXTURE_RENDER_SHADER.getProgram(), "stereo"), 0);
		if (RenderOption.STEREO.isEnabled()){	//Bind second texture
			gl.glActiveTexture(GL.GL_TEXTURE1);
			gl.glBindTexture(GL.GL_TEXTURE_2D, fboRight.getTextureName());		    
		    gl.glUniform1i(gl.glGetUniformLocation(Shader.TEXTURE_RENDER_SHADER.getProgram(), "right"), 1);
		    gl.glUniform1i(gl.glGetUniformLocation(Shader.TEXTURE_RENDER_SHADER.getProgram(), "stereo"), 1);
		}
		//Bind background
		gl.glActiveTexture(GL.GL_TEXTURE2);
		gl.glBindTexture(GL.GL_TEXTURE_2D, fboBackground.getTextureName());
		gl.glUniform1i(gl.glGetUniformLocation(Shader.TEXTURE_RENDER_SHADER.getProgram(), "back"), 2);
	    
		gl.glBegin(GL.GL_TRIANGLE_STRIP);
		gl.glTexCoord2f(1f, 0f); gl.glVertex3f(width,0f,0f);
		gl.glTexCoord2f(1f, 1f); gl.glVertex3f(width,height,0f);
		gl.glTexCoord2f(0f, 0f); gl.glVertex3f(0f,0f,0f);
		gl.glTexCoord2f(0f, 1f); gl.glVertex3f(0f,height,0f);
		gl.glEnd();
		
		gl.glDisable(GL.GL_TEXTURE_2D);
		Shader.DEFAULT_SHADER.enable(gl);
	}
	
	private void renderScene(GL2 gl, boolean picking, int eye) {
		if (GLContext.getCurrent() == null) {
			System.out.println("Current context is null");
			System.exit(0);
		}

		//Update sphere render data if necessary
		createSphereRenderData(gl, Configuration.Options.BILLBOARD_RENDERING.isEnabled(), 
			Configuration.Options.HQ_RENDERING.isEnabled());
		
		// clear colour and depth buffers
		if (RenderOption.PRINTING_MODE.isEnabled() || picking) gl.glClearColor(1f, 1f, 1f, 1f); 
		else gl.glClearColor(0f, 0f, 0f, 0f);
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
		
		if (atomData == null) return;
		
		setupModelView(gl);
		if (!RenderOption.STEREO.isEnabled())
			if (RenderOption.PERSPECTIVE.isEnabled())
				setupProjectionStereo(gl, 0);
			else setupProjection(gl, true); 
		else setupProjectionStereo(gl, eye);
		
		draw(gl, picking);
	}

	private void draw(GL2 gl, boolean picking) {
		//Renderpass 1 for transparent objects
		if (!picking){
			gl.glDepthMask(false);
			if (RenderOption.INDENTER.isEnabled()) drawIndent(gl);
			if (RenderOption.STACKING_FAULT.isEnabled() && ImportStates.SKELETONIZE.isActive()) drawSurfaces(gl, picking, true);
			if (RenderOption.DDT.isEnabled() && ImportStates.DISLOCATION_DENSITY.isActive()) drawDDT(gl, picking);
			if (RenderOption.GRAINS.isEnabled() && ImportStates.POLY_MATERIAL.isActive()) drawGrain(gl, picking, true);
			gl.glDepthMask(true);
		}
		
		//Render solid objects
		if (RenderOption.DISLOCATIONS.isEnabled() && ImportStates.SKELETONIZE.isActive()) drawCores(gl, picking);
		
		drawAtoms(gl, picking);
		
		for (DataContainer dc : atomData.getAdditionalData()){
			dc.draw(this, gl, renderInterval, picking);
		}
		
		gl.glLineWidth(height*0.0015f<1f ?1f :height*0.0015f);
		//Simulation box
		if (RenderOption.BOUNDING_BOX.isEnabled() && !picking) {
			gl.glColor3f(0.7f,0.7f,0.7f);
			
			float[] b = atomData.getBounds();
			gl.glBegin(GL.GL_LINES);
				gl.glVertex3f(0f, 0f, 0f); 		gl.glVertex3f(0f, 0f, b[2]);
				gl.glVertex3f(0f, b[1], 0f); 	gl.glVertex3f(0f, b[1], b[2]);
				gl.glVertex3f(b[0], b[1], 0f); 	gl.glVertex3f(b[0], b[1], b[2]);
				gl.glVertex3f(b[0], 0f, 0f); 	gl.glVertex3f(b[0], 0f, b[2]);
				gl.glVertex3f(0f, 0f, b[2]); 	gl.glVertex3f(0f, b[1], b[2]);
				gl.glVertex3f(0f, 0f, b[2]); 	gl.glVertex3f(b[0], 0f, b[2]);
				gl.glVertex3f(b[0], 0f, b[2]); 	gl.glVertex3f(b[0], b[1], b[2]);
				gl.glVertex3f(0f, b[1], b[2]); 	gl.glVertex3f(b[0], b[1], b[2]);                                      
				gl.glVertex3f(0f, 0f, 0f); 		gl.glVertex3f(0f, b[1], 0f);
				gl.glVertex3f(0f, 0f, 0f); 		gl.glVertex3f(b[0], 0f, 0f);
				gl.glVertex3f(b[0], 0f, 0f); 	gl.glVertex3f(b[0], b[1], 0f);
				gl.glVertex3f(0f, b[1], 0f); 	gl.glVertex3f(b[0], b[1], 0f);
			gl.glEnd();
		}
		
		//Renderpass 2 for transparent objects
		if (RenderOption.STACKING_FAULT.isEnabled() && ImportStates.SKELETONIZE.isActive()) drawSurfaces(gl, picking, false);
		if (RenderOption.DDT.isEnabled() && ImportStates.DISLOCATION_DENSITY.isActive()) drawDDT(gl, picking);
		if (RenderOption.GRAINS.isEnabled() && ImportStates.POLY_MATERIAL.isActive()) drawGrain(gl, picking, false);
		
		if (!picking){
			if (RenderOption.INDENTER.isEnabled()) drawIndent(gl);
			if (RenderOption.COORDINATE_SYSTEM.isEnabled()) drawCoordinateSystem(gl);
			if (RenderOption.THOMPSON_TETRAEDER.isEnabled()) drawThompsonTetraeder(gl);
			if (drawLegendThisFrame && RenderOption.LEGEND.enabled){
				drawLegend(gl);
				drawLegendThisFrame = false;
			}
		}
	}
	
	private void drawIndent(GL2 gl){
		//Test indenter geometry, sphere or cylinder
		
		Object o = atomData.getFileMetaData("extpot_cylinder");
		if (o != null) {
			float[] indent = null;
			if (o instanceof float[]) indent = (float[])o;
			else return;
			Shader.DEFAULT_ADS_PPL_SHADER.enable(gl);
			gl.glColor4f(0f, 1f, 0.5f, 0.4f);
			GLMatrix mvm = modelViewMatrix.clone();
			mvm.translate(indent[1], indent[2], indent[3]);
			
			float length = 0f;
			if (indent[0] == 0f){
				length = halfbounds[0] * 2f;
				mvm.rotate(90f, 0f, -1f, 0f);
			} else if (indent[1] == 0f){
				length = halfbounds[1] * 2f;
				mvm.rotate(90f, -1f, 0f, 0f);
			} else if (indent[2] == 0f){
				length = halfbounds[2] * 2f;
			}
			
			mvm.applyMatrix(gl, GL2.GL_MODELVIEW);
			gl.glDisable(GL2.GL_CULL_FACE);
			glut.glutSolidCylinder(indent[4],length, 50, 3);
			gl.glEnable(GL2.GL_CULL_FACE);
			
			modelViewMatrix.applyMatrix(gl, GL2.GL_MODELVIEW);
			Shader.DEFAULT_SHADER.enable(gl);
		}
		o = atomData.getFileMetaData("extpot");
		if (o != null) {
			float[] indent = null;
			if (o instanceof float[]) indent = (float[])o;
			else return;
			Shader.DEFAULT_ADS_PPL_SHADER.enable(gl);
			gl.glColor4f(0f, 1f, 0.5f, 0.4f);
			GLMatrix mvm = modelViewMatrix.clone();
			mvm.translate(indent[1], indent[2], indent[3]);
			mvm.applyMatrix(gl, GL2.GL_MODELVIEW);
			glut.glutSolidSphere(indent[4], 30, 30);
			modelViewMatrix.applyMatrix(gl, GL2.GL_MODELVIEW);
			Shader.DEFAULT_SHADER.enable(gl);
		}
		o = atomData.getFileMetaData("wall");
		if (o != null) {
			float[] indent = null;
			if (o instanceof float[]) indent = (float[])o;
			else return;
			Shader.DEFAULT_PPL_SHADER.enable(gl);
			gl.glCullFace(GL.GL_FRONT);
			gl.glColor4f(0f, 1f, 0.5f, 1f);
			GLMatrix mvm = modelViewMatrix.clone();
			float hx = atomData.getBounds()[0]*0.5f;
			float hy = atomData.getBounds()[1]*0.5f;
			mvm.translate(hx, hy, indent[1]+2f*indent[2]);
			mvm.scale(2f*hx, 2*hy, -5*indent[2]);
			mvm.applyMatrix(gl, GL2.GL_MODELVIEW);
			glut.glutSolidCube(1f);
			modelViewMatrix.applyMatrix(gl, GL2.GL_MODELVIEW);
			gl.glCullFace(GL.GL_BACK);
			
//			Shader.DEFAULT_ADS_PPL_SHADER.enable(gl);
//			gl.glColor4f(0.5f, 0.5f, 0.5f, .2f);
//			gl.glBegin(GL.GL_TRIANGLE_STRIP);
//			for (int i=0; i<=30;i++){
//				float s = (float)Math.sin(i/30.*2.*Math.PI);
//				float c = (float)Math.cos(i/30.*2.*Math.PI);
//				gl.glNormal3f(s, c, 0f);
//				gl.glVertex3f(s*200f+hx, c*200f+hy, 0f);
//				gl.glVertex3f(s*200f+hx, c*200f+hy, indent[0]);
//			}
//			gl.glEnd();
//			gl.glNormal3f(atomData.getBounds()[0]*0.5f, 0f, 1f);
//			gl.glBegin(GL.GL_TRIANGLE_FAN);
//			for (int i=0; i<=30;i++){
//				float s = (float)Math.sin(i/30.*2.*Math.PI);
//				float c = (float)Math.cos(i/30.*2.*Math.PI);
//				gl.glVertex3f(s*200f+hx, c*200f+hy, 0f);
//			}
//			gl.glEnd();
//			gl.glNormal3f(atomData.getBounds()[0]*0.5f, 0f, -1f);
//			gl.glBegin(GL.GL_TRIANGLE_FAN);
//			for (int i=30; i>=0;i--){
//				float s = (float)Math.sin(i/30.*2.*Math.PI);
//				float c = (float)Math.cos(i/30.*2.*Math.PI);
//				gl.glVertex3f(s*200f+hx, c*200f+hy, indent[0]);
//			}
//			gl.glEnd();
			
			Shader.DEFAULT_SHADER.enable(gl);
		}
	}
	
	private void drawSurfaces(GL2 gl, boolean picking, boolean firstPass){
		if (firstPass && picking) return;
		Skeletonizer skel = atomData.getSkeletonizer();
		if (skel.getDislocations()==null) return;
		
		gl.glDisable(GL.GL_CULL_FACE);
		for (int i=0; i<skel.getPlanarDefects().size(); i++) {
			PlanarDefect s = skel.getPlanarDefects().get(i);		
			if (picking) getNextPicking(gl, s);
			if (Configuration.getCrystalStructure().hasMultipleStackingFaultTypes()){
				float[] c = ColorTable.getGLColor(s.getPlaneComposedOfType());
				if (highLightObjects.contains(s))
					gl.glColor4f(c[0], c[1], c[2], 0.35f);
				else gl.glColor4f(c[0], c[1], c[2], 0.15f);
			} else {
				if (highLightObjects.contains(s))
					gl.glColor4f(0.8f,0.0f,0.0f,0.15f);
				else gl.glColor4f(0.5f,0.5f,0.5f,0.15f);
			}
			
			gl.glBegin(GL.GL_TRIANGLES);
			for (int j = 0; j < s.getFaces().length; j+=3) {
				if (renderInterval.isInInterval(s.getFaces()[j]) && 
						isVectorInPBC(VectorOps.sub(s.getFaces()[j], s.getFaces()[j+1])) && 
						isVectorInPBC(VectorOps.sub(s.getFaces()[j], s.getFaces()[j+2])) && 
						isVectorInPBC(VectorOps.sub(s.getFaces()[j+1], s.getFaces()[j+2]))) {
					gl.glVertex3f(s.getFaces()[j].getX(), s.getFaces()[j].getY(), s.getFaces()[j].getZ());
					gl.glVertex3f(s.getFaces()[j+1].getX(), s.getFaces()[j+1].getY(), s.getFaces()[j+1].getZ());
					gl.glVertex3f(s.getFaces()[j+2].getX(), s.getFaces()[j+2].getY(), s.getFaces()[j+2].getZ());
				}
			}
			gl.glEnd();
		}
		gl.glEnable(GL.GL_CULL_FACE);
	}
	
	private void drawDDT(GL2 gl, boolean picking) {
		if (DislocationDensityTensorData.JDislocationDensityTensorControls.getInstance() == null) return;
		float[] labels = DislocationDensityTensorData.JDislocationDensityTensorControls.getInstance().renderDDT(gl, picking);
		if (labels != null && RenderOption.LEGEND.isEnabled()){
			drawLegendThisFrame(">10^"+Float.toString(labels[0])+"/m²", "", "<10^"+Float.toString(labels[1])+"/m²");
		}
	}
		
		
	private void drawCores(GL2 gl, boolean picking) {
		Skeletonizer skel = atomData.getSkeletonizer();
		if (skel.getDislocations() == null) return;
		
		if (!picking) {
			if (Configuration.Options.ADS_SHADING.isEnabled()) Shader.DEFAULT_ADS_PPL_SHADER.enable(gl);
			else Shader.DEFAULT_PPL_SHADER.enable(gl);
		}
		for (int i=0; i<skel.getDislocations().size(); i++) {
			Dislocation dis = skel.getDislocations().get(i);
			
			if (dis.getBurgersVectorInfo()!=null){
				gl.glColor3fv(dis.getBurgersVectorInfo().getBurgersVector().getType().getColor(),0);
			} else {
				gl.glColor3f(0f,0f,0f);
			}
			
			if (highLightObjects.contains(dis) && !picking) {
				ArrayList<Atom> atoms = new ArrayList<Atom>();
				ArrayList<float[]> colors = new ArrayList<float[]>();

				for (int  j=0; j<dis.getLine().length; j++) {
					SkeletonNode sn = dis.getLine()[j];
					for (Atom a : sn.getMappedAtoms()){
						if (j==0) colors.add(new float[]{0f, 1f, 0f, 1f});
						else if (j==dis.getLine().length-1) colors.add(new float[]{0f, 0f, 1f, 1f});
						else colors.add(new float[]{1f, 0f, 0f, 1f});
						
						atoms.add(a);
					}
				}
				
				AtomRenderData ard = new AtomRenderData(atoms, colors, maxSphereSize, false);
				if (openGLVersion >= 3.3)
					drawSpheres(gl.getGL3(), ard, false);
				else drawSpheres(gl, ard, false);
			}
			
			if (picking) this.getNextPicking(gl, dis);
			
			ArrayList<float[]> path = new ArrayList<float[]>();
			ArrayList<float[]> dir = new ArrayList<float[]>();
			
			for (int j = 0; j < dis.getLine().length; j++) {
				SkeletonNode c = dis.getLine()[j];
				if (renderInterval.isInInterval(c)) {
					path.add(new float[]{ c.getX(), c.getY(), c.getZ()});
					dir.add(dis.getDirectionAtNode(j));
				} else {
					if (path.size()>1) 
						TubeRenderer.drawTube(gl, path, dir, CORE_THICKNESS);
					path.clear();
					dir.clear();
				}
				//Check PBC and restart tube if necessary
				if (j<dis.getLine().length-1){
					SkeletonNode c1 = dis.getLine()[j+1];
					if (!isVectorInPBC(VectorOps.sub(c, c1))){
						if (path.size()>1)
							TubeRenderer.drawTube(gl, path, dir, CORE_THICKNESS);
						path.clear();
						dir.clear();
						path.add(new float[]{ c1.getX(), c1.getY(), c1.getZ()});
						dir.add(dis.getDirectionAtNode(j+1));
					}
				}
			}
			if (path.size()>1)
				TubeRenderer.drawTube(gl, path, dir, CORE_THICKNESS);	
			
			if (!picking && RenderOption.BURGERS_VECTORS_ON_CORES.isEnabled() && atomData.isRbvAvailable()){
				float[] f;
				if (dis.getBurgersVectorInfo().getBurgersVector().isFullyDefined()) {
					gl.glColor4f(0f,1f,0f,1f);
					if (!ImportStates.POLY_MATERIAL.isActive() || dis.getGrain() == null) 
						f = dis.getBurgersVectorInfo().getBurgersVector().getInXYZ(Configuration.getCrystalRotationTools());
					else f = dis.getBurgersVectorInfo().getBurgersVector().getInXYZ(dis.getGrain().getCystalRotationTools());
				} else {
					gl.glColor4f(1f,0f,0f,1f);
					f = dis.getBurgersVectorInfo().getAverageResultantBurgersVector();
				}
				
				f = VectorOps.multiply(f, CORE_THICKNESS);
				SkeletonNode c = dis.getLine()[dis.getLine().length/2]; //node "in the middle"
				if (renderInterval.isInInterval(c)) {
					ArrowRenderer.renderArrow(gl, c, f, 0.4f);
				}
			}
		}
		if (!picking) Shader.DEFAULT_SHADER.enable(gl);
	}
	
	private void drawAtoms(GL2 gl, boolean picking){
		boolean allDeactivate = true;
		for (int i=0; i < Configuration.getCrystalStructure().getNumberOfTypes(); ++i){
			if (!this.isTypeIgnored(i)) allDeactivate = false;
			if (!this.isBVIgnored(i)) allDeactivate = false;
		}
		if (allDeactivate || atomData.getAtoms().size() == 0) return;
		
		ArrayList<Atom> atomsToRender = new ArrayList<Atom>();
		ArrayList<float[]> colorsOfRenderAtoms = new ArrayList<float[]>();
		
		CrystalStructure cs = Configuration.getCrystalStructure();
		
		if (atomRenderType == AtomRenderType.ATOMS && updateAtoms) {
			for (int i = 0; i < atomData.getAtoms().size(); i++) {
				Atom c = atomData.getAtoms().get(i);
				if (!this.isTypeIgnored(c.getType()) && renderInterval.isInInterval(c)) {
					colorsOfRenderAtoms.add(cs.getGLColor(c.getType())); 
					atomsToRender.add(c);
				}
			}
		} else if (atomRenderType == AtomRenderType.LATTICE_ROTATION) {
			if (updateAtoms)
				for (int i = 0; i < atomData.getAtoms().size(); i++) {
					Atom c = atomData.getAtoms().get(i);
					if (!this.isTypeIgnored(c.getType()) && renderInterval.isInInterval(c)) {
						float r;
						if (latticeRotationComponent == 3) r = c.getLatticeTilt();
						else r = c.getAxisRotation(latticeRotationComponent);
						if (!Configuration.isFilterRange() || 
								( (r >= latticeRotationMin && r<=latticeRotationMax) ^ Configuration.isFilterInversed())){
							colorsOfRenderAtoms.add(ColorTable.getIntensityGLColor(latticeRotationMin, latticeRotationMax, r));
							atomsToRender.add(c);
						}
					}
				}
			if (RenderOption.LEGEND.isEnabled()){
				drawLegendThisFrame(
						Float.toString(latticeRotationMin)+"°", 
						Float.toString((latticeRotationMin+latticeRotationMax)*0.5f)+"°",
						Float.toString(latticeRotationMax)+"°");
			}
		} else if (atomRenderType == AtomRenderType.ELEMENTS && updateAtoms) {
			float[][] colorTable = ColorTable.createColorTable(Configuration.getNumElements());
			
			for (int i = 0; i < atomData.getAtoms().size(); i++) {
				Atom c = atomData.getAtoms().get(i);
				if (!this.isElementIgnored(c.getElement()) && !this.isTypeIgnored(c.getType())
						&& renderInterval.isInInterval(c)) {
					colorsOfRenderAtoms.add(colorTable[c.getElement()]);
					atomsToRender.add(c);
				}
			}
		} else if (atomRenderType == AtomRenderType.GRAINS && updateAtoms) {
			for (int i = 0; i < atomData.getAtoms().size(); i++) {
				Atom c = atomData.getAtoms().get(i);
				if (!this.isElementIgnored(c.getElement()) && !this.isTypeIgnored(c.getType())
						&& !this.isGrainIgnored(c.getGrain()) && renderInterval.isInInterval(c)) {
					colorsOfRenderAtoms.add(this.grainColorTable.get(c.getGrain()));
					atomsToRender.add(c);
				}
			}
		} else if (atomRenderType == AtomRenderType.CUSTOM) {
			CustomColumnInfo customInfo = Configuration.getSelectedColumn();
			int selected = customInfo.getColumn();
			
			float min = customInfo.getLowerLimit();
			float max = customInfo.getUpperLimit();
			boolean filter = Configuration.isFilterRange();
			boolean inversed = Configuration.isFilterInversed();
			
			//Set custom color scheme of the custom value if present
			ColorBarScheme scheme = ColorTable.getColorBarScheme();
			boolean swapped = ColorTable.isColorBarSwapped();
			if (customInfo.getScheme()!=null){
				ColorTable.setColorBarScheme(customInfo.getScheme());
				ColorTable.setColorBarSwapped(false);
			}
			
			if (updateAtoms)
				for (int i = 0; i < atomData.getAtoms().size(); i++) {
					Atom c = atomData.getAtoms().get(i);
					if (!this.isTypeIgnored(c.getType()) && !this.isElementIgnored(c.getElement()) && renderInterval.isInInterval(c)) {
						if (filter) {
							 if ( (c.getCustom(selected)>=min && c.getCustom(selected)<=max) ^ inversed){
								 colorsOfRenderAtoms.add(ColorTable.getIntensityGLColor(min, max, c.getCustom(selected)));
								 atomsToRender.add(c);
							 }
						}
						else {
							colorsOfRenderAtoms.add(ColorTable.getIntensityGLColor(min, max, c.getCustom(selected)));
							atomsToRender.add(c);
						}
					}
				}
			
			if (customInfo.getScheme()!=null){
				ColorTable.setColorBarScheme(scheme);
				ColorTable.setColorBarSwapped(swapped);
			} else if (RenderOption.LEGEND.isEnabled()){
				drawLegendThisFrame(
						Float.toString(min)+" "+customInfo.getUnit(), 
						Float.toString((max+min)*0.5f)+" "+customInfo.getUnit(),
						Float.toString(max)+" "+customInfo.getUnit()
						);
			}
		}
		
		if (updateAtoms){
			renderData = new AtomRenderData(atomsToRender, colorsOfRenderAtoms, maxSphereSize, true);
			this.updateAtoms = false;
		}
		

		if (openGLVersion>=3.3)
			drawSpheres(gl.getGL3(), renderData, picking);
		else 
			drawSpheres(gl, renderData, picking);
			
		if (atomData.isRbvAvailable()){
			if (!picking) Shader.DEFAULT_PPL_SHADER.enable(gl);
			
			float lsVectorLength = cs.getLatticeConstant()/3f;
			float minLength = cs.getPerfectBurgersVectorLength() * 0.2f;
			minLength *= minLength; 
			float maxLength = cs.getPerfectBurgersVectorLength() * 2f;
			maxLength *= maxLength;
			for (int i=0; i<atomData.getAtoms().size(); i++){
				Atom c = atomData.getAtoms().get(i);
				if (c.getRBV()!=null && !this.isBVIgnored(c.getType()) && renderInterval.isInInterval(atomData.getAtoms().get(i))){
					float l = c.getRBV().getSquaredRBVLengh();
					if (l<minLength || l>maxLength) continue; 
					
					if (picking) this.getNextPicking(gl, c);
					gl.glColor3fv(cs.getGLColor(c.getType()), 0);

					ArrowRenderer.renderArrow(gl, c,
							new float[]{c.getRBV().rbv_x, c.getRBV().rbv_y, c.getRBV().rbv_z}, 0.1f);
					gl.glColor4f(0.5f,0.5f,0.5f,1f);
					ArrowRenderer.renderArrow(gl, c,
							new float[]{c.getRBV().lineDirection_x*lsVectorLength, 
							            c.getRBV().lineDirection_y*lsVectorLength,
							            c.getRBV().lineDirection_z*lsVectorLength}, 0.05f);
				}
			}
		}
		if (!picking) Shader.DEFAULT_SHADER.enable(gl);
	}
	
	
	private void drawGrain(GL2 gl, boolean picking, boolean firstPass){
		if (picking) firstPass = true; //A picking operation is equal to the first pass rendering
		gl.glLineWidth(height*0.001f<1f ?1f :height*0.0010f);
		if (!picking) Shader.DEFAULT_ADS_PPL_SHADER.enable(gl);
		else Shader.PICKING_SHADER.enable(gl);
		
		for (Grain grain : atomData.getGrains()){
//			if (grain.getMesh().getVolume()<1000.) continue;
			
			if (isGrainIgnored(grain.getGrainNumber())) continue;
			if (picking) this.getNextPicking(gl, grain);
			
			Mesh mesh = grain.getMesh();
			
			gl.glDepthMask(!firstPass);
			gl.glBlendFunc(GL.GL_ONE_MINUS_SRC_ALPHA, GL.GL_SRC_ALPHA);	
			
			gl.glColor4fv(grainColorTable.get(grain.getGrainNumber()), 0);
			mesh.getFinalMesh().renderMesh(gl);
			
			gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
			gl.glDepthMask(true);
			
//			if ( !firstPass && !picking){
//				gl.glColor4f(0.0f, 0.0f, 0.0f, 0.5f);
//				
//				gl.glEnable(GL2.GL_POLYGON_OFFSET_LINE);
//				gl.glPolygonOffset(0f, -500f);
//				gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL2.GL_LINE);
//			
//				mesh.getFinalMesh().renderMesh(gl);
//			
//				gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL2.GL_FILL);
//				gl.glPolygonOffset(0f, 0f);
//				gl.glDisable(GL2.GL_POLYGON_OFFSET_LINE);
//			}
		}
		if (!picking) Shader.DEFAULT_SHADER.enable(gl);
	}
	
	private void drawThompsonTetraeder(GL2 gl){
		gl.glLineWidth(height*0.0015f<2? 2f :height*0.0015f);
		gl.glDisable(GL.GL_DEPTH_TEST);
		
		GLMatrix pm = projectionMatrix.clone();
		pm.scale(1f/zoom, 1f/zoom, 1f/zoom);
		pm.applyMatrix(gl, GL2.GL_PROJECTION);
		
		GLMatrix mvm = new GLMatrix();
		
		float aspect = width / (float) height;
		if (aspect > 1) {
			mvm.translate(-aspect*0.8f, -0.8f, 0.8f);
		} else {
			mvm.translate(-0.8f, -(1/aspect)*0.8f, 0.8f);
		}

		mvm.scale(0.2f, 0.2f, 0.2f);
		mvm.mult(rotMatrix);
		mvm.applyMatrix(gl, GL2.GL_MODELVIEW);
		
		float[][] a = Configuration.getCrystalRotationTools().getThompsonTetraeder();
		float[] alpha = VectorOps.multiply(VectorOps.add(VectorOps.add(a[1], a[2]), a[3]), 0.333f);
		float[] beta  = VectorOps.multiply(VectorOps.add(VectorOps.add(a[0], a[2]), a[3]), 0.333f);
		float[] gamma = VectorOps.multiply(VectorOps.add(VectorOps.add(a[0], a[1]), a[3]), 0.333f);
		float[] delta = VectorOps.multiply(VectorOps.add(VectorOps.add(a[0], a[1]), a[2]), 0.333f);		
		
		gl.glBegin(GL.GL_TRIANGLES);
			gl.glColor4f(1f,0f,0f,1f);
			gl.glVertex3fv(a[3],0); gl.glVertex3fv(a[2],0); gl.glVertex3fv(a[1],0);
			gl.glColor4f(0f,0f,1f,1f);
			gl.glVertex3fv(a[0],0); gl.glVertex3fv(a[2],0); gl.glVertex3fv(a[3],0);
			gl.glColor4f(1f,1f,0f,1f);
			gl.glVertex3fv(a[1],0); gl.glVertex3fv(a[0],0); gl.glVertex3fv(a[3],0);
			gl.glColor4f(0f,1f,0f,1f);
			gl.glVertex3fv(a[0],0); gl.glVertex3fv(a[1],0); gl.glVertex3fv(a[2],0);
		gl.glEnd();
		
		gl.glColor3f(0.5f,0.5f,0.5f);
		gl.glEnable(GL2.GL_POLYGON_OFFSET_LINE);
		gl.glPolygonOffset(0f, -500f);
		gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL2.GL_LINE);
		gl.glBegin(GL.GL_TRIANGLES);
			gl.glVertex3fv(alpha,0); gl.glVertex3fv(a[2],0); gl.glVertex3fv(a[1],0);
			gl.glVertex3fv(alpha,0); gl.glVertex3fv(a[1],0); gl.glVertex3fv(a[3],0);
			gl.glVertex3fv(alpha,0); gl.glVertex3fv(a[3],0); gl.glVertex3fv(a[2],0);
			gl.glVertex3fv(beta ,0); gl.glVertex3fv(a[3],0); gl.glVertex3fv(a[0],0);
			gl.glVertex3fv(beta ,0); gl.glVertex3fv(a[0],0); gl.glVertex3fv(a[2],0);
			gl.glVertex3fv(beta ,0); gl.glVertex3fv(a[2],0); gl.glVertex3fv(a[3],0);
			gl.glVertex3fv(gamma,0); gl.glVertex3fv(a[0],0); gl.glVertex3fv(a[3],0);
			gl.glVertex3fv(gamma,0); gl.glVertex3fv(a[3],0); gl.glVertex3fv(a[1],0);
			gl.glVertex3fv(gamma,0); gl.glVertex3fv(a[1],0); gl.glVertex3fv(a[0],0);
			gl.glVertex3fv(delta,0); gl.glVertex3fv(a[1],0); gl.glVertex3fv(a[2],0);
			gl.glVertex3fv(delta,0); gl.glVertex3fv(a[2],0); gl.glVertex3fv(a[0],0);
			gl.glVertex3fv(delta,0); gl.glVertex3fv(a[0],0); gl.glVertex3fv(a[1],0);
		gl.glEnd();
		gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL2.GL_FILL);
		gl.glPolygonOffset(0f, 0f);
		gl.glDisable(GL2.GL_POLYGON_OFFSET_LINE);
		
		modelViewMatrix.applyMatrix(gl, GL2.GL_MODELVIEW);
		projectionMatrix.applyMatrix(gl, GL2.GL_PROJECTION);
		gl.glEnable(GL.GL_DEPTH_TEST);
	}
	
	private void drawCoordinateSystem(GL2 gl) {
		gl.glDisable(GL.GL_DEPTH_TEST);
		
		GLMatrix pm = projectionMatrix.clone();
		pm.scale(1f/zoom, 1f/zoom, 1f/zoom);
		pm.applyMatrix(gl, GL2.GL_PROJECTION);
		
		GLMatrix mvm = new GLMatrix();
		
		float aspect = width / (float) height;
		if (aspect > 1) {
			mvm.translate(aspect*0.5f, -0.8f, 1f);
		} else {
			mvm.translate(0.5f, -(1/aspect)*0.8f, 1f);
		}
		
		mvm.scale(0.40f, 0.40f, 0.40f);
		mvm.mult(rotMatrix);
		mvm.applyMatrix(gl, GL2.GL_MODELVIEW);
		
		gl.glColor3f(0.5f,0.5f,0.5f);
		
		//Coordinate lines
		Shader.DEFAULT_PPL_SHADER.enable(gl);
		ArrowRenderer.renderArrow(gl, new TrivialCoordinate(0f, 0f, 0f), new float[]{1f, 0f, 0f}, 0.02f);
		ArrowRenderer.renderArrow(gl, new TrivialCoordinate(0f, 0f, 0f), new float[]{0f, 1f, 0f}, 0.02f);
		ArrowRenderer.renderArrow(gl, new TrivialCoordinate(0f, 0f, 0f), new float[]{0f, 0f, 1f}, 0.02f);
		
		gl.glColor4f(1f, 1f, 1f, 1f);
		tr.setColor(0f, 0f, 0f, 1f);
		
		Shader.DEFAULT_TEXTURE_SHADER.enable(gl);
		gl.glActiveTexture(GL.GL_TEXTURE0);
		//Labels
		for (int i=0; i<3; i++){
			String s = "["+(int)Configuration.getCrystalOrientation()[i][0]+" "
				+(int)Configuration.getCrystalOrientation()[i][1]+" "
				+(int)Configuration.getCrystalOrientation()[i][2] +"]";
			if (i==0) s = "x="+s;
			if (i==1) s = "y="+s;
			if (i==2) s = "z="+s;
			
			GLMatrix mvm_new = mvm.clone();
			mvm_new.translate(i==0 ? 1f : 0f, i==1 ? 1f : 0f, i==2 ? 1f : 0f);
			GLMatrix rotInverse = new GLMatrix(rotMatrix.getMatrix());
			rotInverse.inverse();
			mvm_new.mult(rotInverse);
			mvm_new.translate(-0.5f, 0.1f, 0f);
			mvm_new.applyMatrix(gl, GL2.GL_MODELVIEW);
			tr.begin3DRendering();
	        tr.draw3D(s, 0f, 0f, 0f, 0.005f);
	        tr.end3DRendering();
		}
		
		Shader.DEFAULT_SHADER.enable(gl);
		modelViewMatrix.applyMatrix(gl, GL2.GL_MODELVIEW);
		projectionMatrix.applyMatrix(gl, GL2.GL_PROJECTION);
		
		gl.glEnable(GL.GL_DEPTH_TEST);
	}
	
	private void drawLegend(GL2 gl) {
		Shader.DEFAULT_SHADER.enable(gl);
		setupProjectionFlat(gl);
		GLMatrix mvm = new GLMatrix();
		mvm.loadIdentity();
		mvm.applyMatrix(gl, GL2.GL_MODELVIEW);
		
		float[][] scale = ColorTable.getColorBarScheme().getColorBar();
		
		//Z-coordinate is always in front of everything else
        gl.glBegin(GL.GL_TRIANGLE_STRIP);
    	for (int i=0; i<scale.length; i++){
    		if (!ColorTable.isColorBarSwapped())
    			gl.glColor3f(scale[i][0],scale[i][1],scale[i][2]);
    		else gl.glColor3f(scale[scale.length-1-i][0],scale[scale.length-1-i][1],scale[scale.length-1-i][2]);
    		float h = height*(0.015f + 0.24f*(i/(float)(scale.length-1)));
    		gl.glVertex3f(width*0.005f,                h, zoom*1.01f); 
    		gl.glVertex3f(width*0.005f + height*0.06f, h, zoom*1.01f);
    		
    	}
        gl.glEnd();
        
        gl.glColor4f(1f, 1f, 1f, 1f);
		tr.setColor(0f, 0f, 0f, 1f);
		gl.glActiveTexture(GL.GL_TEXTURE0);
		Shader.DEFAULT_TEXTURE_SHADER.enable(gl);
        tr.begin3DRendering();
        tr.draw3D(this.legendLabels[0], width*0.01f+height*0.06f, height*0.01f, zoom*1.01f, 0.0007f*height);
        tr.draw3D(this.legendLabels[1], width*0.01f+height*0.06f, height*0.13f, zoom*1.01f, 0.0007f*height);
        tr.draw3D(this.legendLabels[2], width*0.01f+height*0.06f, height*0.25f, zoom*1.01f, 0.0007f*height);
        tr.end3DRendering();
        Shader.DEFAULT_SHADER.enable(gl);
        modelViewMatrix.applyMatrix(gl, GL2.GL_MODELVIEW);
	}
	
	public void drawLegendThisFrame(String lowerLabel, String middleLabel, String upperLabel){
		this.drawLegendThisFrame = true;
		this.legendLabels[0] = lowerLabel;
		this.legendLabels[1] = middleLabel;
		this.legendLabels[2] = upperLabel;
	}
	
	public void reDraw(){
		reRenderTexture = true;
		this.repaint();
	}
	
	public void updateAtoms(){
		this.updateAtoms = true;
		this.reDraw();
	}
	
	// region MouseListener
	public void mouseClicked(MouseEvent arg0) {
		this.pickPosition = arg0.getPoint();
		this.pickModifier = arg0.getModifiersEx();
		this.performPicking = true;
		this.reDraw();
	}

	public void mouseEntered(MouseEvent arg0) {}

	public void mouseExited(MouseEvent arg0) {}

	public void mousePressed(MouseEvent arg0) {
		this.requestFocusInWindow();
		if (!mousePressed) {
			this.dragMatrix = new GLMatrix();
			this.dragMatrix.mult(rotMatrix);
			arcBall.setRotationStartingPoint(arg0.getPoint());
		}
		mousePressed = true;
		startDragPosition = arg0.getPoint();
	}

	public void mouseReleased(MouseEvent arg0) {
		mousePressed = false;
	}

	// endregion MouseListener

	// region MouseMotionListener
	@Override
	public void mouseDragged(MouseEvent e) {
		if (mousePressed) {
			Point newDragPosition = e.getPoint();
			if ((e.getModifiersEx() & InputEvent.ALT_DOWN_MASK) == InputEvent.ALT_DOWN_MASK){
				zoom *= 1 + (newDragPosition.y - startDragPosition.y) / 50f;
				if (zoom>50f) zoom = 50f;
				if (zoom<0.05f) zoom = 0.05f;
			}
			else if ((e.getModifiersEx() & InputEvent.CTRL_DOWN_MASK) == InputEvent.CTRL_DOWN_MASK){
				moveX -= (newDragPosition.x - startDragPosition.x) * 0.01f;
				moveY += (newDragPosition.y - startDragPosition.y) * 0.01f;
			} else {
				if ((e.getModifiersEx() & InputEvent.SHIFT_DOWN_MASK) == InputEvent.SHIFT_DOWN_MASK){
					GLMatrix mat = new GLMatrix();
					mat.rotate(newDragPosition.x - startDragPosition.x, 0, 1, 0);
					mat.mult(rotMatrix);
					rotMatrix.loadIdentity();
					rotMatrix.mult(mat);
				} else {
					//ArcBall update
					float[] quat = arcBall.drag(newDragPosition);
					if (quat != null) {
						rotMatrix.loadIdentity();
					    rotMatrix.setRotationFromQuaternion(quat);   
					    rotMatrix.mult(dragMatrix);
					}
				}
			}
			startDragPosition = newDragPosition;
			this.reDraw();
		}
	}
	
	public void mouseMoved(MouseEvent arg0) {}

	// endregion MouseMotionListener

	// region MouseWheelListener
	public void mouseWheelMoved(MouseWheelEvent arg0) {
		zoom *= 1 + arg0.getWheelRotation() / 20f;
		if (zoom>50f) zoom = 50f;
		if (zoom<0.05f) zoom = 0.05f;
		this.reDraw();
	}

	// endregion MouseWheelListener
	
	public void setAtomData(AtomData ada, boolean newSet){
		this.atomData = ada;
		this.halfbounds[0] = ada.getBounds()[0]*0.5f;
		this.halfbounds[1] = ada.getBounds()[1]*0.5f;
		this.halfbounds[2] = ada.getBounds()[2]*0.5f;
		if (newSet){
			ignoreTypes = new boolean[Configuration.getCrystalStructure().getNumberOfTypes()];
			ignoreBV = new boolean[Configuration.getCrystalStructure().getNumberOfTypes()];
			ignoreElement = new boolean[Configuration.getNumElements()];
			setSphereSize(Configuration.getCrystalStructure().getLatticeConstant()*0.4f);
			if (!ImportStates.POLY_MATERIAL.isActive()) RenderOption.GRAINS.enabled = false;
			if (!ImportStates.SKELETONIZE.isActive()) RenderOption.DISLOCATIONS.enabled = false;
			if (!ImportStates.SKELETONIZE.isActive()) RenderOption.STACKING_FAULT.enabled = false;
			if (!ImportStates.DISLOCATION_DENSITY.isActive()) RenderOption.DDT.enabled = false;
			
			//Find global maximum boundaries
			globalMaxBounds = new float[3];
			AtomData tmp = ada;
			//Iterate over the whole set
			while (tmp.getPrevious()!=null) tmp = tmp.getPrevious();
			do {
				float[] bounds = tmp.getBounds();
				if (bounds[0]>globalMaxBounds[0]) globalMaxBounds[0] = bounds[0];
				if (bounds[1]>globalMaxBounds[1]) globalMaxBounds[1] = bounds[1];
				if (bounds[2]>globalMaxBounds[2]) globalMaxBounds[2] = bounds[2];
				tmp = tmp.getNext();
			} while (tmp != null);
			
			ignoreGrain.clear();
			if (Configuration.getGrainIndices().size()!=0){
				for (int g : Configuration.getGrainIndices()){
					ignoreGrain.put(g, false);
				}
				ignoreGrain.put(Atom.IGNORED_GRAIN, false);
				ignoreGrain.put(Atom.DEFAULT_GRAIN, false);
			}
			
			grainColorTable.clear();
			

			ArrayList<Integer> sortGrainIndices = new ArrayList<Integer>(Configuration.getGrainIndices());
			Collections.sort(sortGrainIndices);
			float[][] colors = ColorTable.createColorTable(sortGrainIndices.size()+2, 0.3f);
			int j=0;
			for (int i: sortGrainIndices){
				grainColorTable.put(i, colors[j]);
				j++;
			}
		}
		
		if (this.renderInterval==null)
			this.renderInterval = new RenderRange(ada.getBounds());
		else {
			renderInterval.setGlobalLimit(3, this.atomData.getBounds()[0]);
			renderInterval.setGlobalLimit(4, this.atomData.getBounds()[1]);
			renderInterval.setGlobalLimit(5, this.atomData.getBounds()[2]);
			if (newSet || renderInterval.isNoLimiting()) renderInterval.reset();		
		}
		this.updateAtoms();
	}
	
	public RenderRange getRenderRange(){
		return renderInterval;
	}
	
	public void setAtomRenderMethod(AtomRenderType value){
		atomRenderType = value;
		this.updateAtoms();
	}

	public float getSphereSize() {
		return defaultSphereSize;
	}

	public void setSphereSize(float sphereSize) {
		this.defaultSphereSize = sphereSize;
		this.sphereSize = Configuration.getCrystalStructure().getSphereSizeScalings();
		this.maxSphereSize = 0f;
		for (int i=0; i<this.sphereSize.length; i++){
			this.sphereSize[i] *= defaultSphereSize;
			if (this.sphereSize[i]>maxSphereSize)
				maxSphereSize = this.sphereSize[i];
		}
		this.reDraw();
	}
	
	public boolean isTypeIgnored(int i){
		if (i>Configuration.getCrystalStructure().getNumberOfTypes()) return true;
		return ignoreTypes[i];
	}
	
	public boolean isElementIgnored(int i){
		if (i>ignoreElement.length) return true;
		return ignoreElement[i];
	}
	
	public boolean isBVIgnored(int i){
		if (i>Configuration.getCrystalStructure().getNumberOfTypes()) return true;
		return ignoreBV[i];
	}
	
	public boolean isGrainIgnored(int i){
		if (ignoreGrain.containsKey(i))
			return ignoreGrain.get(i);
		else return true;
	}
	
	public float[] getGrainColor(int i){
		if (ignoreGrain.containsKey(i))
			return grainColorTable.get(i);
		else return new float[]{0f, 0f, 0f};
	}
	
	public void setTypeIgnored(int i, boolean ignore){
		if (i>Configuration.getCrystalStructure().getNumberOfTypes()) return;
		ignoreTypes[i] = ignore;
	}
	
	public void setElementIgnored(int i, boolean ignore){
		if (i>ignoreElement.length) return;
		ignoreElement[i] = ignore;
	}
	
	public void setBVIgnored(int i, boolean ignore){
		if (i>Configuration.getCrystalStructure().getNumberOfTypes()) return;
		ignoreBV[i] = ignore;
	}
	
	public void setGrainIgnored(int i, boolean ignore){
		ignoreGrain.put(i, ignore);
	}
	
	public void resetZoom(){
		zoom = 1f;
		this.reDraw();
	}
	
	public void setLatticeRotationMinMax(float latticeRotationMin, float latticeRotationMax) {
		this.latticeRotationMin = latticeRotationMin;
		this.latticeRotationMax = latticeRotationMax;
		if (atomRenderType == AtomRenderType.LATTICE_ROTATION) this.updateAtoms();
	}
	
	public void setLatticeRotationComponent(int latticeRotationComponent) {
		this.latticeRotationComponent = latticeRotationComponent;
		if (atomRenderType == AtomRenderType.LATTICE_ROTATION) this.updateAtoms();
	}
	
	public int getLatticeRotationComponent() {
		return latticeRotationComponent;
	}
	
	public float getLatticeRotationMax() {
		return latticeRotationMax;
	}
	
	public float getLatticeRotationMin() {
		return latticeRotationMin;
	}
	
	/**
	 * Sets the selected point of view, resets model shift, but leaves the zoom unchanged
	 * @param rotX
	 * @param rotY
	 * @param rotZ
	 */
	public void setPOV(float rotX, float rotY, float rotZ){
		this.rotMatrix.loadIdentity();
		this.rotMatrix.rotate(rotX, 1.0f, 0.0f, 0.0f);
		this.rotMatrix.rotate(rotY, 0.0f, 1.0f, 0.0f);
		this.rotMatrix.rotate(rotZ, 0.0f, 0.0f, 1.0f);
		this.moveX = 0f; this.moveY = 0;
		
		this.reDraw();
	}
	
	public void setPOV(float[] m){
		if (m == null || m.length!=19) return; //0-15 rot, 16 zoom, 17-18 moveXY
		this.rotMatrix.loadIdentity();
		this.rotMatrix.mult(m);
		this.moveX = m[17]; this.moveY = m[18];
		this.zoom = m[16];
		this.reDraw();
	}
	
	public float[] getPov(){
		float[] m = new float[19];
		rotMatrix.getMatrix().get(m, 0, 16);
		rotMatrix.getMatrix().rewind();
		m[16] = zoom;
		m[17] = moveX;
		m[18] = moveY;
		return m;
	}
	
	/**
	 * 
	 * @param gl
	 * @param p
	 * @return true if repaint is required
	 */
	private boolean performPicking(GL2 gl, Point p){
		if (atomData==null) return false;
		
		//Extract viewport
		int[] viewport = new int[4];
		gl.glGetIntegerv(GL.GL_VIEWPORT, viewport, 0);
		
		
		//Save current Framebuffer before rendering in picking mode in
		//temporary framebuffer
		FrameBufferObject tmp_fboLeft = fboLeft;
		FrameBufferObject tmp_fboRight = fboRight;
		FrameBufferObject tmp_fboBack = fboBackground;
		fboLeft = new FrameBufferObject(width, height, gl);
		fboRight = new FrameBufferObject(width, height, gl);
		fboBackground = new FrameBufferObject(width, height, gl);
		
		//Render in Picking mode
		gl.glDisable(GL.GL_BLEND);
		gl.glEnable(GL.GL_SCISSOR_TEST);
		gl.glScissor(p.x-3, viewport[3] - p.y-3, 7, 7);
		renderScene(gl, true ,0);
		gl.glDisable(GL.GL_SCISSOR_TEST);
		gl.glEnable(GL.GL_BLEND);
		
		if (pickMap.size() > 16777214){
			JLogPanel.getJLogPanel().addLog("Too many objects in the scene, show less objects to enable picking");
			pickMap.clear();
			return false;
		}
		
		FloatBuffer selectBuffer = Buffers.newDirectFloatBuffer(3*49);
		gl.glReadPixels(p.x-3, viewport[3] - p.y-3, 7, 7, GL.GL_RGB, GL.GL_FLOAT, selectBuffer);
		
		//Delete temporary framebuffer and restore the standard ones
		if (fboLeft != null) fboLeft.destroy(gl);
		if (fboRight != null) fboRight.destroy(gl);
		if (fboBackground != null) fboBackground.destroy(gl);
		fboLeft = tmp_fboLeft;
		fboRight = tmp_fboRight;
		fboBackground = tmp_fboBack;
		
		float[] selectBuf = new float[3*49];
		selectBuffer.get(selectBuf);
		
		int hits = 0;
		
		//Identify unique objects
		TreeSet<Integer> hitMap = new TreeSet<Integer>();
		for (int i=0; i<selectBuf.length/3; i++){
			int num = (int)(selectBuf[i*3]*255f)<<16 | ((int)(selectBuf[i*3+1]*255f))<<8 | ((int)(selectBuf[i*3+2]*255f));
			hitMap.add(num);
		}
		
		boolean repaintRequired = false;
		boolean cleared = false;
		
		//Process hits
		for (Integer num : hitMap){			
			if (num == 0xffffff) continue; //Pure white > Background
			
			Pickable picked = pickMap.get(num);
			JLogPanel.getJLogPanel().addLog(picked.printMessage(pickModifier));
			
			if (picked.isHighlightable()){
				if (!cleared) {
					highLightObjects.clear();
					cleared = true;
				}
				repaintRequired = true;
				highLightObjects.add(picked);
			}
			if (picked.getHighlightedObjects() != null){
				if (!cleared) {
					highLightObjects.clear();
					cleared = true;
				}
				repaintRequired = true;
				highLightObjects.addAll(picked.getHighlightedObjects());
			}
			hits++;
		}
		if (hits == 0) {
			if (highLightObjects.size() > 0){
				highLightObjects.clear();
				repaintRequired = true;
			}
		}
		pickMap.clear();
		return repaintRequired;
	}
	
    private boolean isVectorInPBC(float[] v){
		return (!(Configuration.getPbc()[0] && Math.abs(v[0]) > halfbounds[0])
				&& !(Configuration.getPbc()[1] && Math.abs(v[1]) > halfbounds[1])
				&& !(Configuration.getPbc()[2] && Math.abs(v[2]) > halfbounds[2]));
	}
    
    //region export methods
	public void makeScreenshot(String filename, String type, boolean sequence, int w, int h){
		this.exportFilename = filename;
		this.exportType = type;
		this.exportSequence = sequence;
		//Save current panel size, then set to screenshot dimensions
		this.sizeBeforeExport = new Dimension(this.getWidth(),this.getHeight());
		this.setSize(w,h);
		this.reDraw();
	}
	
	private void exportScreenshot(GL2 gl) throws GLException, IOException{
		if (!exportSequence) {
			ImageIO.write( Screenshot.readToBufferedImage(width, height), exportType, new java.io.File(exportFilename) );
		} else {
			if (this.atomData!=null){
				AtomData currentAtomData = this.atomData;
				
				//rewind to beginning
				while (atomData.getPrevious()!=null)
					this.atomData = atomData.getPrevious();
				
				do {
					renderTexture(gl, RenderOption.STEREO.isEnabled());
					renderSceneFromTexture(gl);
					this.updateAtoms();
					ImageIO.write( Screenshot.readToBufferedImage(width, height), exportType, new java.io.File(exportFilename+atomData.getName()+"."+exportType) );
					this.atomData = this.atomData.getNext();
				} while (this.atomData != null);
				
				this.atomData = currentAtomData;
				this.updateAtoms();
			}
		}
	}
	
	//endregion export methods
	
	private void createSphereRenderData(GL gl, boolean billboards, boolean highQuality){
		int[] t;
		float[] v, tex = null;
		if (!updateSphereRenderData && !billboards) return;
		
		if (billboards){
			GLMatrix rotInverse = rotMatrix.clone();
			rotInverse.inverse();
			float[] tmp = new float[16];
			rotInverse.getMatrix().get(tmp, 0, 16);
			
			
			tex = new float[]{0.5f,0.5f,  0f, 1f,   1f, 0f,  1f, 1f,  0f, 0f};
			
			t = new int[]{0,1,4,2,3,1};
			float[] v_orig = new float[]{0f,0f,1f, -1f,1f,0f,  1f,-1f,0f, 1f,1f,0f, -1f,-1f,0f};
			v = new float[15];
			v[0] = v_orig[0] * tmp[0] + v_orig[1] * tmp[4] + v_orig[2] * tmp[8]; 
			v[1] = v_orig[0] * tmp[1] + v_orig[1] * tmp[5] + v_orig[2] * tmp[9];
			v[2] = v_orig[0] * tmp[2] + v_orig[1] * tmp[6] + v_orig[2] * tmp[10];
			
			v[3] = v_orig[3] * tmp[0] + v_orig[4] * tmp[4] + v_orig[5] * tmp[8]; 
			v[4] = v_orig[3] * tmp[1] + v_orig[4] * tmp[5] + v_orig[5] * tmp[9];
			v[5] = v_orig[3] * tmp[2] + v_orig[4] * tmp[6] + v_orig[5] * tmp[10];
			
			v[6] = v_orig[6] * tmp[0] + v_orig[7] * tmp[4] + v_orig[8] * tmp[8]; 
			v[7] = v_orig[6] * tmp[1] + v_orig[7] * tmp[5] + v_orig[8] * tmp[9];
			v[8] = v_orig[6] * tmp[2] + v_orig[7] * tmp[6] + v_orig[8] * tmp[10];
			
			v[9] = v_orig[9] * tmp[0] + v_orig[10] * tmp[4] + v_orig[11] * tmp[8]; 
			v[10] = v_orig[9] * tmp[1] + v_orig[10] * tmp[5] + v_orig[11] * tmp[9];
			v[11] = v_orig[9] * tmp[2] + v_orig[10] * tmp[6] + v_orig[11] * tmp[10];
			
			v[12] = v_orig[12] * tmp[0] + v_orig[13] * tmp[4] + v_orig[14] * tmp[8]; 
			v[13] = v_orig[12] * tmp[1] + v_orig[13] * tmp[5] + v_orig[14] * tmp[9];
			v[14] = v_orig[12] * tmp[2] + v_orig[13] * tmp[6] + v_orig[14] * tmp[10];
			
		} else {
			Tupel<int[], float[]> sphere = SphereGenerator.getSphericalVBOData(highQuality ? 1 : 0);
			t = sphere.getO1();
			v = sphere.getO2();
		}
		
		sphereVBOIndexCount = t.length;
		if (sphereVboIndices != null)
			gl.glDeleteBuffers(3, sphereVboIndices, 0);
		else  sphereVboIndices = new int[3];
		
		gl.glGenBuffers(3, sphereVboIndices, 0);
		
		IntBuffer tb = Buffers.newDirectIntBuffer(t.length);
		tb.put(t);
		tb.rewind();
		gl.glBindBuffer(GL.GL_ARRAY_BUFFER, sphereVboIndices[0]);
		gl.glBufferData(GL.GL_ARRAY_BUFFER, tb.capacity()*Integer.SIZE/8, tb, GL.GL_STATIC_DRAW);
		gl.glUnmapBuffer(GL.GL_ARRAY_BUFFER);
		
		FloatBuffer vb = Buffers.newDirectFloatBuffer(v.length);
		vb.put(v);
		vb.rewind();
		gl.glBindBuffer(GL.GL_ARRAY_BUFFER, sphereVboIndices[1]);
		gl.glBufferData(GL.GL_ARRAY_BUFFER, vb.capacity()*Float.SIZE/8, vb, GL.GL_STATIC_DRAW);
		gl.glUnmapBuffer(GL.GL_ARRAY_BUFFER);
		
		if (tex != null){
			FloatBuffer texb = Buffers.newDirectFloatBuffer(tex.length);
			texb.put(tex);
			texb.rewind();
			gl.glBindBuffer(GL.GL_ARRAY_BUFFER, sphereVboIndices[2]);
			gl.glBufferData(GL.GL_ARRAY_BUFFER, texb.capacity()*Float.SIZE/8, texb, GL.GL_STATIC_DRAW);
			gl.glUnmapBuffer(GL.GL_ARRAY_BUFFER);
		}
		
		this.updateSphereRenderData = false;
	}
	
	public float[] getNextPicking(GL2 gl, Pickable pickableObject){
		float[] f = getNextPickingColor(pickableObject);
		gl.glUniform4f(pickingShaderUniform, f[0], f[1], f[2], 1.0f);
		return f;
	}
	
	public float[] getNextPickingColor(Pickable pickableObject){
		int pickingIndex = pickMap.size();
		int r = (pickingIndex & 0xff0000) >> 16;
		int g = (pickingIndex & 0xff00) >> 8;
		int b = pickingIndex & 0xff;
		
		pickMap.add(pickableObject);
		
		float[] f = {r*0.003921569f, g*0.003921569f, b*0.003921569f};
		return f;
	}
	
	private void drawSpheres(GL2 gl, AtomRenderData ard, boolean picking){
		int numEle = Configuration.getCrystalStructure().getNumberOfElements();
		
		gl.glDisable(GL.GL_BLEND); //Transparency can cause troubles and should be avoided, disabling blending might be faster then
		
		//Activate vbos and shaders
		gl.glEnableClientState(GL2.GL_VERTEX_ARRAY);
		gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, sphereVboIndices[0]);
		gl.glBindBuffer(GL.GL_ARRAY_BUFFER, sphereVboIndices[1]);
		gl.glVertexPointer(3,GL.GL_FLOAT, 0, 0);
		
		//Sphere render uses triangles, the billboards triangle_strips
		int renderFormat = GL.GL_TRIANGLES;
		Shader activeShader;
		
		//Select the rendering shader (Billboards or spheres, instanced or not, picking)
		if (Configuration.Options.BILLBOARD_RENDERING.isEnabled()) {
			if (picking){
				activeShader = Shader.BILLBOARD_PICKING_SHADER;
				pickingShaderUniform = gl.glGetUniformLocation(activeShader.getProgram(), "pickcolor");
			} else {
				activeShader = Shader.BILLBOARD_SHADER;
			}
			//Billboards need their texture coordinates binded
			gl.glEnableClientState(GL2.GL_TEXTURE_COORD_ARRAY);
			gl.glBindBuffer(GL2.GL_ARRAY_BUFFER, sphereVboIndices[2]);
			gl.glTexCoordPointer(2,GL.GL_FLOAT, 0, 0);
			
			gl.glActiveTexture(GL.GL_TEXTURE2);
			gl.glBindTexture(GL.GL_TEXTURE_2D, Configuration.Options.HQ_RENDERING.isEnabled()
					?sphereTexHQ.getTextureName():sphereTexLQ.getTextureName());
			renderFormat = GL.GL_TRIANGLE_FAN;
			
			gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MAG_FILTER, GL2.GL_LINEAR);
			gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MIN_FILTER, GL2.GL_LINEAR);
			
		} else {
			if (picking){
				activeShader = Shader.SPHERE_PICKING_SHADER;
				pickingShaderUniform = gl.glGetUniformLocation(activeShader.getProgram(), "pickcolor");
			} else {
				if (Configuration.Options.ADS_SHADING.isEnabled())
					activeShader = Shader.SPHERE_ADS_SHADER;
				else activeShader = Shader.SPHERE_SHADER;
			}
		}
		
		//Set all uniforms, depending on the active shader, some may not be used
		GLMatrix pmvp = projectionMatrix.clone();
		pmvp.mult(modelViewMatrix);
		activeShader.enable(gl);
		gl.glUniformMatrix4fv(gl.glGetUniformLocation(activeShader.getProgram(), "mvm"), 1, false, modelViewMatrix.getMatrix());
		gl.glUniformMatrix3fv(gl.glGetUniformLocation(activeShader.getProgram(), "nm"), 1, false, modelViewMatrix.getNormalMatrix());
		gl.glUniformMatrix4fv(gl.glGetUniformLocation(activeShader.getProgram(), "mvpm"), 1, false, pmvp.getMatrix());
		if (Configuration.Options.BILLBOARD_RENDERING.isEnabled()) {
			gl.glUniform1i(gl.glGetUniformLocation(activeShader.getProgram(), "Texture0"), 2);
		}
		
		int sphereColorUniform = gl.glGetUniformLocation(activeShader.getProgram(), "Color");
		int sphereTranslateUniform;
		sphereTranslateUniform = gl.glGetUniformLocation(activeShader.getProgram(), "Move");
		
		for (int i=0; i<ard.getRenderableCells().size(); ++i){
			Cell c = ard.getRenderableCells().get(i);
			for (int j=0; j<c.getNumAtoms(); j++){
				Atom a = c.getAtoms().get(j).atom;
				if (picking) getNextPicking(gl, a);
				else {
					float[] col = c.getAtoms().get(j).color;
					gl.glUniform4f(sphereColorUniform, col[0], col[1], col[2], 1f);
				}
				gl.glUniform4f(sphereTranslateUniform, a.getX(), a.getY(), a.getZ(),sphereSize[a.getElement()%numEle]);
				gl.glDrawElements(renderFormat, sphereVBOIndexCount, GL.GL_UNSIGNED_INT, 0);
			}
		}
		
		
		//Disable vbos and switch active shader
		gl.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
		gl.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
		gl.glDisableClientState(GL2.GL_VERTEX_ARRAY);
		
		if (Configuration.Options.BILLBOARD_RENDERING.isEnabled()) {
			gl.glActiveTexture(GL.GL_TEXTURE2);
			gl.glBindTexture(GL.GL_TEXTURE_2D, 0);
			gl.glActiveTexture(GL.GL_TEXTURE0);
			gl.glDisableClientState(GL2.GL_TEXTURE_COORD_ARRAY);
			gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MAG_FILTER, GL2.GL_NEAREST);
			gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MIN_FILTER, GL2.GL_NEAREST);
		}
		
		if (picking) {
			Shader.PICKING_SHADER.enable(gl);
			pickingShaderUniform = gl.glGetUniformLocation(Shader.PICKING_SHADER.getProgram(), "pickcolor");   
		}
		else {
			Shader.DEFAULT_PPL_SHADER.enable(gl);
		}

		if (!picking) gl.glEnable(GL.GL_BLEND);
	}
	
	private void drawSpheres(GL3 gl, AtomRenderData ard, boolean picking){
		int numEle = Configuration.getCrystalStructure().getNumberOfElements();
		
		gl.glDisable(GL.GL_BLEND); //Transparency can cause troubles and should be avoided, disabling blending might be faster then
		
		//Select the rendering shader
		int renderFormat = 0;
		GL3Shader shader;
		if (Configuration.Options.BILLBOARD_RENDERING.isEnabled()){
			if (picking) shader = GL3Shader.BILLBOARD_PICKING_SHADER;
			else shader = GL3Shader.BILLBOARD_SHADER;
			
			gl.glActiveTexture(GL.GL_TEXTURE2);
			gl.glBindTexture(GL.GL_TEXTURE_2D, Configuration.Options.HQ_RENDERING.isEnabled()
					?sphereTexHQ.getTextureName():sphereTexLQ.getTextureName());
			renderFormat = GL.GL_TRIANGLE_FAN;
			
			gl.glTexParameteri(GL2.GL_TEXTURE_2D, GL2.GL_TEXTURE_MAG_FILTER, GL2.GL_LINEAR);
			
		} else {
			if (picking) shader = GL3Shader.SPHERE_PICKING_SHADER;
			else {
				if (Configuration.Options.ADS_SHADING.isEnabled())
					shader = GL3Shader.SPHERE_ADS_SHADER;
				else shader = GL3Shader.SPHERE_SHADER;
			}
			renderFormat = GL.GL_TRIANGLES;
		}
		
		//Set all uniforms, depending on the active shader, some may not be used
		GLMatrix pmvp = projectionMatrix.clone();
		pmvp.mult(modelViewMatrix);
		
		GL3 gl3 = gl.getGL3();
		
		GL3Shader visTestShader = GL3Shader.VISIBILTY_TEST_SHADER;
		visTestShader.enable(gl3);
		int visShiftUniform = gl.glGetUniformLocation(visTestShader.getProgram(), "shift");
		int visScaleUniform = gl.glGetUniformLocation(visTestShader.getProgram(), "scale");
		gl.glUniformMatrix4fv(gl.glGetUniformLocation(visTestShader.getProgram(), "mvpm"), 1, false, pmvp.getMatrix());

		shader.enable(gl3);
		gl.glUniformMatrix4fv(gl.glGetUniformLocation(shader.getProgram(), "mvm"), 1, false, modelViewMatrix.getMatrix());
		gl.glUniformMatrix3fv(gl.glGetUniformLocation(shader.getProgram(), "nm"), 1, false, modelViewMatrix.getNormalMatrix());
		gl.glUniformMatrix4fv(gl.glGetUniformLocation(shader.getProgram(), "mvpm"), 1, false, pmvp.getMatrix());
		if (Configuration.Options.BILLBOARD_RENDERING.isEnabled()) {
			gl.glUniform1i(gl.glGetUniformLocation(shader.getProgram(), "Texture0"), 2);
		}
				
		gl3.glBindBuffer(GL.GL_ARRAY_BUFFER, sphereVboIndices[1]);
		gl3.glVertexAttribPointer(ShaderCollection.ATTRIB_VERTEX, 3, GL.GL_FLOAT, false, 0, 0);
		
		if (Configuration.Options.BILLBOARD_RENDERING.isEnabled()){
			gl3.glBindBuffer(GL.GL_ARRAY_BUFFER, sphereVboIndices[2]);
			gl3.glVertexAttribPointer(ShaderCollection.ATTRIB_TEX0, 2, GL.GL_FLOAT, false, 0, 0);
		}

		gl3.glVertexAttribDivisor(ShaderCollection.ATTRIB_VERTEX, 0);
		gl3.glVertexAttribDivisor(ShaderCollection.ATTRIB_VERTEX_OFFSET, 1);
		gl3.glVertexAttribDivisor(ShaderCollection.ATTRIB_COLOR, 1);
		
		if (Configuration.Options.BILLBOARD_RENDERING.isEnabled()){
			gl3.glVertexAttribDivisor(ShaderCollection.ATTRIB_TEX0, 0);
		}
		
		CellRenderBuffer cbr = cellRenderRingBuffer.getCurrent();
		ard.sortCells(modelViewMatrix);
		
		HashMap<Cell, int[]> currentActiveRenderingTest = new HashMap<AtomRenderData.Cell, int[]>();
		
//		int cellsDrawn = 0;
		
		for (int j=0; j<ard.getRenderableCells().size(); j++){
			Cell c = ard.getRenderableCells().get(j);
			
			boolean renderCell = true;
			//Test if the cell needs to be rendered 
			if (currentActiveRenderingTest.containsKey(c)){
				
				int[] result = new int[1];
				int[] query = currentActiveRenderingTest.remove(c);
				gl3.glGetQueryObjectuiv(query[0], GL3.GL_QUERY_RESULT, result, 0);
				if (result[0]==0) 
					renderCell=false;
				gl3.glDeleteQueries(1, query, 0);
			}
			
			if (renderCell){
//				cellsDrawn++;
				//Full renderer 
				shader.enable(gl3);
				if (cbr.fence != -1){
					// Wait until last call using the buffer is finished
					gl.glClientWaitSync(cbr.fence, GL3.GL_SYNC_FLUSH_COMMANDS_BIT, GL3.GL_TIMEOUT_IGNORED);
					gl.glDeleteSync(cbr.fence);
				}
				gl3.glBindBuffer(cbr.bufferType, cbr.buffer);
				gl3.glBufferSubData(cbr.buffer, 0, cbr.bufferSize, null);
				FloatBuffer buf = gl3.glMapBuffer(cbr.bufferType, GL.GL_WRITE_ONLY).asFloatBuffer();
				for (int i=0; i<c.getNumAtoms(); i++){
					RenderableAtom ra = c.getAtoms().get(i);
					Atom a = ra.atom;
					float[] col = picking ? getNextPickingColor(a) : ra.color;
					buf.put(col[0]); buf.put(col[1]); buf.put(col[2]); buf.put(1f);
					buf.put(a.getX()); buf.put(a.getY()); buf.put(a.getZ()); buf.put(sphereSize[a.getElement()%numEle]);
				}
				gl3.glUnmapBuffer(cbr.bufferType);
				gl3.glVertexAttribPointer(ShaderCollection.ATTRIB_COLOR, 4, GL.GL_FLOAT, false, 8*Float.SIZE/8, 0);
				gl3.glVertexAttribPointer(ShaderCollection.ATTRIB_VERTEX_OFFSET, 4, GL.GL_FLOAT, false, 8*Float.SIZE/8, 4*Float.SIZE/8);
				gl3.glBindBuffer(GL.GL_ARRAY_BUFFER, sphereVboIndices[1]);
				gl3.glVertexAttribPointer(ShaderCollection.ATTRIB_VERTEX, 3, GL.GL_FLOAT, false, 0, 0);
				gl3.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, sphereVboIndices[0]);
				gl3.glDrawElementsInstanced(renderFormat, sphereVBOIndexCount, GL.GL_UNSIGNED_INT, null, c.getNumAtoms());
				
				cbr.fence = gl.glFenceSync(GL3.GL_SYNC_GPU_COMMANDS_COMPLETE, 0);
				cbr = cellRenderRingBuffer.getNext();
				gl.glFlush();
			}
			
			if (j+8<ard.getRenderableCells().size()){
				Cell futureCell = ard.getRenderableCells().get(j+8);
				int[] query = new int[1]; 
				gl3.glGenQueries(1, query, 0);
				gl3.glColorMask(false, false, false, false);
				gl3.glDepthMask(false);
				gl3.glBeginQuery(GL3.GL_SAMPLES_PASSED, query[0]);
				//Test renderer
//				gl3.glPolygonMode(GL.GL_FRONT_AND_BACK, GL3.GL_LINE);
				visTestShader.enable(gl3);
				gl3.glBindBuffer(GL.GL_ARRAY_BUFFER, cubeVboIndices[1]);
				gl3.glVertexAttribPointer(ShaderCollection.ATTRIB_VERTEX, 3, GL.GL_FLOAT, false, 0, 0);
				gl3.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, cubeVboIndices[0]);
				float[] f = futureCell.getOffset();
				gl.glUniform3f(visShiftUniform, f[0], f[1], f[2]);
				f = futureCell.getSize();
				gl.glUniform3f(visScaleUniform, f[0], f[1], f[2]);
				gl3.glDrawElements(GL.GL_TRIANGLES, 36, GL.GL_UNSIGNED_INT, 0);
//				gl3.glPolygonMode(GL.GL_FRONT_AND_BACK, GL3.GL_FILL);
				gl3.glEndQuery(GL3.GL_SAMPLES_PASSED);
				gl3.glColorMask(true, true, true, true);
				gl3.glDepthMask(true);
				
				currentActiveRenderingTest.put(futureCell, query);
			}
		}
		
		//System.out.println("Drawn "+cellsDrawn+" of "+ard.getRenderableCells().size());
		
		gl3.glBindBuffer(GL.GL_ARRAY_BUFFER, 0);
		gl3.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0);
		
		gl3.glVertexAttribDivisor(ShaderCollection.ATTRIB_VERTEX_OFFSET, 0);
		gl3.glVertexAttribDivisor(ShaderCollection.ATTRIB_COLOR, 0);
		gl.glActiveTexture(0);
		
		shader.disable(gl3);
		
		if (picking) {
			Shader.PICKING_SHADER.enable(gl);
			pickingShaderUniform = gl.glGetUniformLocation(Shader.PICKING_SHADER.getProgram(), "pickcolor");   
		}
		else {
			Shader.DEFAULT_PPL_SHADER.enable(gl);
		}
		
		if (!picking) gl.glEnable(GL.GL_BLEND);
	}
	
	
	@Override
	public void dispose(GLAutoDrawable arg0) {
		//Delete all GL context related objects
		GL2 gl = arg0.getGL().getGL2();
		ShaderCollection.dispose(gl);
		if (sphereVboIndices != null)
			gl.glDeleteBuffers(3, sphereVboIndices, 0);
		sphereVboIndices = null;
		if (cubeVboIndices != null)
			gl.glDeleteBuffers(2, cubeVboIndices, 0);
		cubeVboIndices = null;
		
		if (cellRenderRingBuffer != null){
			for (int i=0; i<cellRenderRingBuffer.size(); i++)
				cellRenderRingBuffer.getNext().dispose(gl);
		}
		cellRenderRingBuffer = null;
	}
	
	private class CellRenderBuffer{
		int buffer;
		int bufferSize;
		long fence = -1;
		int bufferType;
		
		void init(GL2GL3 gl, int size, int bufferType){
			this.bufferType = bufferType;
			int[] buf = new int[1];
			gl.glGenBuffers(1, buf, 0);
	
			gl.glBindBuffer(bufferType, buf[0]);
			gl.glBufferData(bufferType, size, null, GL.GL_DYNAMIC_DRAW);
			gl.glUnmapBuffer(bufferType);
	
			this.buffer = buf[0];
			this.bufferSize = size;
		}
		
		void dispose(GL2GL3 gl){
			int[] buffers = new int[1];
			buffers[0] = buffer;
			gl.glDeleteBuffers(1, buffers, 0);
			if (fence != -1 && openGLVersion >= 3)
				gl.getGL3().glDeleteSync(fence);
		}
	}
}