// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model;

import java.awt.event.InputEvent;
import java.util.*;

import common.CommonUtils;
import common.VectorOps;

/**
 * This is the basic class for storing data per atom
 * @author Christoph Begau
 *
 */
public class Atom implements Coordinate, Comparable<Atom>, Pickable{
	public static final int IGNORED_GRAIN = 0xFFFF;
	public static final int DEFAULT_GRAIN = 0xFFFE;
	
	private final int number;
	private Atom[] neigh;
	private final float x,y,z;
	private RBV rbv;
	private float[] customValues;
	
	//Storing some small types together in one int byte 0-7: type, 8-15: element, 16-31: grain
	//This saves some memory (min 8bytes per atom) at the cost of bitmask operations
	private int grainElementType = 0;
	
	/**
	 * @param p float array with x,y,z
	 * @param typeNumEle array with type, number, element
	 */
	public Atom(float[] p, int[] typeNumEle) {
		//No element/type > 255 and <0 is allowed
		assert(typeNumEle[0]>=0 && typeNumEle[0]<256);
		assert(typeNumEle[2]>=0 && typeNumEle[2]<256);

		this.x = p[0];
		this.y = p[1];
		this.z = p[2];
		this.number = typeNumEle[1];
		this.grainElementType = (DEFAULT_GRAIN<<16) | (typeNumEle[2] << 8) | typeNumEle[0];
		if (ImportStates.CUSTOM_VALUES.isActive())
			customValues = new float[Configuration.getSizeCustomColumns()];
	}
	
	@Override
	public float getX() {
		return x;
	}
	
	@Override
	public float getY() {
		return y;
	}
	
	@Override
	public float getZ() {
		return z;
	}
	
	public RBV getRBV() {
		return rbv;
	}
	
	public void setRBV( float[] rbv, float[] lineSense ){
		if (rbv == null || lineSense == null){
			this.rbv = null;
		} else {
			if (lineSense[0] != 0f && lineSense[1] != 0f && lineSense[2] != 0f)  
				this.rbv = new model.RBV(rbv, lineSense);
		}
	}
	
	public void setCustom(float v, int n){
		assert(n<customValues.length);
		customValues[n]=v;
	};
	
	public float getCustom(int n){
		assert(n<customValues.length);
		return customValues[n];
	}
	
	/**
	 * ZYX notation
	 * @param eulerAngles
	 */
	public void setEulerAngles(float[] eulerAngles) {
		if (ImportStates.LATTICE_ROTATION.isActive()){
			int index = Configuration.getLatticeRotationColumn();
			setCustom(eulerAngles[0], index);
			setCustom(eulerAngles[1], index+1);
			setCustom(eulerAngles[2], index+2);
		}
	}
	
	public float[] getEulerAngles() {
		if (ImportStates.LATTICE_ROTATION.isActive()){
			int index = Configuration.getLatticeRotationColumn();
			return new float[]{getCustom(index), getCustom(index+1), getCustom(index+2)};
		}
		else return null; 
	}
	
	public float[] getAxisRotations() {
		float[] axisRotation = new float[3];
		axisRotation[0] = this.getAxisRotation(0);
		axisRotation[1] = this.getAxisRotation(1);
		axisRotation[2] = this.getAxisRotation(2);
		return axisRotation;
	}
	
	/**
	 * Get the rotation axis around the coordinate systems axes
	 * @param a
	 * @param axis 0: X-axis, 1: Y-axis, 2: Z-axis
	 * @return
	 */
	public float getAxisRotation(int axis){
		float[] euler = this.getEulerAngles();
		if (euler == null) return 0;
		
		double y = -Math.asin(-Math.cos(euler[1])*Math.sin(euler[0]));
		switch(axis){
			case 0: return (float)Math.toDegrees(Math.asin(Math.sin(euler[1])/Math.cos(y)));
			case 1: return (float)Math.toDegrees(y);
			case 2: return (float)Math.toDegrees(Math.asin( ((Math.sin(euler[2])*Math.cos(euler[1])) / Math.cos(y))));
		}
		return 0f;
	}
	
	public float getLatticeTilt() {
		//Calculate via the trace of the rotation matrix
		float[] euler = getEulerAngles();
		if (euler == null) return 0f;
		double s1 = Math.sin(euler[0]);
		double s2 = Math.sin(euler[1]);
		double s3 = Math.sin(euler[2]);
		double c1 = Math.cos(euler[0]);
		double c2 = Math.cos(euler[1]);
		double c3 = Math.cos(euler[2]);
		double a = (s1*s2*s3 + c1*c2 + c1*c3 + c2*c3 - 1.)*0.5;
		return (float) Math.toDegrees(Math.acos(a));
	}
	
	public int getElement() {
		return (grainElementType & 0x0000FF00)>>>8;
	}
	
	/**
	 * The grain the atom is located in 
	 * @return the index of the grain the atom is placed inside
	 */
	public int getGrain() {
		return (grainElementType & 0xFFFF0000)>>>16;
	}
	
	/**
	 * No value > 65533 and <0 is allowed
	 * (65534 & 65535 are reserved for IGNORED_GRAIN and DEFAULT_GRAIN)
	 * @param grain
	 */
	public void setGrain(int grain) {
		assert(grain>=0 && grain<=65535);
		this.grainElementType = (grainElementType & 0x0000FFFF) | (grain << 16);
		if (grain == IGNORED_GRAIN) rbv = null;
	}
	
	/**
	 * No value > 255 and <0 is allowed
	 * @param type
	 */
	public void setType(int type) {
		assert(type>=0 && type<256);
		this.grainElementType = ((grainElementType & 0xFFFFFF00) | type);
	}
	
	public int getType() {
		return (grainElementType & 0x000000FF);
	}
	
	public Atom[] getNeigh() {
		return neigh;
	}
	
	public void buildNeighOfSameElement(NearestNeighborBuilder<Atom> nnb){
		ArrayList<Atom> neigh = nnb.getNeigh(this);
		ArrayList<Atom> neighSameElement = new ArrayList<Atom>(12);
		int numEle = Configuration.getCrystalStructure().getNumberOfElements();
		int element = getElement() % numEle;
		for (int i=0; i< neigh.size(); i++){
			if (element == neigh.get(i).getElement() % numEle)
				neighSameElement.add(neigh.get(i));
		}
		
		this.neigh = neighSameElement.toArray(new Atom[neighSameElement.size()]);
	}
	
	public void buildNeigh(NearestNeighborBuilder<Atom> nnb){
		ArrayList<Atom> neigh = nnb.getNeigh(this);
		this.neigh = neigh.toArray(new Atom[neigh.size()]);
	}
	
	public void removeNeigh(){
		this.neigh = null;
	}
	
	public int getNumber(){
		return number;
	}
	
	@Override
	public boolean equals(Object obj){
		if (!(obj instanceof Atom)) return false;
		Atom a = (Atom)obj;
		return a.number == this.number;
	}
	
	@Override
	public int hashCode() {
		return number;
	}
	
	@Override
	public int compareTo(Atom o) {
		return this.number - o.number;
	}
	
	@Override
	public Collection<? extends Object> getHighlightedObjects() {
		return null;
	}

	@Override
	public boolean isHighlightable() {
		return false;
	}
	
	@Override
	public String printMessage(int modifier) {
		StringBuilder sb = new StringBuilder();
		
		float[] offset = Configuration.getCurrentAtomData().getOffset();
		sb.append(String.format("Nr=%d, ", number));
		sb.append(String.format("xyz=( %.4f | %.4f | %.4f ),", x+offset[0], y+offset[1], z+offset[2]));
		sb.append(String.format(" (%s),", Configuration.getCrystalStructure().getNameForType(getType())));
		if (Configuration.getNumElements()>1) sb.append(String.format(" Element=%d,", getElement()));
		if (ImportStates.POLY_MATERIAL.isActive() && getGrain() != DEFAULT_GRAIN) 
			sb.append(String.format(" Grain=%s", getGrain()==IGNORED_GRAIN?"None":Integer.toString(getGrain())));
		if (getRBV()!=null) {
			CrystalRotationTools crt = null;
			if (ImportStates.POLY_MATERIAL.isActive()){
				//Poly crystal
				if (getGrain() == DEFAULT_GRAIN)
					crt = Configuration.getCrystalRotationTools();
				else crt = Configuration.getCurrentAtomData().getGrains(getGrain()).getCystalRotationTools();
			} else {
				//Single crystal
				crt = Configuration.getCrystalRotationTools();
			}
			
			float[] bv = crt.getInCrystalCoordinates(this.getRBV().getRBVAsVector());
			sb.append(String.format(" RBV=[ %.3f | %.3f | %.3f ],", bv[0], bv[1], bv[2]));
			sb.append(String.format(" Lenght=%.3f", VectorOps.length(this.getRBV().getRBVAsVector())));
			float[] ld = VectorOps.normalize(crt.getInCrystalCoordinates(this.getRBV().getLineDirectionAsVector()));
			sb.append(String.format(" LineDir=[ %.3f | %.3f | %.3f ],", ld[0], ld[1], ld[2]));
			
			BurgersVector tbv = crt.rbvToBurgersVector(this.getRBV());
			sb.append(" TBV="+tbv.toString());
		}
		
		if (ImportStates.LATTICE_ROTATION.isActive()) {
			float[] euler = getEulerAngles();
			if (euler != null){
				sb.append(String.format(" Lattice Tilt (%.4f)", getLatticeTilt()));
				sb.append(String.format(" e1(%.4f)", Math.toDegrees(euler[0])));
				sb.append(String.format(" e2(%.4f)", Math.toDegrees(euler[1])));
				sb.append(String.format(" e3(%.4f)", Math.toDegrees(euler[2])));
				float[] axisRotation = getAxisRotations();
				sb.append(String.format(" RotX(%.4f)", axisRotation[0]));
				sb.append(String.format(" RotY(%.4f)", axisRotation[1]));
				sb.append(String.format(" RotZ(%.4f)", axisRotation[2]));
			}
		}
		
		if (ImportStates.CUSTOM_VALUES.isActive()){
			for (int i=0; i< Configuration.getSizeCustomColumns(); i++){
				CustomColumnInfo c = Configuration.getCustomColumnInfo(i);
				if (!c.isSpecialColoumn())
					sb.append(String.format(" %s: %s%s", c.getName(), CommonUtils.outputDecimalFormatter.format(getCustom(i)), c.getUnit()));	
			}
		}
		
		if ((modifier & InputEvent.CTRL_DOWN_MASK) == InputEvent.CTRL_DOWN_MASK){
			sb.append("\n");
			sb.append(Configuration.getCurrentAtomData().plotNeighborsGraph(this));
		}
		return sb.toString();
	}
}