// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model;

import java.io.*;
import java.util.*;
import java.util.concurrent.Callable;
import java.util.concurrent.CyclicBarrier;
import java.util.zip.GZIPOutputStream;

import javax.swing.JOptionPane;

import model.dataContainer.*;
import model.dislocationDensity.DislocationDensityTensorData;
import model.polygrain.*;
import model.skeletonizer.*;
import common.*;
import crystalStructures.MonoclinicNiTi;

public class AtomData {
	/**
	 * Simulation box size, origin is at (0,0,0)
	 * Atoms outside this box can be displayed, but are ignored during most other calculations
	 */
	private float[] boxSize = new float[3];
	/**
	 * In lammps the coordinate system does not necessarily start at (0,0,0)
	 * Store the shift so correct coordinates can be given when picking an atom 
	 */
	private float[] offset = new float[3];
	/**
	 * All atoms are stored in this list 
	 */
	private ArrayList<Atom> atoms = new ArrayList<Atom>();
	/**
	 * Only used in polycrystalline / polyphase material:
	 * Each subgrain is assigned a number, each atom is assigned a number of the grain its belongs to
	 */
	private HashMap<Integer, Grain> grains = new HashMap<Integer, Grain>();
	/**
	 * The largest (virtual) elements number
	 */
	private int maxNumElements = 1;
	//Some flags for imported or calculated values
	private boolean rbvAvailable = false;
	private boolean atomTypesAvailable = false;
	private boolean grainsImported = false;
	private boolean meshImported = false;
	private boolean latticeRotationImported = false;
	/**
	 * Meta data found in the file header
	 */
	private Map<String, Object> fileMetaData = null;
	/**
	 * Identifier of the simulation, usually the filename
	 */
	private String name;
	
	private ArrayList<DataContainer> additionalData = new ArrayList<DataContainer>();
	
	/**
	 * Skeletonizer: stores the dislocation network
	 */
	private Skeletonizer skeletonizer = null;
	
	private int[] atomsPerElement = new int[0];
	private int[] n_types;
	private Tupel<NearestNeighborFilter, Float> lastNearestNeighborBuild = new Tupel<NearestNeighborFilter, Float>(null, -1f);
	private AtomData next, previous;
	private DislocationDensityTensorData ddtd;
	
	private static NearestNeighborFilter defectNearestNeighborFilter = new NearestNeighborFilter() {
		private final int defaultType = Configuration.getCrystalStructure().getDefaultType();
		
		@Override
		public boolean accept(Atom a) {
			if (a.getType() != defaultType && (!ImportStates.POLY_MATERIAL.isActive() || a.getGrain()!=Atom.IGNORED_GRAIN))
				return true;
			else return false;
		}
	};
	
	private static NearestNeighborFilter nearestNeighborNonFiltered = new NearestNeighborFilter() {
		@Override
		public boolean accept(Atom a) {
			return true;
		}
	};
	
	public AtomData(AtomData previous, MDFileLoader.ImportDataContainer idc){
		this.n_types = new int[Configuration.getCrystalStructure().getNumberOfTypes()];
		this.boxSize = idc.boxSize;
		this.offset = idc.offset;
		this.atoms = idc.atoms;
		this.grains = idc.grains;
		this.meshImported = idc.meshImported;
		this.maxNumElements = idc.maxNumElements;
		this.rbvAvailable = idc.rbvAvailable;
		this.atomTypesAvailable = idc.atomTypesAvailable;
		this.grainsImported = idc.grainsImported;
		this.latticeRotationImported = idc.latticeRotationImported;
		this.fileMetaData = idc.fileMetaData;
		this.name = idc.name;
		Configuration.setCurrentAtomData(this);
		
		if (Configuration.getNumElements()<idc.maxNumElements)
			Configuration.setNumElements(idc.maxNumElements);
		
		if (previous!=null){
			previous.next = this;
			this.previous = previous;
		}
		
		processInputData();
	}

	public ArrayList<DataContainer> getAdditionalData() {
		return additionalData;
	}
	
	private void processInputData() {
		//Short circuit for empty files
		if (this.atoms.size() == 0) return;
		
		//Set ranges for customColums
		if (ImportStates.CUSTOM_VALUES.isActive()){
			for (int i=0; i<Configuration.getSizeCustomColumns(); i++){
				Configuration.getCustomColumnInfo(i).findRange(true);
			}
		}
		
		//Copy the atom-arrayList into a new ArrayList which fits exactly in size
		//otherwise the memory overhead can be quite large
		if (!ImportStates.KILL_ALL_ATOMS.isActive() && !ImportStates.DISPOSE_DEFAULT.isActive())
			this.atoms.trimToSize();
		
		//Bond Angle Analysis
		if (!atomTypesAvailable)
			StructuralAnalysisBuilder.performStructureAnalysis(this);
		
		if (ImportStates.FILTER_SURFACE.isActive()){
			this.filterSurface();
		}
		
		if (ImportStates.POLY_MATERIAL.isActive()){
			if (ImportStates.BURGERS_VECTORS.isActive() && 
					Configuration.getCrystalStructure().createRBVbeforeGrains() && !rbvAvailable){
				RbvBuilder.createRBV(this);
				rbvAvailable = true;
			}
			
			final List<Grain> gr = Configuration.getCrystalStructure().identifyGrains(this);
			for (Grain g : gr){
				this.grains.put(g.getGrainNumber(), g);
			}
		
			if (ImportStates.BURGERS_VECTORS.isActive() && !rbvAvailable) {
				// RBV poly-crystal
				for (Grain g : gr)
					RbvBuilder.createRBV(g, boxSize);
				rbvAvailable = true;
			}
			
			Grain.processGrains(this);
			
			for (Grain g : gr) {
				// Make sure all meshes are calculated if grains are imported
				g.getMesh().finalizeMesh();
				if (!Configuration.getCrystalStructure().orderGrainsBySize())
					g.getAtomsInGrain().clear();	//Atoms are not needed anymore if there is no need for reordering
			}
			
			if (!Configuration.getCrystalStructure().orderGrainsBySize()){
				for (Grain g : gr) {
					Configuration.addGrainIndex(g.getGrainNumber());
				}
			}
		} else if (ImportStates.BURGERS_VECTORS.isActive() && !rbvAvailable){
			//RBV single-crystal
			RbvBuilder.createRBV(this);
			rbvAvailable = true;
		}
		
		if (ImportStates.SKELETONIZE.isActive()){
			this.getSkeletonizer().transform();
		}
		
		if (ImportStates.LATTICE_ROTATION.isActive() && !latticeRotationImported){
			LatticeRotationBuilder.calculateLatticeRotation(this);
		}
		
		if (ImportStates.DETECT_MARTENSITE_VARIANTS.isActive()){
			MonoclinicNiTi.identifyVariants(this);
		}
		
		//DDTs
		if (ImportStates.DISLOCATION_DENSITY.isActive()){
			ddtd = new DislocationDensityTensorData(this);
		}
		
		this.removeNearestNeigbors();
		
		if (ImportStates.KILL_ALL_ATOMS.isActive()) 			//Dispose all atoms
			atoms.clear();
		else if (ImportStates.DISPOSE_DEFAULT.isActive()){ //Dispose perfect atoms
			int defaultType = Configuration.getCrystalStructure().getDefaultType();
			ArrayList<Atom> atoms = new ArrayList<Atom>();
			for (int i=0; i<this.atoms.size();i++){
				if (this.atoms.get(i).getType() != defaultType)
					atoms.add(this.atoms.get(i));
			}
			this.atoms = atoms;
		}
		
		//Count all atoms types
		for (int i=0; i<atoms.size();i++){
			if (atoms.get(i).getType() >= 0 && atoms.get(i).getType() < n_types.length)
				n_types[atoms.get(i).getType()]++;			
		}
		
		//Count atoms of different (virtual) elements
		atomsPerElement = new int[maxNumElements];
		if (maxNumElements>1){
			for (int i = 0; i< atoms.size(); i++){
				atomsPerElement[atoms.get(i).getElement()]++;
			}
		} else atomsPerElement[0] = atoms.size();
		
		this.atoms.trimToSize();
		
		if (Configuration.getCrystalStructure().orderGrainsBySize()){
			ArrayList<Grain> sortedGrains = new ArrayList<Grain>(this.getGrains());
			Collections.sort(sortedGrains, new Comparator<Grain>() {
				@Override
				public int compare(Grain o1, Grain o2) {
					double diff = o1.getMesh().getVolume() - o2.getMesh().getVolume();
					if (diff<0.) return 1;
					if (diff>0.) return -1;
					return 0;
				}
			});
			
			for (int i=0; i<sortedGrains.size(); i++){
				sortedGrains.get(i).renumberGrain(i);
				Configuration.addGrainIndex(i);
				for (Atom a : sortedGrains.get(i).getAtomsInGrain()){
					a.setGrain(i);
				}
				sortedGrains.get(i).getAtomsInGrain().clear();	//Atoms in the grain are not needed anymore
			}
		}
	}
	
	/**
	 * Build a nearest neighbor graph and return the neighbor configuration
	 * @param atomsToPlot
	 * @return
	 */
	public StringBuilder plotNeighborsGraph(Atom... atomsToPlot){
		StringBuilder sb = new StringBuilder();
		final float d = Configuration.getCrystalStructure().getNearestNeighborSearchRadius();
		final NearestNeighborBuilder<Atom> nnb = new NearestNeighborBuilder<Atom>(this.boxSize, d, Configuration.getPbc());
		
		for (int i=0; i<atoms.size();i++)
			nnb.add(atoms.get(i));
		
		for (Atom a : atomsToPlot){
			ArrayList<Tupel<Atom, float[]>> t =nnb.getNeighAndNeighVec(a);
			sb.append("#Neighbor configuration ");
			sb.append(t.size());
			sb.append("\n");	
			sb.append("#number\t#dx\t#dy\t#dz\t#element");
			sb.append("\n");	
			for (int i=0; i< t.size(); i++){
				String s = String.format("%d\t%.6f\t%.6f\t%.6f\t%d", t.get(i).getO1().getNumber(),
						t.get(i).getO2()[0], t.get(i).getO2()[1], t.get(i).getO2()[2],
						t.get(i).getO1().getElement());
				sb.append(s);
				sb.append("\n");
			}
			
			int k=0;
			sb.append("#Bond angles");
			float[][] angles = new float[t.size()*(t.size()-1)/2][3];
			sb.append(angles.length);
			sb.append("\n");
			sb.append("#number1\t#number2\t#angle\n");
			for (int i=0; i<t.size(); i++){
				for (int j=i+1; j< t.size(); j++){
					angles[k][0] = (float)Math.toDegrees(VectorOps.angle(t.get(i).getO2(), t.get(j).getO2()));
					angles[k][1] = t.get(i).getO1().getNumber();
					angles[k][2] = t.get(j).getO1().getNumber();
					k++;
				}	
			}
			Arrays.sort(angles, new Comparator<float[]>() {
				@Override
				public int compare(float[] o1, float[] o2) {
					if (o1[0] < o2[0]) return 1;
					if (o1[0] == o2[0]) return 0;
					return -1;
				}
			});
			
			for (int i=0; i<angles.length; i++){
				String s = String.format("%d\t%d\t%.6f°", (int)angles[i][1], (int)angles[i][2],
						angles[i][0]);
				sb.append(s);
				sb.append("\n");
			}
			
			
		}
		return sb;
	}
	
	private void filterSurface(){
		findNearestNeighborsComplete(Configuration.getCrystalStructure().getNearestNeighborSearchRadius());

		final int surfaceType = Configuration.getCrystalStructure().getSurfaceType();
		final CyclicBarrier cb = new CyclicBarrier(ThreadPool.availProcessors());
		
		Vector<Callable<Void>> parallelTasks = new Vector<Callable<Void>>();
		for (int i=0; i<ThreadPool.availProcessors(); i++){
			final int j = i;
			parallelTasks.add(new Callable<Void>() {
				@Override
				public Void call() throws Exception {
					//parallel filtering of atoms in contact with the surface
					ArrayList<Atom> filteredAtoms = new ArrayList<Atom>();
					int start = (atoms.size() * j)/ThreadPool.availProcessors();
					int end = (atoms.size() * (j+1))/ThreadPool.availProcessors();

					for (int i=start; i<end; i++){
						Atom a = atoms.get(i);
						int count = 0;
						for (int j=0; j<a.getNeigh().length; j++){
							if (a.getType()!=surfaceType)
							if (a.getNeigh()[j].getType() == surfaceType) {
								count++;
							}
						}
						if (count>1) filteredAtoms.add(a);
					}
					//wait if all processes are finished with the identification of 
					//surface atoms, then relabel them and delete the RBV is applicable 
					cb.await();
					for (Atom a: filteredAtoms){
						a.setType(surfaceType);
						a.setRBV(null, null);
					}
					return null;
				}
			});
		}
		ThreadPool.executeParallel(parallelTasks);		
	}
	
	public float[] getBounds() {
		return boxSize;
	}
	
	public List<Atom> getAtoms(){
		return atoms;
	}
	
	public String getName() {
		return name;
	}
	
	public AtomData getNext() {
		return next;
	}
	
	public AtomData getPrevious() {
		return previous;
	}
	
	public float[] getOffset() {
		return offset;
	}
	
	public int getNumberOfType(int i){
		if (i>n_types.length) return 0;
		return n_types[i];
	}
	
	public int getNumAtomsOfElements(int i){
		if (i>=atomsPerElement.length) return 0;
		return atomsPerElement[i];
	}
	
	public Collection<Grain> getGrains() {
		return grains.values();
	}
	
	public Grain getGrains(int grain) {
		return grains.get(grain);
	}
	
	public DislocationDensityTensorData getDislocationDensityData() {
		return ddtd;
	}
	
	public boolean isGrainsImported() {
		return grainsImported;
	}
	
	public boolean isMeshImported() {
		return meshImported;
	}
	
	public void findNearestNeighbors(float d){
		findNearestNeighbors(d, defectNearestNeighborFilter);
	}
	
	public void findNearestNeighborsComplete(float d){
		findNearestNeighbors(d, nearestNeighborNonFiltered);
	}
	
	public void findNearestNeighbors(float d, final NearestNeighborFilter filter){
		if (this.lastNearestNeighborBuild.getO2()==d && this.lastNearestNeighborBuild.getO1().equals(filter)) return;
		
		final NearestNeighborBuilder<Atom> nnb = new NearestNeighborBuilder<Atom>(this.boxSize, d, Configuration.getPbc());
		
		for (int i=0; i<atoms.size();i++)
			if (filter.accept(atoms.get(i)))
				nnb.add(atoms.get(i));		
		//parallel build of nearest neighbors
		Vector<Callable<Void>> parallelTasks = new Vector<Callable<Void>>();
		for (int i=0; i<ThreadPool.availProcessors(); i++){
			final int j = i;
			parallelTasks.add(new Callable<Void>() {
				@Override
				public Void call() throws Exception {
					int start = (atoms.size() * j)/ThreadPool.availProcessors();
					int end = (atoms.size() * (j+1))/ThreadPool.availProcessors();

					for (int i=start; i<end; i++){
						if (Thread.interrupted()) return null;
						if (filter.accept(atoms.get(i)))
							atoms.get(i).buildNeigh(nnb);
						else atoms.get(i).removeNeigh();
					}
					return null;
				}
			});
		}
		ThreadPool.executeParallel(parallelTasks);
		
		this.lastNearestNeighborBuild.setO1(filter);
		this.lastNearestNeighborBuild.setO2(d);
	}
	
	public void removeNearestNeigbors(){
		this.lastNearestNeighborBuild.setO1(null);
		this.lastNearestNeighborBuild.setO2(0f);
		for (int i = 0; i< atoms.size(); i++){
			atoms.get(i).removeNeigh();
		}
	}
	
	public Skeletonizer getSkeletonizer(){
		if (skeletonizer == null)
			skeletonizer = new Skeletonizer(this);
		return skeletonizer;
	}
	
	public Object getFileMetaData(String s) {
		if (fileMetaData == null) return null;
		return fileMetaData.get(s);
	}

	/**
	 * Frees the atom list to make more memory available and removes references to the skeletonizer
	 * and next and previous AtomDatas
	 * Helps if you want to load another large file, without memory shortcomings 
	 */
	public void clear(){
		maxNumElements = 1;
		this.atoms.clear();
		this.skeletonizer = null;
		this.next = null;
		this.previous = null;
		this.ddtd = null;
	}
	
	public boolean isRbvAvailable() {
		return rbvAvailable;
	}
	
	public void addAdditionalData(DataContainer dc){
		dc = dc.deriveNewInstance();
		
		boolean success = false;
		try {
			if (dc.isExternalFileRequired()) {
				File f = dc.selectFile(Configuration.getLastOpenedFolder());
				if (f == null) return;
				success = dc.processData(f, this);
			} else success = dc.processData(null, this);
		} catch (IOException e) {
			// TODO proper handling
			e.printStackTrace();
		}
		if (success) {
			AtomData a = this;
//			while (a!=null){
				boolean replaced = false;
				for (int i=0; i<a.additionalData.size(); i++){
					if (a.additionalData.get(i).getClass() == dc.getClass()){
						a.additionalData.set(i,dc);
						replaced = true;
						break;
					}
				}
				if (!replaced) a.additionalData.add(dc);
//				a = a.next;
//			}
			
			
//			boolean replaced = false;
//			for (int i=0; i<this.additionalData.size(); i++){
//				if (this.additionalData.get(i).getClass() == dc.getClass()){
//					this.additionalData.set(i,dc);
//					replaced = true;
//					break;
//				}
//			}
//			if (!replaced) this.additionalData.add(dc);
		}
		else JOptionPane.showMessageDialog(null, "Creating additional data failed");
	}
	
	public void printToFile(File out, boolean binary, boolean compressed) throws IOException{
		DataOutputStream dos;
		if (compressed)
			dos = new DataOutputStream(new BufferedOutputStream(new GZIPOutputStream(new FileOutputStream(out))));
		else dos = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(out)));
		
		int numData = 1+Configuration.getSizeCustomColumns()+(ImportStates.POLY_MATERIAL.isActive()?1:0);
		
		if (binary){
			dos.writeBytes(String.format("#F b 1 1 0 3 0 %d\n", numData));
		} else dos.writeBytes(String.format("#F A 1 1 0 3 0 %d\n", numData));
		dos.writeBytes("#C number type x y z ada_type");
		
		if (Configuration.getSizeCustomColumns()!=0){
			for (int i=0; i<Configuration.getSizeCustomColumns(); i++){
				String id = " "+Configuration.getCustomColumnInfo(i).getId();
				dos.writeBytes(id);
			}
		}
		if (ImportStates.POLY_MATERIAL.isActive())
			dos.writeBytes(" grain");
		
		//rbv always at the end
		if (rbvAvailable)
			dos.writeBytes(" rbv_data");
		dos.writeBytes("\n");
		
		dos.writeBytes("#X "+boxSize[0]+" 0.0 0.0\n");
		dos.writeBytes("#Y 0.0 "+boxSize[1]+" 0.0\n");
		dos.writeBytes("#Z 0.0 0.0 "+boxSize[2]+"\n");
		
		dos.writeBytes("##META\n");
		if (ImportStates.POLY_MATERIAL.isActive()){
			for (Grain g: getGrains()){
				float[][] rot = g.getCystalRotationTools().getDefaultRotationMatrix();
				dos.writeBytes(String.format("##grain %d %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f %.4f\n",
						g.getGrainNumber(), rot[0][0], rot[1][0], rot[2][0]
						, rot[0][1], rot[1][1], rot[2][1]
						, rot[0][2], rot[1][2], rot[2][2]));
				dos.writeBytes(String.format("##atomsInGrain %d %d\n",
						g.getGrainNumber(), g.getNumberOfAtoms()));
				g.getMesh().getFinalMesh().printMetaData(dos, g.getGrainNumber());
			}
		}
		if (fileMetaData != null && fileMetaData.containsKey("extpot")){
			float[] indent = (float[]) fileMetaData.get("extpot");
			dos.writeBytes(String.format("##extpot %.4f %.4f %.4f %.4f %.4f\n", indent[0], indent[1], indent[2], indent[3], indent[5]));
		}
		if (fileMetaData != null && fileMetaData.containsKey("wall")){
			float[] wall = (float[]) fileMetaData.get("wall");
			dos.writeBytes(String.format("##wall %.4f %.4f %.4f\n", wall[0], wall[1], wall[2]));
		}
		if (fileMetaData != null && fileMetaData.containsKey("timestep")){
			float[] timestep = (float[])fileMetaData.get("timestep");
			dos.writeBytes(String.format("##timestep %f\n", timestep[0]));
		}
		
		dos.writeBytes("##METAEND\n");
		dos.writeBytes("#E\n");
		
		if (binary){
			for (Atom a : atoms){
				//enable highest filtering
//				if ( (a.getRBV() == null || a.getRBV().getSquaredRBVLengh()<0.36) && a.getType()!=6) continue;
//				if (a.getRBV() == null && a.getType()!=6) continue;
				
				dos.writeInt(a.getNumber()); dos.writeInt(a.getElement());
				dos.writeFloat(a.getX()); dos.writeFloat(a.getY()); dos.writeFloat(a.getZ());
				dos.writeInt(a.getType());
				
				if (Configuration.getSizeCustomColumns()!=0)
					for (int i=0; i<Configuration.getSizeCustomColumns();i++)
						dos.writeFloat(a.getCustom(i));
				if (ImportStates.POLY_MATERIAL.isActive())
					dos.writeInt(a.getGrain());
				
				if (rbvAvailable){
					if (a.getRBV()!=null){
						dos.writeInt(1);
						dos.writeFloat(a.getRBV().lineDirection_x); dos.writeFloat(a.getRBV().lineDirection_y); dos.writeFloat(a.getRBV().lineDirection_z);
						dos.writeFloat(a.getRBV().rbv_x); dos.writeFloat(a.getRBV().rbv_y); dos.writeFloat(a.getRBV().rbv_z);
					} else {
						dos.writeInt(0);
					}
				}
			}
		} else {
			for (Atom a : atoms){
				dos.writeBytes(String.format("%d %d %.8f %.8f %.8f %d", a.getNumber(), a.getElement(), 
						a.getX(), a.getY(), a.getZ(), a.getType()));
				
				if (Configuration.getSizeCustomColumns()!=0)
					for (int i=0; i<Configuration.getSizeCustomColumns();i++)
						dos.writeBytes(String.format(" %.8f",a.getCustom(i)));
				
				if (ImportStates.POLY_MATERIAL.isActive())
					dos.writeBytes(" "+Integer.toString(a.getGrain()));
				
				if (rbvAvailable){
					if (a.getRBV()==null){
						dos.writeBytes(" 0");
					} else {
						dos.writeBytes(String.format(" 1 %.4f %.4f %.4f %.4f %.4f %.4f",
								a.getRBV().lineDirection_x, a.getRBV().lineDirection_y, a.getRBV().lineDirection_z,
								a.getRBV().rbv_x, a.getRBV().rbv_y, a.getRBV().rbv_z));
					}
				}
				
				dos.writeBytes("\n");
			}
		}
		dos.close();
	}
}