// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.Callable;

import model.polygrain.Grain;

import Jama.Matrix;
import Jama.SingularValueDecomposition;

import common.MatrixOps;
import common.ThreadPool;
import common.VectorOps;

import crystalStructures.CrystalStructure;

public class LatticeRotationBuilder {
	
	public static void calculateLatticeRotation(final AtomData data) {
//		final double PHI_MAX = Math.cos(30*Math.PI/180.);
		
		final HashMap<Integer, float[][]> p = new HashMap<Integer, float[][]>();
		final HashMap<Integer, float[]> p_l = new HashMap<Integer, float[]>();
		
		CrystalStructure cs = Configuration.getCrystalStructure();
		float[][] rot = Configuration.getCrystalRotationTools().getDefaultRotationMatrix();
		double[][] perf_0 = cs.getPerfectNearestNeighborsUnrotated();
		
		//Adding the default grain orientation
		float[][] p_0 = new float[perf_0.length][3];
		for (int i=0; i<p_0.length; i++){
			float[] d = new float[3];
			d[0] = (float)(perf_0[i][0]*rot[0][0] + perf_0[i][1]*rot[1][0] + perf_0[i][2]*rot[2][0]);
			d[1] = (float)(perf_0[i][0]*rot[0][1] + perf_0[i][1]*rot[1][1] + perf_0[i][2]*rot[2][1]);
			d[2] = (float)(perf_0[i][0]*rot[0][2] + perf_0[i][1]*rot[1][2] + perf_0[i][2]*rot[2][2]);
			p_0[i][0] = d[0] * Configuration.getCrystalStructure().getLatticeConstant();
			p_0[i][1] = d[1] * Configuration.getCrystalStructure().getLatticeConstant();
			p_0[i][2] = d[2] * Configuration.getCrystalStructure().getLatticeConstant();
		}
		float[] p_l_0 = new float[p_0.length];
		for (int i=0; i<p_0.length; i++) p_l_0[i] = VectorOps.length(p_0[i]);
		p.put(Atom.DEFAULT_GRAIN, p_0);
		p_l.put(Atom.DEFAULT_GRAIN, p_l_0);
		
		//Add different grains for polycrystals
		if (data.getGrains()!=null){
			for (Grain g : data.getGrains()){
				rot = g.getCystalRotationTools().getDefaultRotationMatrix();
				perf_0 = g.getCrystalStructure().getPerfectNearestNeighborsUnrotated();
				
				p_0 = new float[perf_0.length][3];
				for (int i=0; i<p_0.length; i++){
					float[] d = new float[3];
					d[0] = (float)(perf_0[i][0]*rot[0][0] + perf_0[i][1]*rot[1][0] + perf_0[i][2]*rot[2][0]);
					d[1] = (float)(perf_0[i][0]*rot[0][1] + perf_0[i][1]*rot[1][1] + perf_0[i][2]*rot[2][1]);
					d[2] = (float)(perf_0[i][0]*rot[0][2] + perf_0[i][1]*rot[1][2] + perf_0[i][2]*rot[2][2]);
					p_0[i][0] = d[0] * Configuration.getCrystalStructure().getLatticeConstant();
					p_0[i][1] = d[1] * Configuration.getCrystalStructure().getLatticeConstant();
					p_0[i][2] = d[2] * Configuration.getCrystalStructure().getLatticeConstant();
				}
				p_l_0 = new float[p_0.length];
				for (int i=0; i<p_0.length; i++) p_l_0[i] = VectorOps.length(p_0[i]);
				p.put(g.getGrainNumber(), p_0);
				p_l.put(g.getGrainNumber(), p_l_0);
			}
		}
		
		final NearestNeighborBuilder<Atom> nnb = new NearestNeighborBuilder<Atom>(data.getBounds(), 
				cs.getNearestNeighborSearchRadius(), Configuration.getPbc());
		for (int i=0; i<data.getAtoms().size(); i++){
			nnb.add(data.getAtoms().get(i));
		}
		
		Vector<Callable<Void>> parallelTasks = new Vector<Callable<Void>>();
		for (int k=0; k<ThreadPool.availProcessors(); k++){
			final int l = k;
			parallelTasks.add(new Callable<Void>() {
				@Override
				public Void call() throws Exception {
					List<Atom> atoms = data.getAtoms();
					int start = (atoms.size() * l)/ThreadPool.availProcessors();
					int end = (atoms.size() * (l+1))/ThreadPool.availProcessors();

					double[][] lcm = new double[3][3];
					double[] a = new double[9];
					double[] b = new double[9];
					Matrix lcmMatrix = new Matrix(lcm);
					
					for (int k=start; k<end; k++){
						if (Thread.interrupted()) return null;
						
						for (int i=0;i<9;i++){
							a[i] = 0f; b[i] = 0f;
						}
						
						Atom atom = atoms.get(k);
						int grain = atom.getGrain();
						if (grain == Atom.IGNORED_GRAIN) continue;
						
						//Pick the current grain data
						float[][] p_0 = p.get(grain);
						float[] p_l_0 = p_l.get(grain);
						
						ArrayList<float[]> neighVecList = nnb.getNeighVec(atom);

						for (int i=0; i<neighVecList.size(); i++){
							float bestAngle = -1;
							int best = 0;
							float[] n = neighVecList.get(i);
							float l = VectorOps.length(n);
							for (int j=0; j<p_0.length; j++){
								float angle = VectorOps.product(n, p_0[j]) / (l* p_l_0[j]);
								if (angle>bestAngle){
									best = j;
									bestAngle = angle;
								}
							}
							//if (bestAngle>PHI_MAX){
								a[0] += n[0] * p_0[best][0]; a[1] += n[0] * p_0[best][1]; a[2] += n[0] * p_0[best][2];
								a[3] += n[1] * p_0[best][0]; a[4] += n[1] * p_0[best][1]; a[5] += n[1] * p_0[best][2];
								a[6] += n[2] * p_0[best][0]; a[7] += n[2] * p_0[best][1]; a[8] += n[2] * p_0[best][2];
								
								b[0] += n[0] * n[0]; b[1] += n[0] * n[1]; b[2] += n[0] * n[2];
								b[3] += n[1] * n[0]; b[4] += n[1] * n[1]; b[5] += n[1] * n[2];
								b[6] += n[2] * n[0]; b[7] += n[2] * n[1]; b[8] += n[2] * n[2];
//							}
						}
						
						if (MatrixOps.invert3x3matrix(a, 0.00001)){
							lcm[0][0] = a[0] * b[0] + a[1] * b[3] +a[2] * b[6];
							lcm[1][0] = a[0] * b[1] + a[1] * b[4] +a[2] * b[7];
							lcm[2][0] = a[0] * b[2] + a[1] * b[5] +a[2] * b[8];
							         
							lcm[0][1] = a[3] * b[0] + a[4] * b[3] +a[5] * b[6];
							lcm[1][1] = a[3] * b[1] + a[4] * b[4] +a[5] * b[7];
							lcm[2][1] = a[3] * b[2] + a[4] * b[5] +a[5] * b[8];
							         
							lcm[0][2] = a[6] * b[0] + a[7] * b[3] +a[8] * b[6];
							lcm[1][2] = a[6] * b[1] + a[7] * b[4] +a[8] * b[7];
							lcm[2][2] = a[6] * b[2] + a[7] * b[5] +a[8] * b[8];
						}
						else continue;
						
						SingularValueDecomposition svd = new SingularValueDecomposition(lcmMatrix);
						Matrix r = svd.getU().times(svd.getV().transpose());
						if (r!=null){
							float[][] foo = new float[3][3];
							foo[0][0] = (float)r.get(0, 0); foo[0][1] = (float)r.get(0, 1); foo[0][2] = (float)r.get(0, 2);
							foo[1][0] = (float)r.get(1, 0); foo[1][1] = (float)r.get(1, 1); foo[1][2] = (float)r.get(1, 2);
							foo[2][0] = (float)r.get(2, 0); foo[2][1] = (float)r.get(2, 1); foo[2][2] = (float)r.get(2, 2);
							atom.setEulerAngles(MatrixOps.getEulerAngles(foo));
						}
					}
					return null;
				}
			});
		}
		ThreadPool.executeParallel(parallelTasks);
	}
}
