// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model;

import java.util.*;

import common.Tupel;

public class NearestNeighborBuilder<T extends Coordinate> {
	private float[] cellSize = new float[3];
	private float invCellSizeX, invCellSizeY, invCellSizeZ;
	private float boxSizeX, boxSizeY, boxSizeZ;
	private int dimX, dimY, dimZ;
	private boolean pbcX, pbcY, pbcZ;
	
	private float sqrMinDist; 
	
	private int numberOfElements = 0;
	
	private ArrayList<T>[][][] cells;
	
	
	@SuppressWarnings("unchecked")
	public NearestNeighborBuilder(float[] boxSize, float minDist, boolean pbc[]) {
		this.boxSizeX = boxSize[0];
		this.boxSizeY = boxSize[1];
		this.boxSizeZ = boxSize[2];
		
		this.sqrMinDist = minDist*minDist;
		dimX = (int)(boxSizeX / minDist);
		if (dimX==0) dimX=1;
		dimY = (int)(boxSizeY / minDist);
		if (dimY==0) dimY=1;
		dimZ = (int)(boxSizeZ / minDist);
		if (dimZ==0) dimZ=1;
		
		
		cellSize[0] = boxSize[0] / (float)dimX;
		cellSize[1] = boxSize[1] / (float)dimY;
		cellSize[2] = boxSize[2] / (float)dimZ;
		
		invCellSizeX = 1f/cellSize[0];
		invCellSizeY = 1f/cellSize[1];
		invCellSizeZ = 1f/cellSize[2];
		         
		cells = new ArrayList[dimX][dimY][dimZ];
		
		this.pbcX = pbc[0];
		this.pbcY = pbc[1];
		this.pbcZ = pbc[2];
	}

	public int getNumberOfElements() {
		return getNumberOfElements();
	}

	public void add(T c){
		int x = (int)(c.getX()*invCellSizeX);
		int y = (int)(c.getY()*invCellSizeY);
		int z = (int)(c.getZ()*invCellSizeZ);
		
		if (x<0) x=0;
		if (y<0) y=0;
		if (z<0) z=0;
		if (x>=dimX) x=dimX-1;
		if (y>=dimY) y=dimY-1;
		if (z>=dimZ) z=dimZ-1;
		
		if (cells[x][y][z] == null){
			cells[x][y][z] = new ArrayList<T>(5);
		}
		cells[x][y][z].add(c);
		numberOfElements++;
		
	}
	
	public boolean remove(T c) {
		int x = (int)(c.getX()*invCellSizeX);
		int y = (int)(c.getY()*invCellSizeY);
		int z = (int)(c.getZ()*invCellSizeZ);
		
		if (x<0) x=0;
		if (y<0) y=0;
		if (z<0) z=0;
		if (x>=dimX) x=dimX-1;
		if (y>=dimY) y=dimY-1;
		if (z>=dimZ) z=dimZ-1;
		
		if (cells[x][y][z] != null){
			if (cells[x][y][z].remove(c)){
				numberOfElements--;
				return true;
			}
		}
		
		return false;
	};
	
	public ArrayList<T> getNeigh(T c){
		int x = (int)(c.getX()*invCellSizeX);
		int y = (int)(c.getY()*invCellSizeY);
		int z = (int)(c.getZ()*invCellSizeZ);
		
		if (x<0) x=0;
		if (y<0) y=0;
		if (z<0) z=0;
		if (x>=dimX) x=dimX-1;
		if (y>=dimY) y=dimY-1;
		if (z>=dimZ) z=dimZ-1;
		
		ArrayList<T> neigh = new ArrayList<T>(15);
		
		for (int i=-1; i<=1; i++){
			for (int j=-1; j<=1; j++){
				for (int k=-1; k<=1; k++){
					ArrayList<T> possibleNeigh = accessCell(x+i, y+j, z+k);
					if (possibleNeigh!=null){
						for (int l=0; l<possibleNeigh.size(); l++){
							T s = possibleNeigh.get(l);
							if (s!=c){
								float tx = s.getX(); 
								float ty = s.getY();
								float tz = s.getZ();
								if (pbcX){
									if (x+i<0) tx -= this.boxSizeX;
									else if (x+i>=this.dimX) tx += this.boxSizeX;
								}
								
								if (pbcY){
									if (y+j<0) ty -= this.boxSizeY;
									else if (y+j>=this.dimY) ty += this.boxSizeY;
								}
								
								if (pbcZ){
									if (z+k<0) tz -= this.boxSizeZ;
									else if (z+k>=this.dimZ) tz += this.boxSizeZ;
								}
								
								float distance = (c.getX()-tx)*(c.getX()-tx)
									+ (c.getY()-ty)*(c.getY()-ty)
									+ (c.getZ()-tz)*(c.getZ()-tz);
								if (distance<=sqrMinDist) neigh.add(s);
							}
						}
					}
				}	
			}	
		}
		
		return neigh;
	}
	
	public ArrayList<float[]> getNeighVec(T c){
		int x = (int)(c.getX()*invCellSizeX);
		int y = (int)(c.getY()*invCellSizeY);
		int z = (int)(c.getZ()*invCellSizeZ);
		
		if (x<0) x=0;
		if (y<0) y=0;
		if (z<0) z=0;
		if (x>=dimX) x=dimX-1;
		if (y>=dimY) y=dimY-1;
		if (z>=dimZ) z=dimZ-1;
		
		ArrayList<float[]> neigh = new ArrayList<float[]>(15);
		
		for (int i=-1; i<=1; i++){
			for (int j=-1; j<=1; j++){
				for (int k=-1; k<=1; k++){
					ArrayList<T> possibleNeigh = accessCell(x+i, y+j, z+k);
					if (possibleNeigh!=null){
						for (int l=0; l<possibleNeigh.size(); l++){
							T s = possibleNeigh.get(l);
							if (s!=c){
								float tx = s.getX(); 
								float ty = s.getY();
								float tz = s.getZ();
								if (pbcX){
									if (x+i<0) tx -= this.boxSizeX;
									else if (x+i>=this.dimX) tx += this.boxSizeX;
								}
								
								if (pbcY){
									if (y+j<0) ty -= this.boxSizeY;
									else if (y+j>=this.dimY) ty += this.boxSizeY;
								}
								
								if (pbcZ){
									if (z+k<0) tz -= this.boxSizeZ;
									else if (z+k>=this.dimZ) tz += this.boxSizeZ;
								}
								float distance = (c.getX()-tx)*(c.getX()-tx)
									+ (c.getY()-ty)*(c.getY()-ty)
									+ (c.getZ()-tz)*(c.getZ()-tz);
								if (distance<=sqrMinDist) neigh.add(new float[]{c.getX()-tx, c.getY()-ty, c.getZ()-tz});
							}
						}
					}
				}	
			}	
		}
		
		return neigh;
	}
	
	public ArrayList<Tupel<T, float[]>> getNeighAndNeighVec(T c){
		int x = (int)(c.getX()*invCellSizeX);
		int y = (int)(c.getY()*invCellSizeY);
		int z = (int)(c.getZ()*invCellSizeZ);
		
		if (x<0) x=0;
		if (y<0) y=0;
		if (z<0) z=0;
		if (x>=dimX) x=dimX-1;
		if (y>=dimY) y=dimY-1;
		if (z>=dimZ) z=dimZ-1;
		
		ArrayList<Tupel<T, float[]>> neigh = new ArrayList<Tupel<T, float[]>>(15);
		
		for (int i=-1; i<=1; i++){
			for (int j=-1; j<=1; j++){
				for (int k=-1; k<=1; k++){
					ArrayList<T> possibleNeigh = accessCell(x+i, y+j, z+k);
					if (possibleNeigh!=null){
						for (int l=0; l<possibleNeigh.size(); l++){
							T s = possibleNeigh.get(l);
							if (s!=c){
								float tx = s.getX(); 
								float ty = s.getY();
								float tz = s.getZ();
								if (pbcX){
									if (x+i<0) tx -= this.boxSizeX;
									else if (x+i>=this.dimX) tx += this.boxSizeX;
								}
								
								if (pbcY){
									if (y+j<0) ty -= this.boxSizeY;
									else if (y+j>=this.dimY) ty += this.boxSizeY;
								}
								
								if (pbcZ){
									if (z+k<0) tz -= this.boxSizeZ;
									else if (z+k>=this.dimZ) tz += this.boxSizeZ;
								}
								
								float distance = (c.getX()-tx)*(c.getX()-tx)
									+ (c.getY()-ty)*(c.getY()-ty)
									+ (c.getZ()-tz)*(c.getZ()-tz);
								if (distance<=sqrMinDist) neigh.add(
										new Tupel<T, float[]>(s, new float[]{c.getX()-tx, c.getY()-ty, c.getZ()-tz}));
							}
						}
					}
				}	
			}	
		}
		
		return neigh;
	}
	
	private ArrayList<T> accessCell(int x, int y, int z){
		if (x>=0 && y>=0 && z>=0 && x<dimX && y<dimY && z<dimZ)
			return cells[x][y][z];
		
		if (pbcX && x>=dimX) x -= dimX;
		if (pbcX && x<0) x += dimX;
		if (pbcY && y>=dimY) y -= dimY;
		if (pbcY && y<0) y += dimY;
		if (pbcZ && z>=dimZ) z -= dimZ;	
		if (pbcZ && z<0) z += dimZ;
		
		if (x>=0 && y>=0 && z>=0 && x<dimX && y<dimY && z<dimZ)
			return cells[x][y][z];
		return null;
	}
}
