// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.dislocationDensity;

import java.util.ArrayList;

import model.*;

import common.MatrixOps;
import common.Tupel;
import common.VectorOps;
import crystalStructures.CrystalStructure;

public class CuboidSectorDensityTensorBuilder {
	
	private final double[][] neighPerf;
	private final double[] pnl;
	
	private final boolean pbc_x, pbc_y, pbc_z;
	private float[] bounds;
	private float halfBound_x, halfBound_y, halfBound_z;
	
	private AtomData data;
	private int rasterX, rasterY, rasterZ;
	
	
	public CuboidSectorDensityTensorBuilder(AtomData data, int rasterX, int rasterY, int rasterZ) {
		this.data = data;
		
		this.pbc_x = Configuration.getPbc()[0];
		this.pbc_y = Configuration.getPbc()[1];
		this.pbc_z = Configuration.getPbc()[2];
		
		this.rasterX = rasterX;
		this.rasterY = rasterY;
		this.rasterZ = rasterZ;
		
		this.bounds = data.getBounds();
		
		this.halfBound_x = bounds[0]*0.5f;
		this.halfBound_y = bounds[1]*0.5f;
		this.halfBound_z = bounds[2]*0.5f;

		
		CrystalStructure cs = Configuration.getCrystalStructure();
		float[][] rot = Configuration.getCrystalRotationTools().getDefaultRotationMatrix();
		double[][] perf = cs.getPerfectNearestNeighborsUnrotated();
		neighPerf = new double[perf.length][3];
		for (int i=0; i<neighPerf.length; i++){
			double[] d = new double[3];
			d[0] = (perf[i][0]*rot[0][0] + perf[i][1]*rot[1][0] + perf[i][2]*rot[2][0]);
			d[1] = (perf[i][0]*rot[0][1] + perf[i][1]*rot[1][1] + perf[i][2]*rot[2][1]);
			d[2] = (perf[i][0]*rot[0][2] + perf[i][1]*rot[1][2] + perf[i][2]*rot[2][2]);
			neighPerf[i][0] = d[0] * Configuration.getCrystalStructure().getLatticeConstant();
			neighPerf[i][1] = d[1] * Configuration.getCrystalStructure().getLatticeConstant();
			neighPerf[i][2] = d[2] * Configuration.getCrystalStructure().getLatticeConstant();
		}

		this.pnl = new double[neighPerf.length];
		for (int i=0; i<this.pnl.length; i++){
			this.pnl[i] = VectorOps.length(neighPerf[i]);
		}
	}
	
	
	
	@SuppressWarnings("unchecked")
	public DislocationDensityTensor[][][] createCuboids() {
		NearestNeighborBuilder<Coordinate> nnb = new NearestNeighborBuilder<Coordinate>(data.getBounds(), 
				Configuration.getCrystalStructure().getNearestNeighborSearchRadius(), Configuration.getPbc());
		for (int i=0; i<data.getAtoms().size(); i++){
			nnb.add(data.getAtoms().get(i));
		}
		ArrayList<Tupel<float[], double[][]>> points = new ArrayList<Tupel<float[],double[][]>>();
		
		
		Tupel<float[],double[][]>[][][] raster = new Tupel[rasterX+1][rasterY+1][rasterZ+1];
		
		for (int i=0; i<=rasterX; ++i){
			for (int j=0; j<=rasterY; ++j){
				for (int k=0; k<=rasterZ; ++k){
					float[] p = new float[3];
					p[0] = (bounds[0] / rasterX) * i;
					p[1] = (bounds[1] / rasterY) * j;
					p[2] = (bounds[2] / rasterZ) * k;
					Tupel<float[], double[][]> tupel = new Tupel<float[], double[][]>(p, null); 
					raster[i][j][k] = tupel;
					points.add(tupel);
				}
			}	
		}

		//Create cuboids
		Cuboid[][][] cuboids = new CuboidSectorDensityTensorBuilder.Cuboid[rasterX][rasterY][rasterZ];
		
		float xFrac = bounds[0] / rasterX;
		float yFrac = bounds[1] / rasterY;
		float zFrac = bounds[2] / rasterZ;
		
		for (int i=0; i<rasterX; ++i){
			for (int j=0; j<rasterY; ++j){
				for (int k=0; k<rasterZ; ++k){
					float[] p = new float[3];
					p[0] = (xFrac) * (i+0.5f);
					p[1] = (yFrac) * (j+0.5f);
					p[2] = (zFrac) * (k+0.5f);
					Tupel<float[], double[][]> tupel = new Tupel<float[], double[][]>(p, null); 
					points.add(tupel);
					Tupel<float[], double[][]>[] corners = new Tupel[8];
					corners[0] = raster[i+0][j+0][k+0];
					corners[1] = raster[i+0][j+0][k+1];
					corners[2] = raster[i+0][j+1][k+0];
					corners[3] = raster[i+0][j+1][k+1];
					corners[4] = raster[i+1][j+0][k+0];
					corners[5] = raster[i+1][j+0][k+1];
					corners[6] = raster[i+1][j+1][k+0];
					corners[7] = raster[i+1][j+1][k+1];
					cuboids[i][j][k] = new Cuboid(corners, tupel);
				}
			}	
		}

		
		for (Tupel<float[], double[][]> t : points){
			double[][] dm = getDeformationMatrix(t.getO1(), nnb);
			t.setO2(dm);
		}
		
		DislocationDensityTensor[][][] cuboidsDens = new DislocationDensityTensor[rasterX][rasterY][rasterZ];
		
		for (int i = 0; i < rasterX; ++i) {
			for (int j = 0; j < rasterY; ++j) {
				for (int k = 0; k < rasterZ; ++k) {
					Cuboid c = cuboids[i][j][k];
					Tupel<float[], double[][]>[] cornersExt = new Tupel[14];
					for (int l = 0; l < 8; l++)
						cornersExt[l] = c.corners[l];

					float[] ce = c.center.getO1();

					float[] p = { ce[0] + xFrac, ce[1], ce[2] };
					cornersExt[8] = new Tupel<float[], double[][]>(p, getDeformationMatrix(p, nnb));
					p = new float[] { ce[0] - xFrac, ce[1], ce[2] };
					cornersExt[9] = new Tupel<float[], double[][]>(p, getDeformationMatrix(p, nnb));
					p = new float[] { ce[0], ce[1] + yFrac, ce[2] };
					cornersExt[10] = new Tupel<float[], double[][]>(p, getDeformationMatrix(p, nnb));
					p = new float[] { ce[0], ce[1] - yFrac, ce[2] };
					cornersExt[11] = new Tupel<float[], double[][]>(p, getDeformationMatrix(p, nnb));
					p = new float[] { ce[0], ce[1], ce[2] + zFrac };
					cornersExt[12] = new Tupel<float[], double[][]>(p, getDeformationMatrix(p, nnb));
					p = new float[] { ce[0], ce[1], ce[2] - zFrac };
					cornersExt[13] = new Tupel<float[], double[][]>(p, getDeformationMatrix(p, nnb));

					double[][] nye = calculateNye(c.center, cornersExt);
					// double[][] nye = calculateNye(c.center, c.corners);
					CuboidVolumeElement cs = new CuboidVolumeElement(c.corners[7].getO1(), c.corners[0].getO1());
					cuboidsDens[i][j][k] = new DislocationDensityTensor(nye, cs);
				}
			}
		}
		
		return cuboidsDens;
	}
	
	private double[][] getDeformationMatrix(float[] p, NearestNeighborBuilder<Coordinate> nnb){
		CoordinateWrapper cw = new CoordinateWrapper(p);
		ArrayList<Coordinate> neigh = nnb.getNeigh(cw);
		
		double[][] avDeform = new double[3][3];
		for (Coordinate c : neigh){
			double[][] d = getDeformationMatrix(nnb.getNeighVec(c));
//			float[][] d = CrystalRotationTools.getRotationMatrixFromEulerAngles(((Atom)c).getEulerAngles());
			
			for (int i=0; i<3; i++){
				for (int j=0; j<3; j++){
					avDeform[i][j] += d[i][j];
				}	
			}
		}
		
		if (neigh.size()>0)
			for (int i=0; i<3; i++){
				for (int j=0; j<3; j++){
					avDeform[i][j] /= neigh.size();
				}	
			}
		else avDeform = new double[][]{{1,0,0},{0,1,0},{0,0,1}};
		
		return avDeform;
	}
	
	
	private double[][] getDeformationMatrix(ArrayList<float[]> neighVec){
		double[][] lcm = new double[3][3];
		float[][] neigh = new float[neighVec.size()][3];
		
		for (int i=0; i<neighVec.size(); i++){
			neigh[i] = neighVec.get(i); 
		}
				
		double[] a = new double[9];
		double[][] b = new double[3][3];

		for (int i=0; i<neigh.length; i++){
			double bestAngle = -1;
			int best = 0;
			float[] n = neigh[i];
			double l = VectorOps.length(n);
			for (int j=0; j<neighPerf.length; j++){
				double[] v = neighPerf[j];
				double angle = (n[0]*v[0]+n[1]*v[1]+n[2]*v[2]) / (l* pnl[j]);
				if (angle>bestAngle){
					best = j;
					bestAngle = angle;
				}
			}
			if (bestAngle>Math.cos(20*Math.PI/180.)){	
			a[0] += n[0] * neighPerf[best][0]; a[1] += n[0] * neighPerf[best][1]; a[2] += n[0] * neighPerf[best][2];
			a[3] += n[1] * neighPerf[best][0]; a[4] += n[1] * neighPerf[best][1]; a[5] += n[1] * neighPerf[best][2];
			a[6] += n[2] * neighPerf[best][0]; a[7] += n[2] * neighPerf[best][1]; a[8] += n[2] * neighPerf[best][2];
			
			b[0][0] += n[0] * n[0]; b[0][1] += n[0] * n[1]; b[0][2] += n[0] * n[2];
			b[1][0] += n[1] * n[0]; b[1][1] += n[1] * n[1]; b[1][2] += n[1] * n[2];
			b[2][0] += n[2] * n[0]; b[2][1] += n[2] * n[1]; b[2][2] += n[2] * n[2];
			}
		}
		
		if (MatrixOps.invert3x3matrix(a, 0.001)){
			lcm = new double[3][3];
			lcm[0][0] = a[0] * b[0][0] + a[1] * b[1][0] +a[2] * b[2][0];
			lcm[0][1] = a[0] * b[0][1] + a[1] * b[1][1] +a[2] * b[2][1];
			lcm[0][2] = a[0] * b[0][2] + a[1] * b[1][2] +a[2] * b[2][2];
			
			lcm[1][0] = a[3] * b[0][0] + a[4] * b[1][0] +a[5] * b[2][0];
			lcm[1][1] = a[3] * b[0][1] + a[4] * b[1][1] +a[5] * b[2][1];
			lcm[1][2] = a[3] * b[0][2] + a[4] * b[1][2] +a[5] * b[2][2];
			
			lcm[2][0] = a[6] * b[0][0] + a[7] * b[1][0] +a[8] * b[2][0];
			lcm[2][1] = a[6] * b[0][1] + a[7] * b[1][1] +a[8] * b[2][1];
			lcm[2][2] = a[6] * b[0][2] + a[7] * b[1][2] +a[8] * b[2][2];
		}
		else lcm = new double[][]{{1.,0.,0.},{0.,1.,0.},{0.,0.,1.}};
		return lcm;
	}
	
	
	private double[][] calculateNye(Tupel<float[], double[][]> central, Tupel<float[], double[][]>[] inter ){
		double[][] neigh = new double[inter.length][3];
		
		for (int i=0; i<inter.length; i++){
			neigh[i][0] = inter[i].getO1()[0] - central.getO1()[0]; 
			neigh[i][1] = inter[i].getO1()[1] - central.getO1()[1];
			neigh[i][2] = inter[i].getO1()[2] - central.getO1()[2];
			
			if (pbc_x){
				if (neigh[i][0] < -halfBound_x)
					neigh[i][0] += bounds[0];
				else if (neigh[i][0] > halfBound_x)
					neigh[i][0] -= bounds[0];
			}
			if (pbc_y){
				if (neigh[i][1] < -halfBound_y)
					neigh[i][1] += bounds[1];
				else if (neigh[i][1] > halfBound_y)
					neigh[i][1] -= bounds[1];
			}
			if (pbc_z){
				if (neigh[i][2] < -halfBound_z)
					neigh[i][2] += bounds[2];
				else if (neigh[i][2] > halfBound_z)
					neigh[i][2] -= bounds[2];
			}
		}
		
		double[][][] neighT = new double[inter.length][3][3];
		for (int i=0; i<inter.length; i++){
			//neighT[i]= neigh[i].transpose().times(neigh[i]);
			
			neighT[i][0][0] = neigh[i][0] * neigh[i][0]; 
			neighT[i][0][1] = neigh[i][0] * neigh[i][1]; 
			neighT[i][0][2] = neigh[i][0] * neigh[i][2];
			
			neighT[i][1][0] = neigh[i][1] * neigh[i][0]; 
			neighT[i][1][1] = neigh[i][1] * neigh[i][1];
			neighT[i][1][2] = neigh[i][1] * neigh[i][2];
			
			neighT[i][2][0] = neigh[i][2] * neigh[i][0];
			neighT[i][2][1] = neigh[i][2] * neigh[i][1];
			neighT[i][2][2] = neigh[i][2] * neigh[i][2];
		}
		
		
		double[][][] grd = new double[3][3][3];
		
		for (int i=0; i<3; i++){
			for (int j=0; j<3; j++){
				double[] a = new double[9];
				double[] c = new double[3];
				double e0 = central.getO2()[i][j];
				for (int k=0; k<inter.length; k++){
					a[0] += neighT[k][0][0];
					a[1] += neighT[k][0][1];
					a[2] += neighT[k][0][2];
					
					a[3] += neighT[k][1][0];
					a[4] += neighT[k][1][1];
					a[5] += neighT[k][1][2];
					
					a[6] += neighT[k][2][0];
					a[7] += neighT[k][2][1];
					a[8] += neighT[k][2][2];
					
					double de = inter[k].getO2()[i][j] - e0;
					c[0] += neigh[k][0]*de;
					c[1] += neigh[k][1]*de;
					c[2] += neigh[k][2]*de;
				}
				
				MatrixOps.invert3x3matrix(a,0.001);
				grd[i][j][0] = c[0]*a[0] + c[1]*a[1] + c[2]*a[2];
				grd[i][j][1] = c[0]*a[3] + c[1]*a[4] + c[2]*a[5];
				grd[i][j][2] = c[0]*a[6] + c[1]*a[7] + c[2]*a[8];
			}
		}
		
		double[][] nyeTensor = new double[3][3];
		for (int i=0; i<3; i++){
			nyeTensor[0][i] = -grd[2][i][1] + grd[1][i][2];
			nyeTensor[1][i] = -grd[0][i][2] + grd[2][i][0];
			nyeTensor[2][i] = -grd[1][i][0] + grd[0][i][1];
		}
		return nyeTensor;
	}
	
	public static double product(float[] u, double[] v){
		return u[0]*v[0]+u[1]*v[1]+u[2]*v[2];
	}
	
	private class CoordinateWrapper implements Coordinate {
		private float[] c;
		
		public CoordinateWrapper(float[] c) {
			this.c = c;
		}

		@Override
		public float getX() {
			return c[0];
		}
		
		@Override
		public float getY() {
			return c[1];
		}
		
		@Override
		public float getZ() {
			return c[2];
		}
	}
	
	private class Cuboid{
		public Tupel<float[], double[][]>[] corners;
		public Tupel<float[], double[][]> center;
		public Cuboid(Tupel<float[], double[][]>[] corners, Tupel<float[], double[][]> center) {
			super();
			this.corners = corners;
			this.center = center;
		}
	}
}

