// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.polygrain;

import java.util.*;
import java.util.concurrent.*;

import common.ThreadPool;
import crystalStructures.CrystalStructure;

import model.*;
import model.polygrain.mesh.*;

public class Grain implements Pickable{

	private Mesh mesh;
	private CrystalStructure cs;
	private CrystalRotationTools rotTools;
	private List<Atom> atomsInGrain;
	private int grainNumber, numAtoms;
	private Future<Void> f;
	
	public Grain(Mesh mesh, List<Atom> atomsInGrain, int grainNumber, CrystalStructure cs, float[] boxSize) {
		this.cs = cs;
		this.atomsInGrain = atomsInGrain;
		this.mesh = mesh;
		this.f = ThreadPool.submit(mesh);
		this.grainNumber = grainNumber;
		this.numAtoms = atomsInGrain.size();
		this.rotTools = new CrystalRotationTools(cs, identifyGrainRotation(atomsInGrain, boxSize));
	}
	
	public Grain(Mesh mesh, List<Atom> atomsInGrain, int grainNumber, CrystalStructure cs, float[][] latticeRotation) {
		this.cs = cs;
		this.atomsInGrain = atomsInGrain;
		this.grainNumber = grainNumber;
		this.mesh = mesh;
		this.f = ThreadPool.submit(mesh);
		this.grainNumber = grainNumber;
		this.rotTools = new CrystalRotationTools(cs, latticeRotation);
	}
	
	
	public static void processGrains(final AtomData data){
		if (Configuration.getCrystalStructure().getGrainBoundaryFilterDistance()==0f || 
				data.isMeshImported()) return; 
		
		//The BSPTree implementation is often slower
		final ClosestTriangleSearchAlgorithm ctsa = 
			new ClosestTriangleSearch(Configuration.getCrystalStructure().getGrainBoundaryFilterDistance());
//			new BSPTree(Configuration.getCrystalStructure().getGrainBoundaryFilterDistance());
			
		ArrayList<Triangle> trias = new ArrayList<Triangle>();
		for (Grain g : data.getGrains()){
			for (Triangle t: g.getMesh().getTriangleSet()){
				trias.add(t);
			}
			g.getMesh().finalizeMesh();
		}
		
		Collections.shuffle(trias);
		for (Triangle t : trias){
			ctsa.add(t);
		}
		
		final int defaultType = Configuration.getCrystalStructure().getDefaultType();
		final int surfaceType = Configuration.getCrystalStructure().getSurfaceType();
		
		Vector<Callable<Void>> parallelTasks = new Vector<Callable<Void>>();
		for (int i=0; i<ThreadPool.availProcessors(); i++){
			final int j = i;
			parallelTasks.add(new Callable<Void>() {
				@Override
				public Void call() throws Exception {
					for (int i=j; i<data.getAtoms().size(); i+=ThreadPool.availProcessors()){
						Atom a = data.getAtoms().get(i);
						if (Thread.interrupted()) return null;
						if (a.getType() != defaultType && a.getType() != surfaceType && a.getGrain() != Atom.DEFAULT_GRAIN){
							if (ctsa.isMeshElementWithinThreshold(a)){
								data.getGrains(a.getGrain()).decreaseAtomCount();
								a.setGrain(Atom.IGNORED_GRAIN);
							}
						}
					}
					return null;
				}
			});
		};
		ThreadPool.executeParallel(parallelTasks);	
	}
	
	public Mesh getMesh() {
		if (f!=null){
			try {
				f.get();
				f = null;
			} catch (InterruptedException e) {
				return null;
			} catch (ExecutionException e) {
				e.printStackTrace();
			}
		}
		return mesh;
	}
	
	public int getGrainNumber() {
		return grainNumber;
	}
	
	public CrystalStructure getCrystalStructure() {
		return cs;
	}
	
	public CrystalRotationTools getCystalRotationTools() {
		return rotTools;
	}	
	
	@Override
	public Collection<? extends Object> getHighlightedObjects() {
		return null;
	}
	
	@Override
	public boolean isHighlightable() {
		return false;
	}
	
	/**
	 * Elements in this list are to be deleted once it is not needed anymore
	 * @return
	 */
	public List<Atom> getAtomsInGrain() {
		return atomsInGrain;
	}
	
	/**
	 * Return the number of atoms in the grain
	 * @return
	 */
	public int getNumberOfAtoms() {
		return numAtoms;
	}
	
	/**
	 * Sets number of atoms in the grain, should be used if the grain/mesh is read from
	 * a file and the original number of atoms cannot be reproduced
	 * @param numAtoms
	 */
	public void setNumberOfAtoms(int numAtoms) {
		this.numAtoms = numAtoms;
	}
	
	/**
	 * Assign a new grain number, only effective is CrystalStructure.orderBySize() is true
	 * @param i
	 * @return
	 */
	public void renumberGrain(int i){
		if (Configuration.getCrystalStructure().orderGrainsBySize())
			grainNumber = i;
	}

	/**
	 * Decreases the atom count by one, should be called if an atom's grain associated is removed elsewhere
	 */
	public synchronized void decreaseAtomCount(){
		numAtoms--;
	}
	
	@Override
	public String toString() {
		float[][] o = rotTools.getDefaultRotationMatrix();
		
		StringBuilder sb = new StringBuilder();
		sb.append(String.format("Grain (%d) Volume: %.4f Atoms: %d\n", this.grainNumber,mesh.getVolume(), this.numAtoms));
		sb.append(String.format("(%.3f, %.3f, %.3f)\n", o[0][0], o[0][1], o[0][2]));
		sb.append(String.format("(%.3f, %.3f, %.3f)\n", o[1][0], o[1][1], o[1][2]));
		sb.append(String.format("(%.3f, %.3f, %.3f)\n", o[2][0], o[2][1], o[2][2]));
		return sb.toString();
	}
	
	private float[][] identifyGrainRotation(List<Atom> atoms, float[] boxSize){
		//find an perfect lattice site atom surrounded by only perfect lattice site atoms
		//This will be the reference orientation for the whole grain.
		int defaultType = cs.getDefaultType();
		
		NearestNeighborBuilder<Atom> nnb = new NearestNeighborBuilder<Atom>(boxSize, 
				Configuration.getCrystalStructure().getNearestNeighborSearchRadius(), Configuration.getPbc());
		for (Atom a : atoms) {
			if (a.getType() == defaultType) nnb.add(a);
		}
		
		float[][] rot = null;
		for (Atom a : atoms) {
			if (a.getType() == defaultType) {
				// Identify the local grain rotation
				rot = this.cs.identifyRotation(a, nnb);
			}
			if (rot != null) break;
		}
		
		//if rot is null, then the grain is extremely small or so distorted that no orientation can be identified
		//The default orientation is used in this case
		if (rot == null){
			rot = Configuration.getCrystalOrientation();	
		} 
		
		return rot;
	}
	
	@Override
	public String printMessage(int modifier) {
		return toString();
	}
}
