// Part of AtomViewer: AtomViewer is a tool to display and analyse
// atomistic simulations
//
// Copyright (C) 2013  ICAMS, Ruhr-Universität Bochum
//
// AtomViewer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AtomViewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with AtomViewer. If not, see <http://www.gnu.org/licenses/> 

package model.skeletonizer;

import java.util.*;
import java.util.Map.Entry;
import java.util.concurrent.*;

import model.*;
import model.polygrain.Grain;
import common.ThreadPool;
import common.VectorOps;

/**
 * This implementation of planar defect can identify
 * at this time only stacking faults and twin boundaries in FCC correctly
 * In general, it is only designed to handle defects that are limited to exactly on plane.
 * The more general case of grain- and phaseboundaries are handle in model.polygrain.Grain
 */
public class PlanarDefect implements Pickable{
	/**
	 * Each PlanarDefect instance needs an unique number to easily check for equality
	 * and sorting in sets 
	 */
	private static int uniqueNumber = 0;
	private int number;
	
	private Dislocation[] adjacentDislocations;
	private float[] normal;
	private Atom[] faces = new Atom[0];
	private int numberOfAtoms;
	private Future<Void> faceCreationFuture;
	private Grain grain;
	/**
	 * Each plane is composed of exactly one atom type. 
	 */
	private int planeComposedOfType;
	
	/**
	 * Constructor for import
	 * @param num Unique number for the stacking fault plane
	 * @param planarDefectAtoms The set of atoms in the stacking fault plane
	 * @param normal the planes's normal
	 * @param adjacentDislocations a collection of adjacent dislocations, can be empty but not null
	 * @param grain In case of polycrystalline material, the grain in which the defect exist, may be null in single crystals 
	 */
	public PlanarDefect(int num, Set<Atom> planarDefectAtoms, float[] normal,
			Collection<Dislocation> adjacentDislocations, TreeMap<Atom, PlanarDefect> defectMap, Grain grain){
		this(planarDefectAtoms, normal, defectMap, grain);
		this.number = num;
		this.adjacentDislocations = adjacentDislocations.toArray(new Dislocation[adjacentDislocations.size()]);
		for (Dislocation d : this.adjacentDislocations){
			d.addAdjacentStackingFault(this);
		}
	}
	
	protected PlanarDefect(final Set<Atom> planarDefectAtoms, float[] normal, TreeMap<Atom, PlanarDefect> defectMap, Grain grain) {
		if (planarDefectAtoms.size() < 3) throw new IllegalArgumentException("To small set (need >=3 points)");
		this.grain = grain;
		this.number = uniqueNumber;
		uniqueNumber++;
		
		this.numberOfAtoms = planarDefectAtoms.size();
		this.normal = VectorOps.normalize(normal);
		
		for (Atom a : planarDefectAtoms) {
			defectMap.put(a, this);
		}
		
		Callable<Void> c = new Callable<Void>() {
			@Override
			public Void call() throws Exception {
				createFaces(planarDefectAtoms);
				return null;
			}
		};
		
		faceCreationFuture = ThreadPool.submit(c);
		
		if (numberOfAtoms!=0)
			this.planeComposedOfType = planarDefectAtoms.iterator().next().getType();
	}
	
	/**
	 * Creates planar defects 
	 * nearest neighbor atoms between defects need to be created before calling this method
	 * Important note: the relationship between planar defects and atoms in returned in the defectMap!
	 * @param planarDefectAtoms input of all possible planar defect atoms 
	 * @param defectMap contains the relationship between planar defects and atoms AFTER completion
	 * @return a list of all created planar defects
	 */
	public static ArrayList<PlanarDefect> createPlanarDefects(TreeSet<Atom> planarDefectAtoms, TreeMap<Atom, PlanarDefect> defectMap){
		if (!ImportStates.POLY_MATERIAL.isActive()){
			return createPlanarDefects(planarDefectAtoms, 
					Configuration.getCrystalStructure().getStackingFaultNormals(Configuration.getCrystalRotationTools()), defectMap, null);
		} else {
			//For polycrystalline material, split everything into separate grains
			HashMap<Integer, TreeSet<Atom>> sfPerGrain = new HashMap<Integer, TreeSet<Atom>>();
			for (Atom a : planarDefectAtoms){
				if ((!ImportStates.POLY_MATERIAL.isActive() || a.getGrain() != Atom.IGNORED_GRAIN) && !sfPerGrain.containsKey(a.getGrain())){
					sfPerGrain.put(a.getGrain(), new TreeSet<Atom>());
				}
				sfPerGrain.get(a.getGrain()).add(a);
			}
			planarDefectAtoms.clear();	//Not needed anymore, dereference all its members
			
			ArrayList<PlanarDefect> def = new ArrayList<PlanarDefect>();
			
			for (Entry<Integer, TreeSet<Atom>> e : sfPerGrain.entrySet()){
				sfPerGrain.remove(e);	//reference in the map is not needed anymore
				if (!ImportStates.POLY_MATERIAL.isActive() || (e.getKey() != Atom.DEFAULT_GRAIN  && e.getKey() != Atom.IGNORED_GRAIN)){
					Grain g = Configuration.getCurrentAtomData().getGrains(e.getKey()); 
					
					def.addAll(createPlanarDefects(e.getValue(), 
							Configuration.getCrystalStructure().getStackingFaultNormals(g.getCystalRotationTools()),
							defectMap, g));
				}
			}
			
			return def;
		}
	}
	
	/**
	 * @param planarDefectAtoms
	 * @param planeNormals
	 * @return
	 */
	public static ArrayList<PlanarDefect> createPlanarDefects(TreeSet<Atom> planarDefectAtoms, 
			float[][] planeNormals, TreeMap<Atom, PlanarDefect> defectMap, Grain grain){
		HashSet<Atom> planarDefectAtomsLookup = new HashSet<Atom>(planarDefectAtoms);
		ArrayList<PlanarDefect> planarDefectList = new ArrayList<PlanarDefect>();
		
		TreeSet<Atom> leftOverAtoms = findDefects(planarDefectAtoms, true, planarDefectList, planarDefectAtomsLookup, planeNormals, defectMap, grain);
		findDefects(leftOverAtoms, false, planarDefectList, planarDefectAtomsLookup, planeNormals, defectMap, grain);

		//wait till all planarDefects are processed in worker threads
		for (PlanarDefect pd : planarDefectList)
			pd.getFaces();
		
		return planarDefectList;
	}
	
	private static TreeSet<Atom> findDefects(TreeSet<Atom> atomSet, boolean firstRun, ArrayList<PlanarDefect> planarDefectList, 
			HashSet<Atom> planarDefectAtomsLookup, float[][] planeNormals, TreeMap<Atom, PlanarDefect> defectMap, Grain grain){
		
		int bondThreshold = firstRun ? 3 : 2;
		Atom[] fittingArray = new Atom[20];
	
		TreeSet<Atom> leftOverAtoms = new TreeSet<Atom>();
		
		while (!atomSet.isEmpty()){
			Atom c = atomSet.pollFirst();
			
			//Identify straight bonds between nearest neighbors in the set of planar defects
			//In the first run consider only atoms in the list of planar defects
			//In the second run all atoms are considered
			ArrayList<float[]> straightBonds = new ArrayList<float[]>(); 
			for (int i=0; i<c.getNeigh().length-1; i++){
				if (!firstRun || planarDefectAtomsLookup.contains(c.getNeigh()[i])){
					
					float[] vec1 = Configuration.pbcCorrectedDirection(c, c.getNeigh()[i]);
					float vec1_sqrlength = (vec1[0]*vec1[0]+vec1[1]*vec1[1]+vec1[2]*vec1[2]);
					float v1LengthTimesThreshold = 0.9025f * vec1_sqrlength;
					
					for (int j=i+1; j<c.getNeigh().length; j++){
						if (!firstRun || planarDefectAtomsLookup.contains(c.getNeigh()[j])){
							
							float[] vec2 = Configuration.pbcCorrectedDirection(c, c.getNeigh()[j]);
							float p = vec1[0]*vec2[0]+vec1[1]*vec2[1]+vec1[2]*vec2[2];
							
							if (p < 0f && (p*p) > v1LengthTimesThreshold * VectorOps.lengthSqr(vec2)){ 
								straightBonds.add(Configuration.pbcCorrectedDirection(c.getNeigh()[i], c.getNeigh()[j]));
							}
						}
					}
				}
			}
			if (straightBonds.size()>=bondThreshold){
				//Three straight bonds are found, now test if they are in the same plane or forming multiple planes.
				//If there is only one plane, start BFS to detect the whole surface.
				float[] normal = VectorOps.normalize(VectorOps.crossProduct(straightBonds.get(0),straightBonds.get(1)));
				if (bondThreshold == 3){
					float thirdBondInSamePlane = Math.abs(VectorOps.product(straightBonds.get(2), normal));
					if (thirdBondInSamePlane > 0.1f) continue; //Not a single plane -> do nothing
				}
				
				TreeSet<Atom> atomsInSet = new TreeSet<Atom>();
				Queue<Atom> atomsQueue = new LinkedList<Atom>();
				
				//Surface is detected, now test which glide plane fits best
				int t=0;
				float best = 0;
				for (int i=0; i<planeNormals.length;i++){
					float sim = Math.abs(VectorOps.product(normal, planeNormals[i]));
					if (sim>best){
						best = sim;
						t = i;
					}
				}
				
				//BFS to find the surface atoms
				atomsInSet.add(c);
				atomsQueue.add(c);
				while (!atomsQueue.isEmpty()){
					Atom b = atomsQueue.poll();
					int fittingAtoms = 0;
					int fittingArraySize = 0;
					
					for (int i=0; i<b.getNeigh().length; i++){
						Atom n = b.getNeigh()[i];
						
						//Count nearest neighbor hcp-atoms in the same plane 
						if (planarDefectAtomsLookup.contains(n)){
							// equals (vec/|vec|)*normal < 0.2, but this method is more efficient 
							float[] vec = Configuration.pbcCorrectedDirection(b, n);
							float p = vec[0]*normal[0]+vec[1]*normal[1]+vec[2]*normal[2];
							if ( p*p < 0.04f * VectorOps.lengthSqr(vec) && n.getType() == c.getType()){
								fittingAtoms++;
								if (!atomsInSet.contains(n)){
									fittingArray[fittingArraySize] = n;
									fittingArraySize++;
								}
							}
						}
					}
					//if the atom has at least three neighbors in the same plane, the surface detection can proceed at this atom.
					//If not, the detection should stop here, otherwise the plane might extend itself as a line in other planes
					//Atoms that are ignored in first step by this are attached later to their planes.
					if (fittingAtoms>=3){
						for (int i=0; i<fittingArraySize; i++){
							atomSet.remove(fittingArray[i]);
							atomsInSet.add(fittingArray[i]);
							atomsQueue.add(fittingArray[i]);
						}
					}
				}
				if (atomsInSet.size()>=3){
					planarDefectList.add(new PlanarDefect(atomsInSet, planeNormals[t], defectMap, grain));
					planarDefectAtomsLookup.removeAll(atomsInSet);
				}
				
			}
			else leftOverAtoms.add(c);
		}
		return leftOverAtoms;
	}
	
	/**
	 * Resets internal counters, to get an deterministic behavior if a file is loaded twice
	 */
	protected static void reinit(){
		uniqueNumber = 0;
	}

	/**
	 * Returns an array of adjacent dislocations
	 * Warning: Might be null if they have not been created
	 * @return 
	 */
	public Dislocation[] getAdjacentDislocations() {
		return adjacentDislocations;
	}

	/**
	 * Each plane consists of only one type of atoms
	 * @return the type of all atoms in this planar defect
	 */
	public int getPlaneComposedOfType() {
		return planeComposedOfType;
	}

	/**
	 * Note: To find adjacent dislocations for planar defects, first find the adjacent planar defects
	 * for dislocations
	 * @param dislocations a collection of all adjacent dislocations
	 */
	protected void findAdjacentDislocations(Collection<Dislocation> dislocations) {
		ArrayList<Dislocation> contactDis = new ArrayList<Dislocation>();
		for (Dislocation dis : dislocations){
			if (dis.isPlanarDefectAdjacent(this)) contactDis.add(dis);
		}
		this.adjacentDislocations = contactDis.toArray(new Dislocation[contactDis.size()]);
	}
	
	/**
	 * The plane's surface normal in xyz-coordinates (not in crystal-coordinates) 
	 * @return
	 */
	public float[] getNormal() {
		return normal;
	}
	
	/**
	 * The unique number of the stacking fault plane
	 * @return
	 */
	public int getNumber() {
		return number;
	}
	
	
	private void createFaces(Set<Atom> planarDefectAtoms){
		ArrayList<Atom> faces = new ArrayList<Atom>();
		//Find three atoms in triangular configuration in the same plane
		//if a triangle is found and the first atom has the highest number,
		//save the triangle in the list and for later rendering
		//Creating only triangles if the first atom has the highest number 
		//ensures that the same triangle is not created twice
		for (Atom a : planarDefectAtoms){
			for (int i=0; i<a.getNeigh().length; i++){
				Atom n1 = a.getNeigh()[i];
				if (a.getNumber() > n1.getNumber() && planarDefectAtoms.contains(a.getNeigh()[i])){
					//Look for common neighbors
					for (int j=i; j<a.getNeigh().length; j++){
						for (int k=0; k<n1.getNeigh().length; k++){
							if (a.getNeigh()[j] == n1.getNeigh()[k] && 
									a.getNumber()>a.getNeigh()[j].getNumber() && 
									planarDefectAtoms.contains(a.getNeigh()[j])){
								faces.add(a);
								faces.add(n1);
								faces.add(a.getNeigh()[j]);
							}
						}	
					}
				}
			}
		}
		
		
		this.faces = faces.toArray(new Atom[faces.size()]);
	}
	
	
	/**
	 * Get a transformed representation of the planar defect's atoms as a list of triangles for rendering the 
	 * plane.
	 * The list contains triples of three atoms (storing the triangle's vertices) for each face. 
	 * @return
	 */
	public Atom[] getFaces(){
		//Make sure the faces have been created in the worker thread
		while (!faceCreationFuture.isDone()){
			try {
				Thread.sleep(50);
			} catch (InterruptedException e) { e.printStackTrace(); }
		}
		return faces;
	}
	
	public float getArea(){
		float area = 0f;
		for (int i=0; i<getFaces().length/3;i++){
			float[] d1 = VectorOps.sub(faces[i*3+1], faces[i*3]);
			float[] d2 = VectorOps.sub(faces[i*3+2], faces[i*3]);
			area += 0.5f * VectorOps.length(VectorOps.crossProduct(d1,d2));
		}
		return area;
	}
	
	@Override
	public String toString() {
		float[] o;
		if (grain == null)
			o = Configuration.getCrystalRotationTools().getInCrystalCoordinates(normal);
		else 
			o = grain.getCystalRotationTools().getInCrystalCoordinates(normal);
		o = VectorOps.multiply(o, Configuration.getCrystalStructure().getLatticeConstant());
		return String.format("Planar defect (%d) Normal=(%.3f,%.3f,%.3f) in crystal axis=(%.3f,%.3f,%.3f) #Atoms=%d Area=%f", 
				number, normal[0], normal[1], normal[2], o[0], o[1], o[2], numberOfAtoms, getArea());
	}
	
	@Override
	public Collection<? extends Object> getHighlightedObjects() {
		if (adjacentDislocations == null) return null;
		ArrayList<Dislocation> d = new ArrayList<Dislocation>(adjacentDislocations.length);
		for (int i=0; i<adjacentDislocations.length; i++)
			d.add(adjacentDislocations[i]);
		return d;
	}
	
	@Override
	public boolean isHighlightable() {
		return true;
	}
	
	@Override
	public String printMessage(int modifier) {
		return toString();
	}
}
